/********************/

// Digicube v1.2 - programmable virtual Rubik's cube 3x, cube 2x, pyramid 3x
// (C) Copyright 2018, 2022 Andrei Sorin <a.sorin@andsor.ca>
// Information and download: www.softwareandmind.com/digicube.html
// Documentation files: manual.pdf, notes.txt

// This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
//   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.

/********************/

// set win32 1 or 0 to create 32-bit or 16-bit version
#define win32 1

#define  uns   unsigned short
#define  char  unsigned char
#define  dbl   double
typedef  int (*ptf)();

#define  sp55       55
#define  max_moves  20
#define  max_imov   30
#define  max_rec    160
#define  num_argv   16
#define  num_pairs  4
#define  num_fpairs 2

#define  sz_rep     2000
#define  sz_kbuf    130
#define  sz_cmbuf   130
#define  sz_rdbuf   400
#define  sz_flnm    100
#define  sz_lstr    21
#define  sz_dstr    7
#define  sz_estr    14
#define  sz_ad   (6*9)
#define  sz_fc   (6*4)
#define  sz_py   (4*6)
#define  sz_max  sz_ad

#define  m_ad  1
#define  m_fc  2
#define  m_py  3
#define  m_cube (mdl == m_ad || mdl == m_fc)
#define  m_pyramid (mdl == m_py)

#define  eol(c) (c == '\n' || c == '\r')
#define  spc(c) (c == ' ' || c == '\t')
#define  dig(c) (c >= '0' && c <= '9')

// C library functions called: putch, getch, kbhit, _cgets, creat, open, close,
// read, write, lseek, gettime, setjmp, longjmp, exit

// all functions (incl. library) return void or the default int, except these
dbl cb_frand();  long lseek(), ds_lseek(), cb_time(), ds_time();

// padding for data alignment (before all data), adjust fill1 length 0+ for shortest execution time
//CHAR FILL1[2] = { 0,0 };
// compiler inserts here all global and local initialized data, before the other variables

char cnote[] = "Digicube v1.2 & D16cube v1.2, Copyright (C) 2018, 2022 Andrei Sorin";

// dword-aligned variables, adjust fill2 padding accordingly if needed
//CHAR FILL2[2];
char seqtot[8];

// dword-aligned memory area shared by all models
// pointers initialized in cb_init as 64-byte areas, used as follows:
// ad, ad_goal: char [6][9]  ad_save: max_moves x char [6][9]
// fc, fc_goal: char [6][4]  fc_save: max_moves x char [6][4]
// py, py_goal: char [4][6]  py_save: max_moves x char [4][6]
char memcom[(2+max_moves)*64];
char *ad, *ad_goal, *ad_save;
char *fc, *fc_goal, *fc_save;
char *py, *py_goal, *py_save;

long jmpbuf[2][10];  short cmargc;  uns s_ds;
char cmbuf[sz_cmbuf], *cmargv[num_argv];

short mdl, mstop, custom, disp, stlabel, optlabel;
char mempos[99*sz_max], cmem[99], gmem[99], initpos[sz_max], initgoal[sz_max],
  wprev[sz_max], *w_pp;

short im, lno, step, stype, moveord, movetot, elvl, edif, wilpos;
char move[max_moves];  ptf ycomp;

long ofst;  short rdsz;  uns wrhdl, rdhdl;
char wrfile[sz_flnm+5], rdfile[sz_flnm+5], kbuf[sz_kbuf], rdbuf[sz_rdbuf],
  *rdptr, *kptr, *ekptr;

long cycle, mcount, tcount, ccount;

short recn, reclast, recseq;  char rmove[max_rec];

long repnum;  short repn, replast;  char rval[sz_rep];

dbl bktot, bkcount;  long bkseqs;
short bkwrong, bktype, bkflip, bkstnum;
char bkpcs[28], bklth[10], bkstmove[4], bkrev, bkfill1;

short tm_day, tm_hour;  struct tm_time { char mm, hh, cc, ss; };

long ran_seed[2], sav_seed[2];  char ran_prev[2][2][2], sav_prev[2][2][2];

ptf x_chk, x_gchk, x_solve, x_move, x_turn, x_copy, x_reset, x_gen;
short x_sz, x_fsz, x_npcs;
char *x_dd, *x_oo, *x_gg, *x_nn, *x_cn, *x_cl, *x_cp, *x_ci, *x_ef, *x_cf;

long i_rnum;  short i_mord, i_stop, i_mem, i_loc, i_col, i_mnum, i_optn;
char *i_ept, i_lstr[sz_lstr], i_estr[sz_estr], i_fpos[9], i_mov[max_imov],
  i_stype, i_goal, i_face, i_scr, i_ran, i_wil, i_opt;

char k_ent, k_end, k_wrd, k_wrf, k_nowrf, k_out, k_inp, k_eol, k_fin, k_icmp,
  k_rep, k_rex, k_wil, k_schem, k_pcs, k_rows, k_rev, k_wide, k_dnl, k_six,
  k_int, k_irun, k_read, k_solve, k_gen;

char crev[13] = { 0,2,1,4,3,6,5,8,7,10,9,12,11 }; // reverse of 1 - 12 (ad & fc moves, turns)
char yrev[9] = { 0,5,6,7,8,1,2,3,4 }; // reverse of 1 - 8 (py moves, turns)
char dstop[4] = { 0,15,13,11 }; // default mstop (ad fc py)
char cstep[10] = { 0,2,3,6,0,0,0,0,0,4 }; // title codes for step types 0-9 (see cb_wide_disp)

char ad_epair[num_pairs][3], ad_flist[num_fpairs][3];

char ad_faces[24][6] = // 24 valid cube orientations (face numbers in standard face order)
  { 1,2,3,4,5,6,  1,2,4,3,6,5,  1,2,5,6,4,3,  1,2,6,5,3,4,
    2,1,3,4,6,5,  2,1,4,3,5,6,  2,1,5,6,3,4,  2,1,6,5,4,3,
    3,4,1,2,6,5,  3,4,2,1,5,6,  3,4,5,6,1,2,  3,4,6,5,2,1,
    4,3,1,2,5,6,  4,3,2,1,6,5,  4,3,5,6,2,1,  4,3,6,5,1,2,
    5,6,1,2,3,4,  5,6,2,1,4,3,  5,6,3,4,2,1,  5,6,4,3,1,2,
    6,5,1,2,4,3,  6,5,2,1,3,4,  6,5,3,4,1,2,  6,5,4,3,2,1 };
char ad_rturns[24][3] = // 24 turn sequences to reset cube from the 24 ad_faces orientations
  { 0,0,0,  1,1,0,  2,0,0,  1,0,0,
    3,3,0,  5,5,0,  2,5,5,  1,5,5,
    3,3,5,  6,0,0,  2,6,0,  1,6,0,
    5,0,0,  3,3,6,  2,5,0,  1,5,0,
    3,5,0,  3,6,0,  3,0,0,  3,5,5,
    4,5,0,  4,6,0,  4,0,0,  4,5,5 };

char ad_cn[6][9] = // piece numbers in ad array order
  { 9,6,3,18,15,12,27,24,21,   1, 4, 7,10,13,16,19,22,25,  7,8,9,16,17,18,25,26,27,
    3,2,1,12,11,10,21,20,19,  25,26,27,22,23,24,19,20,21,  1,2,3, 4, 5, 6, 7, 8, 9 };
char ad_cl[27] = // piece colors/locations in piece number order
  { 246,46,146,26,6,16,236,36,136,  24,4,14,2,0,1,23,3,13,  245,45,145,25,5,15,235,35,135 };
char ad_cp[27] = // piece numbers in colors/location numerical order
  { 14,15,13,17,11,23,5,18,12,24,6,16,10,22,4,26,8,20,2,27,9,21,3,25,7,19,1 };
char ad_ci[27] = // index into ad_ef or ad_cf table in piece number order
  { 7,11,3,7,0,3,5,9,1,  5,0,1,0,0,0,4,0,0,  6,10,2,6,0,2,4,8,0 };
char ad_ef[12][2] = // 12 edge pieces offset in ad array in colors/location numerical order
  { 3,23,  5,30,  7,41,  1,50,  14,21,  12,32,  16,39,  10,48,  25,37,  19,52,  34,43,  28,46 };
char ad_cf[8][3] = // 8 corner pieces offset in ad array in colors/location numerical order
  { 6,26,38,  0,20,53,  8,33,44,  2,27,47,  17,24,36,  11,18,51,  15,35,42,  9,29,45 };
short ad_edge[24] = // valid colors for 12 edge pieces
  { 16,61,  13,31,  15,51,  14,41,  36,63,  35,53,  26,62,  24,42,  25,52,  23,32,  46,64,  45,54 };
short ad_corner[48] = // valid colors for 8 corner pieces
  { 136,361,613,  153,531,315,  145,451,514,  164,641,416, // 0 - 23 valid in locations 136 145 235 246
    235,352,523,  263,632,326,  246,462,624,  254,542,425,
    163,631,316,  135,351,513,  154,541,415,  146,461,614, // 24 - 47 valid in locations 135 146 236 245
    253,532,325,  236,362,623,  264,642,426,  245,452,524 };

char fc_cn[6][4] = // piece numbers in fc array order
  { 4,2,8,6,  1,3,5,7,  3,4,7,8,  2,1,6,5,  7,8,5,6,  1,2,3,4 };
char fc_cl[8] = // piece colors/locations in piece number order
  { 246,146,236,136,  245,145,235,135 };
char fc_cp[8] = // piece numbers in colors/location numerical order
  { 8,4,6,2,7,3,5,1 };
char fc_ci[8] = // index into fc_cf table in piece number order
  { 7,3,5,1,  6,2,4,0 };
char fc_cf[8][3] = // 8 corner pieces offset in fc array in colors/location numerical order
  { 2,11,17,  0,9,23,  3,14,19,  1,12,21,  7,10,16,  5,8,22,  6,15,18,  4,13,20 };
short fc_corner[48] = // valid colors for 8 corner pieces
  { 136,361,613,  153,531,315,  145,451,514,  164,641,416, // 0 - 23 valid in locations 136 145 235 246
    235,352,523,  263,632,326,  246,462,624,  254,542,425,
    163,631,316,  135,351,513,  154,541,415,  146,461,614, // 24 - 47 valid in locations 135 146 236 245
    253,532,325,  236,362,623,  264,642,426,  245,452,524 };

char py_cn[4][6] = // piece numbers in py array order
  { 1,3,2,5,6,8,  2,3,4,8,9,10,  4,3,1,10,7,5,  6,8,9,5,7,10 };
char py_cl[10] = // piece colors/locations in piece number order
  { 13,12,123,23,  134,14,34,124,24,234 };
char py_cp[10] = // piece numbers in colors/location numerical order
  { 2,1,6,4,9,7,3,8,5,10 };
char py_ci[10] = // index into py_ef or py_cf table in piece number order
  { 1,0,0,3,  2,2,5,1,4,3 };
char py_ef[6][2] = // 6 edge pieces offset in py array in colors/location numerical order
  { 2,6,  0,14,  4,18,  8,12,  10,20,  16,22 };
char py_cf[4][3] = // 4 axial pieces offset in py array in colors/location numerical order
  { 1,7,13,  5,9,19,  3,17,21,  11,15,23 };
short py_edge[12] = // valid colors for 6 edge pieces
  { 12,21,  13,31,  14,41,  23,32,  43,34,  24,42 };
short py_axial[24] = // valid colors for 4 axial pieces
  { 123,231,312,  134,341,413,  142,421,214,  243,432,324, // 0 - 11 valid in locations 123 134
    132,321,213,  143,431,314,  124,241,412,  234,342,423 }; // 12 - 23 valid in locations 124 234

char ad_orig[] = // original position
  { 1,1,1,1,1,1,1,1,1,  2,2,2,2,2,2,2,2,2,  3,3,3,3,3,3,3,3,3,
    4,4,4,4,4,4,4,4,4,  5,5,5,5,5,5,5,5,5,  6,6,6,6,6,6,6,6,6 };
char ad_anchor[] = // goal: anchor (corner 246 and edges 24 26 46)
  { 9,9,9,9,1,9,9,9,9,  2,2,9,2,2,9,9,9,9,  9,9,9,9,3,9,9,9,9,
    9,4,4,9,4,4,9,9,9,  9,9,9,9,5,9,9,9,9,  6,6,9,6,6,9,9,9,9 };
char ad_acorners[] = // goal: anchor + corners
  { 1,9,1,9,1,9,1,9,1,  2,2,2,2,2,9,2,9,2,  3,9,3,9,3,9,3,9,3,
    4,4,4,9,4,4,4,9,4,  5,9,5,9,5,9,5,9,5,  6,6,6,6,6,9,6,9,6 };
char ad_generic[] = // goal: all 9's
  { 9,9,9,9,1,9,9,9,9,  9,9,9,9,2,9,9,9,9,  9,9,9,9,3,9,9,9,9,
    9,9,9,9,4,9,9,9,9,  9,9,9,9,5,9,9,9,9,  9,9,9,9,6,9,9,9,9 };

char fc_orig[] = // original position
  { 1,1,1,1,  2,2,2,2,  3,3,3,3,  4,4,4,4,  5,5,5,5,  6,6,6,6 };
char fc_anchor[] = // goal: anchor (corner 246)
  { 9,9,9,9,  2,9,9,9,  9,9,9,9,  9,4,9,9,  9,9,9,9,  6,9,9,9 };
char fc_generic[] = // goal: all 9's
  { 9,9,9,9,  9,9,9,9,  9,9,9,9,  9,9,9,9,  9,9,9,9,  9,9,9,9 };

char py_orig[] = // original position
  { 1,1,1,1,1,1,  2,2,2,2,2,2,  3,3,3,3,3,3,  4,4,4,4,4,4 };
char py_aaxials[] = // goal: axials
  { 9,1,9,1,9,1,  9,2,9,2,9,2,  9,3,9,3,9,3,  9,4,9,4,9,4 };
char py_generic[] = // goal: all 9's
  { 9,9,9,9,9,9,  9,9,9,9,9,9,  9,9,9,9,9,9,  9,9,9,9,9,9 };

// initial positions for benchmark solution
// times shown are with 32-bit version, Windows 10, Intel i5-7500 3.4 GHz
char ad_init0[] = // 2.2 sec
  { 3,6,2,6,1,4,5,2,6,  1,2,1,4,2,1,1,3,1,  3,5,5,3,3,1,4,5,2,
    6,5,6,2,4,1,3,4,6,  5,4,4,2,5,6,3,6,2,  4,1,4,5,6,3,5,3,2 };
char fc_init0[] = // 1.7 sec
  { 4,3,1,2,  2,1,4,3,  3,1,6,5,  2,4,6,5,  1,4,2,3,  6,5,5,6 };
char py_init0[] = // 1.4 sec
  { 2,3,1,4,4,3,  3,1,4,1,1,1,  2,2,3,2,1,3,  3,4,2,2,4,4 };

char vbase[] = { 0,0 }; // base marker for offsets shown in obj2asm listing

/********************/
main(argc, argv) short argc; char *argv[];
{
short i, n, sz;
char *cp, *cs, *cd, ch, kopt, a_x, a_h, a_l, a_c, a_z, a_g, a_m, a_s, a_d, a_a,
  a_u, a_j, a_r, a_w, a_n, a_k, a_i, a_o;

_asm  mov word ptr s_ds, ds  // s_ds is needed only in asm sections
w_pp = wprev;
k_ent = 0;
setjmp(jmpbuf[0]);

s0: kopt = 0; // start run
s1: a_x = a_h = a_l = a_c = a_z = a_g = a_m = a_s = a_d = a_a = a_u = a_j = a_r
  = a_w = a_n = a_k = a_i = a_o = 0;
k_schem = k_pcs = k_rows = k_rev = k_wide = 0;
mdl = 1; custom = disp = stlabel = optlabel = moveord = 0;
k_end = k_wrf = k_nowrf = k_out = k_inp = k_wil = k_dnl = k_six = k_int
  = k_irun = k_read = k_solve = k_gen = 0;
k_wrd = k_eol = 1;

cb_frand_init(); cb_ran_save();

if (!kopt) // skip if return to s1 was due to option g
{
// reconstruct argv[] in keyboard buffer, skip argv[0]
*kbuf = 0; cp = kbuf + 1;
for (i = 1; i < argc; i++)
  {
  sz = fl_lth(cs = argv[i]); if (cp + sz > kbuf + sz_kbuf - 3) goto s84;
  fl_move(cs, cp, sz); cp += sz; (*kbuf) += sz;
  if (i < argc - 1) { *cp++ = ' '; (*kbuf)++; }
  }
*cp = 0;

if (argc < 2) // options not entered with command
  {
  k_ent = 1; fl_nl();
  fl_disp("Enter run options (X to end, H for help, I for interactive): ");
  fl_kbstr();
  }
}

// copy options from keyboard buffer to cmbuf/cmargc/cmargv
cs = kbuf + 1; cd = cmbuf;
for (cmargc = 0; *cs != 0; cmargc++)
{
if (cmargc == num_argv) goto s84; cmargv[cmargc] = cd; n = 0;
if (cmargc == 0) { *cd++ = 0; continue; } // create empty cmargv[0]
if (((ch = fl_up(*cs)) == 'I' || ch == 'O') && *(cs + 1) == '"') n = 1;
while (1)
  {
  if (cd >= cmbuf + sz_cmbuf - 1) goto s84;
  if ((ch = *cs) == '"')
    if (n == 1) { n = 2; cs++; continue; } else if (n == 2) { cs++; break; }
  if (ch == 0 || spc(ch) && n == 0) break; *cd++ = ch; cs++;
  }
*cd++ = 0; while (1) { ch = *cs; if (!spc(ch)) break; cs++; }
}

for (i = 1; i < cmargc; i++)
{
sz = fl_lth(cp = cmargv[i]);
switch (fl_up(*cp))
  {
  case 'X': if (a_x) goto s82; if (sz > 1) goto s81; a_x = i; break;
  case 'H': if (a_h) goto s82; if (sz > 1) goto s81; a_h = i; break;
  case 'L': if (a_l) goto s82; if (sz > 1) goto s81; a_l = i; break;
  case 'C': if (a_c) goto s82; a_c = i; break;
  case 'Z': if (a_z) goto s82; if (sz > 1) goto s81; a_z = i; break;
  case 'G': if (a_g) goto s82; a_g = i; break;
  case 'M': if (a_m) goto s82; a_m = i; break;
  case 'S': if (a_s) goto s82; a_s = i; break;
  case 'D': if (a_d) goto s82; a_d = i; break;
  case 'A': if (a_a) goto s82; if (sz > 1) goto s81; a_a = i; break;
  case 'U': if (a_u) goto s82; if (sz > 1) goto s81; a_u = i; break;
  case 'J': if (a_j) goto s82; if (sz > 1) goto s81; a_j = i; break;
  case 'R': if (a_r) goto s82; if (sz > 1) goto s81; a_r = i; break;
  case 'W': if (a_w) goto s82; if (sz > 1) goto s81; a_w = i; break;
  case 'N': if (a_n) goto s82; if (sz > 1) goto s81; a_n = i; break;
  case 'K': if (a_k) goto s82; if (sz > 1) goto s81; a_k = i; break;
  case 'I': if (a_i) goto s82;
    if (sz == 1) k_irun = 1; if (sz > sz_flnm + 1) goto s81; a_i = i; break;
  case 'O': if (a_o) goto s82;
    if (sz == 1 || sz > sz_flnm + 1) goto s81; a_o = i; break;
  default: goto s83;
  }
}

if (a_x) exit();
if (a_h) { fl_nl(); cb_help_run(); goto s8; }
if (a_l) { fl_nl(); cb_help_op(); goto s8; }
if (a_c) { if (*((cp = cmargv[a_c]) + 1) == 0) { fl_nl(); cb_help_cus(); goto s8; }
  if ((n = fl_getnum(cp + 1)) < 1 || n > 99) goto s81; custom = n; }
if (a_z) { fl_nl(); cb_help_lic(); goto s8; }
if (a_g) { if ((n = fl_getnum((cp = cmargv[a_g]) + 1)) < 1 || n > 999) goto s81;
  optlabel = n; }
if (a_m) { if ((n = fl_getnum((cp = cmargv[a_m]) + 1)) < 1 || n > 3) goto s81;
  mdl = n; }
if (a_d) { if (*((cp = cmargv[a_d]) + 1) == 0) disp = 1;
    else { if ((n = fl_getnum(cp + 1)) < 1 || n > 2) goto s81; disp = n + 1; } }
if (a_s) { if ((n = fl_getnum((cp = cmargv[a_s]) + 1)) < 1 || n > 999) goto s81;
  stlabel = n; }
if (a_a) k_schem = 1;
if (a_u) k_pcs = 1;
if (a_j) k_rows = 1;
if (a_r) k_rev = 1;
if (a_w) k_wide = 1;
if (a_k) k_nowrf = 1;

if (optlabel > 0)
{
if (kopt) goto s85;
fl_copy("digopt.txt", rdfile); fl_open_inp(); if (cb_rdopt()) goto s84;
fl_close_inp(); kopt = 1; goto s1;
}

fl_copy(a_i && !k_irun ? cmargv[a_i] + 1 : "diginp", rdfile);
fl_copy(a_o ? cmargv[a_o] + 1 : "digout", wrfile);
fl_copy(".txt", rdfile + fl_lth(rdfile));
fl_copy(".txt", wrfile + fl_lth(wrfile));
fl_lowstr(rdfile); fl_lowstr(wrfile);

if (fl_open_out(a_n > 0)) goto s8;
cb_init(); if (k_ent) fl_nl();
k_wrf = 1;

if (custom == 0)
  { if (!k_irun) { fl_open_inp(); cb_inp_start(); } cb_inp_read(); }
// never returns from cb_inp_read

if (mdl == m_ad) ad_custom();
if (mdl == m_fc) fc_custom();
if (mdl == m_py) py_custom();
k_end = 1; cb_exit();

s8: fl_nl(); if (k_ent) goto s0; exit();
s81: fl_disp("Invalid option: "); fl_disp(cp); cb_exit();
s82: fl_disp("Duplicate option: "); fl_disp(cp); cb_exit();
s83: fl_disp("Unidentified option: "); fl_disp(cp); cb_exit();
s84: fl_disp("Too many options"); cb_exit();
s85: fl_disp("Redirection invalid in options file"); cb_exit();
}

/********************/

/*** ad functions ***/

/********************/
ad_custom()
{
switch (custom)
{
case 1: // solve random position
  cb_scramble(0); ad_solve(0); break;

case 2: // solve several random positions
  cu_positions(); break;

case 3: // solve benchmark position
  ad_copy(ad_init0, ad); ad_solve(0); break;

case 4: // list back sequences
  cu_backseq(); ad_bk_trials(); break;

case 5: // repeat script
  cu_script(); break;
}
}

/********************/
ad_flip()
{
static char tcomb[36][2] = // 36 combinations, 2 out of 9 edge pieces
  { 13,14,  13,15,  13,16,  13,23,  13,25,  13,35,  13,36,  13,45,  14,15,
    14,16,  14,23,  14,25,  14,35,  14,36,  14,45,  15,16,  15,23,  15,25,
    15,35,  15,36,  15,45,  16,23,  16,25,  16,35,  16,36,  16,45,  23,25,
    23,35,  23,36,  23,45,  25,35,  25,36,  25,45,  35,36,  35,45,  36,45 };
static char fcomb[126][4] = // 126 combinations, 4 out of 9 edge pieces
  { 13,14,15,16,  13,14,15,23,  13,14,15,25,  13,14,15,35,  13,14,15,36,  13,14,15,45,  13,14,16,23,  13,14,16,25,  13,14,16,35,
    13,14,16,36,  13,14,16,45,  13,14,23,25,  13,14,23,35,  13,14,23,36,  13,14,23,45,  13,14,25,35,  13,14,25,36,  13,14,25,45,
    13,14,35,36,  13,14,35,45,  13,14,36,45,  13,15,16,23,  13,15,16,25,  13,15,16,35,  13,15,16,36,  13,15,16,45,  13,15,23,25,
    13,15,23,35,  13,15,23,36,  13,15,23,45,  13,15,25,35,  13,15,25,36,  13,15,25,45,  13,15,35,36,  13,15,35,45,  13,15,36,45,
    13,16,23,25,  13,16,23,35,  13,16,23,36,  13,16,23,45,  13,16,25,35,  13,16,25,36,  13,16,25,45,  13,16,35,36,  13,16,35,45,
    13,16,36,45,  13,23,25,35,  13,23,25,36,  13,23,25,45,  13,23,35,36,  13,23,35,45,  13,23,36,45,  13,25,35,36,  13,25,35,45,
    13,25,36,45,  13,35,36,45,  14,15,16,23,  14,15,16,25,  14,15,16,35,  14,15,16,36,  14,15,16,45,  14,15,23,25,  14,15,23,35,
    14,15,23,36,  14,15,23,45,  14,15,25,35,  14,15,25,36,  14,15,25,45,  14,15,35,36,  14,15,35,45,  14,15,36,45,  14,16,23,25,
    14,16,23,35,  14,16,23,36,  14,16,23,45,  14,16,25,35,  14,16,25,36,  14,16,25,45,  14,16,35,36,  14,16,35,45,  14,16,36,45,
    14,23,25,35,  14,23,25,36,  14,23,25,45,  14,23,35,36,  14,23,35,45,  14,23,36,45,  14,25,35,36,  14,25,35,45,  14,25,36,45,
    14,35,36,45,  15,16,23,25,  15,16,23,35,  15,16,23,36,  15,16,23,45,  15,16,25,35,  15,16,25,36,  15,16,25,45,  15,16,35,36,
    15,16,35,45,  15,16,36,45,  15,23,25,35,  15,23,25,36,  15,23,25,45,  15,23,35,36,  15,23,35,45,  15,23,36,45,  15,25,35,36,
    15,25,35,45,  15,25,36,45,  15,35,36,45,  16,23,25,35,  16,23,25,36,  16,23,25,45,  16,23,35,36,  16,23,35,45,  16,23,36,45,
    16,25,35,36,  16,25,35,45,  16,25,36,45,  16,35,36,45,  23,25,35,36,  23,25,35,45,  23,25,36,45,  23,35,36,45,  25,35,36,45 };
static char tseq[36][10] = // 36 16- or 18-move sequences to flip the tcomb combinations of 2 edge pieces (2 moves/byte)
  { 31,46,13,14,24,23,15,13,14,00,  13,11,46,13,14,24,23,15,00,00,  23,15,13,11,46,13,14,24,00,00,  11,42,42,31,51,31,14,61,32,00,
    16,26,31,55,26,46,25,15,00,00,  13,14,24,23,15,13,11,46,00,00,  23,31,54,24,13,13,24,64,00,00,  15,26,26,15,35,15,52,45,00,00,
    13,15,26,26,15,14,61,15,00,00,  11,31,51,31,14,61,31,42,41,00,  24,25,31,14,26,24,13,13,00,00,  25,15,16,24,26,11,53,26,00,00,
    24,23,15,13,11,46,13,14,00,00,  13,62,33,13,53,14,24,23,00,00,  16,26,13,13,54,24,16,11,55,00,  61,14,61,31,42,42,31,51,35,00,
    23,13,62,33,13,53,14,24,00,00,  11,51,51,62,42,61,15,32,61,00,  11,42,62,41,31,32,42,53,00,00,  13,24,64,23,31,54,24,13,00,00,
    15,16,26,31,55,26,46,25,00,00,  14,24,23,15,13,11,46,13,00,00,  15,14,61,15,13,15,26,26,00,00,  11,46,13,14,24,23,15,13,00,00,
    11,33,15,42,41,31,32,46,41,00,  26,31,55,26,46,25,15,16,00,00,  45,35,42,42,61,31,45,33,55,00,  42,46,35,35,46,41,53,36,00,00,
    32,33,15,42,41,31,32,46,33,00,  36,46,45,31,35,33,62,35,00,00,  45,53,16,46,35,35,46,26,00,00,  35,24,55,35,15,36,46,45,00,00,
    55,15,35,15,52,45,15,26,25,00,  51,32,46,42,33,15,42,41,36,00,  13,64,63,55,33,63,24,25,13,00,  46,41,53,36,42,46,35,35,00,00 };
static char fseq[126][10] = // 126 16- 18- or 20-move sequences to flip the fcomb combinations of 4 edge pieces (2 moves/byte)
  { 31,13,32,45,16,23,31,13,16,45,  31,15,32,35,15,52,45,42,64,00,  31,31,54,51,31,14,61,64,23,24,  16,42,33,15,45,13,11,46,00,00,
    31,64,23,31,54,51,31,14,64,00,  31,14,26,36,23,31,35,25,00,00,  31,51,36,31,55,26,41,46,14,00,  32,42,33,15,45,13,11,46,16,14,
    11,53,11,42,63,62,33,13,51,00,  23,23,51,55,24,54,26,11,51,00,  31,46,16,42,33,15,45,13,14,00,  35,11,53,23,51,55,24,54,25,54,
    13,36,23,26,45,53,16,13,52,00,  13,62,32,64,55,31,61,35,32,00,  15,54,25,24,55,35,14,15,33,52,  35,14,15,33,64,25,24,55,00,00,
    33,51,41,53,36,42,52,45,54,00,  53,51,41,53,36,42,52,45,00,00,  13,31,14,26,23,13,52,42,35,16,  11,51,31,62,64,15,16,42,32,55,
    63,16,42,33,15,45,13,11,46,45,  26,23,13,52,42,35,16,13,31,14,  51,14,61,64,23,31,54,51,36,00,  25,31,14,26,36,23,31,35,00,00,
    32,35,15,52,45,42,61,15,00,00,  23,11,42,63,62,33,13,52,51,00,  33,55,41,42,63,53,26,46,31,35,  45,13,11,46,16,42,33,15,00,00,
    33,62,33,13,52,53,11,42,64,00,  26,41,46,11,51,36,31,55,00,00,  31,54,51,31,14,61,64,23,00,00,  13,52,53,11,42,63,62,33,00,00,
    61,42,63,62,33,13,52,53,15,00,  36,23,31,35,25,31,14,26,00,00,  14,26,36,23,31,35,25,31,00,00,  13,11,46,16,42,33,15,45,00,00,
    52,45,13,11,46,16,42,33,15,16,  24,51,31,14,61,64,23,31,51,00,  31,15,13,63,15,52,64,14,64,00,  11,64,14,61,15,13,63,15,51,00,
    42,33,15,45,13,11,46,16,00,00,  23,31,54,51,31,14,61,64,00,00,  54,23,31,54,51,31,14,61,55,00,  11,51,36,31,55,26,41,46,00,00,
    11,46,16,42,33,15,45,13,00,00,  26,36,23,31,35,25,31,14,00,00,  23,31,35,25,31,14,26,36,00,00,  23,13,52,53,11,42,63,62,31,00,
    52,45,42,61,15,32,35,15,00,00,  23,31,13,16,45,31,13,32,45,16,  13,32,63,62,33,13,52,53,11,32,  13,26,36,23,31,35,25,31,13,32,
    32,33,13,52,53,11,42,63,64,00,  53,15,45,13,11,46,16,42,36,00,  55,36,23,31,35,25,31,14,25,00,  51,35,25,31,14,26,36,23,36,00,
    23,51,61,33,11,42,62,31,35,24,  51,46,16,42,33,15,45,13,16,00,  11,64,23,31,54,51,31,14,62,00,  61,51,36,31,55,26,41,46,15,00,
    13,11,42,63,62,33,13,52,52,00,  31,13,52,53,11,42,63,62,33,24,  41,46,11,51,36,31,55,26,00,00,  33,14,61,15,13,63,15,52,63,00,
    15,52,45,42,61,15,32,35,00,00,  13,15,45,13,11,46,16,42,32,00,  11,35,25,31,14,26,36,23,32,00,  61,14,26,36,23,31,35,25,35,00,
    11,53,23,51,55,24,54,26,00,00,  11,42,63,62,33,13,52,53,00,00,  13,63,15,52,64,14,61,15,00,00,  31,13,31,54,51,31,14,61,64,14,
    11,41,53,36,42,52,45,53,52,00,  13,31,46,11,51,36,31,55,26,32,  41,46,16,42,33,15,45,13,13,00,  24,23,31,54,51,31,14,61,61,00,
    11,33,15,45,13,11,46,16,41,00,  16,11,42,63,62,33,13,52,53,52,  15,13,63,15,52,64,14,61,00,00,  14,61,64,23,31,54,51,31,00,00,
    11,63,62,33,13,52,53,11,41,00,  45,53,16,13,53,36,23,26,00,00,  13,13,52,53,11,42,63,62,32,00,  55,41,46,11,51,36,31,55,25,00,
    13,35,13,11,46,16,42,33,15,32,  14,14,26,36,23,31,35,25,31,32,  13,13,11,46,16,42,33,15,45,42,  51,15,32,35,15,52,45,42,55,00,
    45,35,14,15,33,64,25,24,53,00,  35,35,14,15,33,64,25,24,54,00,  13,26,41,46,11,51,36,31,55,42,  41,35,14,15,33,64,25,24,55,23,
    33,62,32,64,55,31,61,35,00,00,  36,23,26,45,53,16,13,53,00,00,  55,42,52,45,53,51,41,53,35,00,  13,51,41,53,36,42,52,45,52,00,
    13,35,14,15,33,64,25,24,55,42,  15,35,14,15,33,64,25,24,52,00,  11,33,25,26,41,31,64,24,15,13,  42,31,14,26,36,23,31,35,25,13,
    32,31,14,26,36,23,31,35,25,14,  15,45,13,11,46,16,42,33,00,00,  31,35,25,31,14,26,36,23,00,00,  15,24,54,26,11,53,23,51,52,00,
    33,51,31,14,61,64,23,31,53,00,  41,42,63,62,33,13,52,53,13,00,  31,31,14,61,64,23,31,54,54,00,  13,23,31,35,25,31,14,26,36,42,
    15,31,54,51,31,14,61,64,23,62,  15,53,62,33,13,52,53,11,42,52,  32,64,14,61,15,13,63,15,54,00,  42,52,45,53,51,41,53,36,00,00,
    32,33,15,45,13,11,46,16,33,00,  25,24,54,26,11,53,23,51,51,00,  53,62,32,64,55,31,61,35,36,00,  33,26,36,23,31,35,25,31,13,00,
    32,63,62,33,13,52,53,11,33,00,  23,23,31,35,25,31,14,26,36,41,  25,31,54,51,31,14,61,64,23,61,  25,53,62,33,13,52,53,11,42,51,
    25,13,52,53,11,42,63,62,33,61,  42,33,13,52,53,11,42,63,63,00,  51,54,51,31,14,61,64,23,36,00,  35,53,62,33,13,52,53,11,42,54,
    53,13,52,53,11,42,63,62,36,00,  35,24,54,26,11,53,23,51,54,00 };
short i, j, q;  char p[4], s1, s2;

// flip edge pieces and display the moves

if (ad_flist[0][2] == 1 && ad_flist[1][2] == 0) // 1 pair to flip
{
p[0] = ad_flist[0][0]; p[1] = ad_flist[0][1];
q = 0; // identify the combination (tcomb and p are sorted)
for (i = 0; i < 2; i++) for (; q < 36; q++) if (tcomb[q][i] == p[i]) break;
if (q == 36) { fl_disp("** Error tcomb **"); cb_exit(); } // should never happen
ad_flip_mv(2, p, tseq[q]);
}

if (ad_flist[0][2] == 1 && ad_flist[1][2] == 1) // 2 pairs to flip
{
p[0] = ad_flist[0][0]; p[1] = ad_flist[0][1];
p[2] = ad_flist[1][0]; p[3] = ad_flist[1][1];
for (i = 0; i < 3; i++) // sort the 4 edge pieces from the 2 pairs
  for (j = i + 1; j < 4; j++)
    if ((s2 = p[j]) < (s1 = p[i])) { p[i] = s2; p[j] = s1; }
q = 0; // identify the combination (fcomb and p are sorted)
for (i = 0; i < 4; i++) for (; q < 126; q++) if (fcomb[q][i] == p[i]) break;
if (q == 126) { fl_disp("** Error fcomb **"); cb_exit(); } // should never happen
ad_flip_mv(4, p, fseq[q]);
}
}

/********************/
ad_flip_mv(nf, p, pseq) short nf; char *p, *pseq;
{
short i, j, m, mm;

cb_disp_lno("Flip"); cb_rec_add(25);
for (i = 0; i < nf; i++) { fl_sp(); fl_num(p[i]); } fl_disp(": ");
for (i = 0; i < 10 && (mm = pseq[i]) > 0; i++)
  for (j = 0; j < 2; j++) { m = j == 0 ? mm / 10 : mm % 10;
    fl_sp(); cb_disp_m(m); ad_move(m); movetot++; cb_rec_add(m); }
fl_nl(); cb_rec_add(0); cb_disp_end();
}

/********************/
ad_solve(st) short st;
{
static char sl[] = // number of moves for position modification 1 - 9
  { 2,8,12,12,12,12,12,12,12 };
short i;

// start solution process (st is the solution type)

stype = st; k_solve = 1; cb_pos_save(); cb_goal_save(); cb_disp_init();
ad_chk(1); // only user-specified cube can be invalid

if (stype == 0 || stype >= 3 && stype <= 5) // solution types s sa sc se
{
cb_begin(); if (stype == 3) lno = -1; ad_solution();
for (i = 0; i < 9; i++) { if (k_fin) break; // k_fin=1 if ended successfully
  cb_end(i); cb_begin(); cb_pos_restore(); cb_modpos(sl[i]); ad_solution(); }
cb_end(i);
}
if (stype == 1) { ad_chk_anchor(ad); // solution type sd
  cb_begin(); lno = -1; ad_trials(0, "Solution: "); cb_end(-1); }
if (stype == 2) { ad_chk_anchor(ad); ad_gchk(1); // solution type sg
  cb_begin(); lno = -1; ad_trials(3, "Goal solution: "); cb_end(-1); }

if (!k_fin) cb_pos_restore(); k_solve = 0; stype = 0;
}

/********************/
ad_solution()
{
short i;  char temp[sz_ad];

// standard solution (types s sa sc se)

ad_copy(ad_goal, temp);
if (fl_ncomp(ad, ad_orig, sz_ad) == 0) // already solved, show "---"
  { lno = -1; ad_trials(0, "Solution: "); goto s8; }
ad_copy(ad_anchor, ad_goal); // ad_goal used only for display
ad_trials(1, "Anchor: "); if (!k_fin || stype == 3) goto s8;
ad_copy(ad_acorners, ad_goal);
ad_trials(2, "Corners: "); if (!k_fin || stype == 4) goto s8;
fl_clr(ad_flist, num_fpairs*3);
for (i = 0; i < num_pairs; i++) { ad_trials(9, "Edges: "); if (!k_fin) goto s8; }
if (stype == 5) goto s8;
ad_flip();
if (fl_ncomp(ad, ad_orig, sz_ad) != 0)
  { fl_disp("Not solved, unidentified error"); cb_exit(); } // should never happen
s8: ad_copy(temp, ad_goal);
}

/********************/
ad_reset(kg, ke, kd) short kg, ke, kd;
{
short i, j, t;

// reset cube to standard orientation (center pieces correct)

if (kg) cb_inp_xchg();
for (i = 0; i < 24; i++)
{
for (j = 0; j < 6; j++)
  if (ad[9 * j + 4] != ad_faces[i][j]) goto s4; goto s8; s4: ; // center pieces
}
// invalid cube/goal, can happen only if ad_reset called before ad_chk/ad_gchk
if (kg) cb_inp_xchg(); cb_msg_noreset(kg); if (ke) cb_xrun(); goto s9;

// i from previous loop (the 24 entries in ad_faces & ad_rturns correspond)
s8: for (j = 0; j < 3 && (t = ad_rturns[i][j]) > 0; j++) ad_turn(t);
if (kg) cb_inp_xchg(); if (kd) cb_msg_reset(kg);
s9: ;
}

/********************/
ad_chk_colors(rr) char rr[6][9];
{
short i, j, r, c;  char tot[6];

// check colors and centers validity

r = 0; fl_clr(tot, 6);
for (i = 0; i < 6; i++)
  for (j = 0; j < 9; j++)
    if (!((c = rr[i][j]) >= 1 && c <= 6 || c == 9 && rr == ad_goal)
      || c != 9 && ++tot[c - 1] > 9) { r = 1; goto s9; }

for (i = 0; i < 24; i++)
{
for (j = 0; j < 6; j++)
  if (rr[j][4] != ad_faces[i][j]) goto s4; goto s9; s4: ; // center pieces
}
r = 7; // not found in ad_faces (table of valid orientations)

s9: return r;
}

/********************/
ad_chk_edges(rr) char rr[6][9];
{
short k, r, psum, qn, c;  char temp[sz_ad], kg;

// check edge pieces validity and parity (ignore 9's if goal)

ad_copy(ad, temp); if (rr != ad) ad_copy(rr, ad); r = psum = qn = kg = 0;
while (qn < 12)
{ // bring the 12 edge pieces to location 16
// turns sequence matches the flip order of pieces in ad_edge table
if ((c = 10 * ad[1] + ad[50]) == 99) { kg = 1; goto s6; }
for (k = 0; k < 24; k++) if (c == ad_edge[k]) goto s4; r = 2; goto s8;
s4: psum += k % 2;
s6: qn++; ad_turn(1);
if (qn == 4 || qn == 10) ad_turn(5); if (qn == 5 || qn == 11) ad_turn(1);
if (qn == 6 || qn == 12) { ad_turn(1); ad_turn(5); }
}
if (!kg && psum & 1) r = 3;

s8: ad_copy(temp, ad); return r;
}

/********************/
ad_chk_corners(rr) char rr[6][9];
{
short k, r, psum, qn, c;  char temp[sz_ad], kg;

// check corner pieces validity and parity (ignore 9's if goal)

ad_copy(ad, temp); if (rr != ad) ad_copy(rr, ad); r = psum = qn = kg = 0;
while (qn < 8)
{ // bring the 8 corner pieces to location 136
// turns sequence matches the flip order of pieces in ad_corner table (0 - 23)
if ((c = 100 * ad[0] + 10 * ad[20] + ad[53]) == 999) { kg = 1; goto s6; }
for (k = 0; k < 24; k++) if (c == ad_corner[k]) goto s4; r = 4; goto s8;
s4: psum += k % 3;
s6: qn++; ad_turn(1); if (qn == 4 || qn == 8) { ad_turn(3); ad_turn(3); }
}
if (!kg && psum % 3) r = 5;

s8: ad_copy(temp, ad); return r;
}

/********************/
ad_chk_swap(rr) char rr[6][9];
{
short i, j, r, c, c1, c2, psum, wcl[27], wcd[12];  char *dd, kn;

// check pieces duplication and swap parity (after the other checks)

r = kn = 0; fl_clr(wcl, 27*2); dd = rr;

for (i = 0; i < 12; i++) // set the 12 edge pieces in wcl, check for duplicates
{
c = cb_col_f2(dd[ad_ef[i][0]], dd[ad_ef[i][1]]);
wcl[((char*)ad_cn)[ad_ef[i][0]] - 1] = wcd[i] = c;
for (j = 0; j < i; j ++) if (c == wcd[j] && c != 99) { r = 11; goto s9; }
if (c == 99) kn = 1;
}

for (i = 0; i < 8; i++) // set the 8 corner pieces in wcl, check for duplicates
{
c = cb_col_f3(dd[ad_cf[i][0]], dd[ad_cf[i][1]], dd[ad_cf[i][2]]);
wcl[((char*)ad_cn)[ad_cf[i][0]] - 1] = wcd[i] = c;
for (j = 0; j < i; j ++) if (c == wcd[j] && c != 999) { r = 12; goto s9; }
if (c == 999) kn = 1;
}

if (kn) goto s9; // check goal swap parity only if no 9's
for (i = psum = 0; i < 27; i++) // compare wcl with ad_cl
{
if ((c1 = wcl[i]) == 0 || c1 == (c2 = ad_cl[i])) continue; // =0 for 6+1 center pieces
for (j = i + 1; j < 27; j++) // edges and corners have combined swap parity
  if (wcl[j] == c2) { wcl[j] = c1; psum++; break; }
}
if (psum & 1) r = 14;

s9: return r;
}

/********************/
ad_chk_anchor(rr) char rr[6][9];
{
short i;  char c;

for (i = 0; i < sz_ad; i++)
  if ((c = ad_anchor[i]) != 9 && c != ((char*)rr)[i]) goto s8;
goto s9;

s8: cb_inp_msg(rr == ad ? "Anchor wrong" : "Goal anchor wrong");
fl_nl(); cb_error();
s9: ;
}

/********************/
ad_chk(ke) short ke;
{
short i, r;

// check ad array

if (r = ad_chk_colors(ad)) goto s6; // check colors and centers

// check orientation (after centers check)
for (i = 0; i < 6; i++) if (ad[9 * i + 4] != i + 1) // center pieces
  { cb_inp_msg("Cube incorrectly oriented"); goto s8; }

// check edges, corners, duplication, swaps
if (r = ad_chk_edges(ad)) goto s6;
if (r = ad_chk_corners(ad)) goto s6;
if (r = ad_chk_swap(ad)) goto s6;

if (!ke) { fl_disp("Cube OK"); fl_nl(); }
goto s9;

s6: cb_inp_msg("Invalid cube"); cb_msg_chk(r);
s8: fl_nl(); if (ke) cb_error();
s9: ;
}

/********************/
ad_gchk(ke) short ke;
{
short i, r, c;  char *gg;

// check ad_goal array

if (r = ad_chk_colors(ad_goal)) goto s6; // check colors and centers

// check orientation (after centers check)
for (i = 0; i < 6; i++) if (ad_goal[9 * i + 4] != i + 1) // center pieces
  { cb_inp_msg("Goal incorrectly oriented"); goto s8; }
// check anchor (after colors and centers check)
if (!k_gen) ad_chk_anchor(ad_goal);

gg = ad_goal;
// check each piece's faces (all or none are 9)
for (i = 0; i < 12; i++) // edges
  if (gg[ad_ef[i][0]] == 9 ^ gg[ad_ef[i][1]] == 9) { r = 8; goto s6; }
for (i = 0; i < 8; i++) // corners
  if ((c = (gg[ad_cf[i][0]] == 9) + (gg[ad_cf[i][1]] == 9)
    + (gg[ad_cf[i][2]] == 9)) == 1 || c == 2) { r = 9; goto s6; }

// check edges, corners, duplication, swaps (after faces check)
if (r = ad_chk_edges(ad_goal)) goto s6;
if (r = ad_chk_corners(ad_goal)) goto s6;
if (r = ad_chk_swap(ad_goal)) goto s6;

if (!ke) cb_msg_goal(); goto s9;

s6: cb_inp_msg("Invalid goal"); cb_msg_chk(r);
s8: fl_nl(); if (ke) cb_error();
s9: ;
}

/********************/
ad_gen()
{
short i, j, k, n, r, psum, qn, en, cn, c, c1, c2, c3;  char tp[27];

// generate valid random position from goal position

k_gen = 1; cb_pos_save(); ad_gchk(1); cb_gen_pos(tp, &en, &cn);
// cb_gen_pos replaced the 99/999 pieces randomly with real ones (now ex-99/999)

// check and correct swap parity (edges and corners have combined parity)
if (ad_chk_swap(ad) == 14) // edge and/or corner pieces wrong
 if (en >= 2) // ex-99 pieces
{
for (i = n = 0; ; i++) // swap first two ex-99 pieces found
  if (tp[i] == 2) if (n == 0) { j = ad_ci[i]; n++; } else { k = ad_ci[i]; break; }
c1 = ad_ef[j][0]; c2 = ad_ef[k][0]; c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = c;
c1 = ad_ef[j][1]; c2 = ad_ef[k][1]; c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = c;
}
 else if (cn >= 2) // ex-999 pieces
{
for (i = n = 0; ; i++) // swap first two ex-999 pieces found
  if (tp[i] == 3) if (n == 0) { j = ad_ci[i]; n++; } else { k = ad_ci[i]; break; }
c1 = ad_cf[j][0]; c2 = ad_cf[k][0]; c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = c;
c1 = ad_cf[j][1]; c2 = ad_cf[k][1]; c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = c;
c1 = ad_cf[j][2]; c2 = ad_cf[k][2]; c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = c;
}
 else { r = 14; goto s6; }

// check and correct edge pieces validity and parity (see ad_chk_edges)
// if parity wrong and not seen in ad_gchk, at least one ex-99 piece exists
psum = qn = 0;
while (qn < 12)
{
c = 10 * ad[1] + ad[50];
for (k = 0; k < 24; k++) if (c == ad_edge[k]) goto s42;
r = 2; goto s6; // should never happen, validity confirmed in ad_gchk
s42: psum += k % 2;
qn++; ad_turn(1);
if (qn == 4 || qn == 10) ad_turn(5); if (qn == 5 || qn == 11) ad_turn(1);
if (qn == 6 || qn == 12) { ad_turn(1); ad_turn(5); }
}
if (psum & 1) // parity wrong
{
if (en == 0) { r = 3; goto s6; } // should never happen
for (i = 0; ; i++) if (tp[i] == 2) // flip first ex-99 piece found
  {
  j = ad_ci[i]; c1 = ad_ef[j][0]; c2 = ad_ef[j][1];
  c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = c; break;
  }
}

// check and correct corner pieces validity and parity (see ad_chk_corners)
// if parity wrong and not seen in ad_gchk, at least one ex-999 piece exists
s45: psum = qn = 0;
while (qn < 8)
{
while (1)
  {
  c = 100 * ad[0] + 10 * ad[20] + ad[53];
  for (k = 0; k < 24; k++) if (c == ad_corner[k]) goto s46;
  c = ad[0]; ad[0] = ad[20]; ad[20] = c; // invalid, change piece orientation
  }
s46: psum += k % 3;
qn++; ad_turn(1); if (qn == 4 || qn == 8) { ad_turn(3); ad_turn(3); }
}
if (psum % 3) // parity wrong
{
if (cn == 0) { r = 5; goto s6; } // should never happen
for (i = 0; ; i++) if (tp[i] == 3) // flip first ex-999 piece found
  {
  j = ad_ci[i]; c1 = ad_cf[j][0]; c2 = ad_cf[j][1]; c3 = ad_cf[j][2];
  c = ad[c1]; ad[c1] = ad[c2]; ad[c2] = ad[c3]; ad[c3] = c;
  goto s45; // repeat check, may need a second flip
  }
}

goto s8;

// error means goal is invalid and this could not be seen in ad_gchk
s6: cb_inp_msg("Invalid goal"); cb_msg_chk(r); fl_nl(); cb_error();
s8: k_gen = 0;
}

/********************/
ad_copy(s, d) char *s, *d;
{
// copy position (sz_ad bytes) from s to d

/* the asm below
fl_move(s, d, sz_ad);
*/
_asm
#if win32
{                         ; sz_ad 54 = 3x16+4+2
        mov   esi, s
        mov   edi, d
        mov   ecx, 3      ; 3x4+1 mov pairs is faster than 13 rep movsd
ad_cp3: mov   eax, [esi]
        mov   [edi], eax
        mov   eax, [esi+4]
        mov   [edi+4], eax
        mov   eax, [esi+8]
        mov   [edi+8], eax
        mov   eax, [esi+12]
        mov   [edi+12], eax
        add   esi, 16
        add   edi, 16
        loop  ad_cp3
        mov   eax, [esi]   ; 4 more bytes
        mov   [edi], eax
        mov   ax, [esi+4]  ; 2 more bytes
        mov   [edi+4], ax
}
#else
{                         ; sz_ad 54 = 3x16+4+2
        mov   si, s
        mov   di, d
        mov   ecx, 3      ; 3x4+1 mov pairs is faster than 13 rep movsd
ad_cp3: mov   eax, [si]
        mov   [di], eax
        mov   eax, [si+4]
        mov   [di+4], eax
        mov   eax, [si+8]
        mov   [di+8], eax
        mov   eax, [si+12]
        mov   [di+12], eax
        add   si, 16
        add   di, 16
        loop  ad_cp3
        mov   eax, [si]   ; 4 more bytes
        mov   [di], eax
        mov   ax, [si+4]  ; 2 more bytes
        mov   [di+4], ax
}
#endif
}

/********************/
ad_turn(t) short t;
{
static char pft[6][6][9] = // 6 lines x 6x9 "from" offsets in ad array
  { 6, 3, 0, 7, 4, 1, 8, 5, 2,  11,14,17,10,13,16, 9,12,15,  36,37,38,39,40,41,42,43,44,  53,52,51,50,49,48,47,46,45,  35,34,33,32,31,30,29,28,27,  18,19,20,21,22,23,24,25,26,
    2, 5, 8, 1, 4, 7, 0, 3, 6,  15,12, 9,16,13,10,17,14,11,  45,46,47,48,49,50,51,52,53,  44,43,42,41,40,39,38,37,36,  18,19,20,21,22,23,24,25,26,  35,34,33,32,31,30,29,28,27,
   51,48,45,52,49,46,53,50,47,  42,39,36,43,40,37,44,41,38,  24,21,18,25,22,19,26,23,20,  29,32,35,28,31,34,27,30,33,   6, 3, 0, 7, 4, 1, 8, 5, 2,  15,12, 9,16,13,10,17,14,11,
   38,41,44,37,40,43,36,39,42,  47,50,53,46,49,52,45,48,51,  20,23,26,19,22,25,18,21,24,  33,30,27,34,31,28,35,32,29,  11,14,17,10,13,16, 9,12,15,   2, 5, 8, 1, 4, 7, 0, 3, 6,
   18,19,20,21,22,23,24,25,26,  27,28,29,30,31,32,33,34,35,   9,10,11,12,13,14,15,16,17,   0, 1, 2, 3, 4, 5, 6, 7, 8,  42,39,36,43,40,37,44,41,38,  47,50,53,46,49,52,45,48,51,
   27,28,29,30,31,32,33,34,35,  18,19,20,21,22,23,24,25,26,   0, 1, 2, 3, 4, 5, 6, 7, 8,   9,10,11,12,13,14,15,16,17,  38,41,44,37,40,43,36,39,42,  51,48,45,52,49,46,53,50,47 };
char *pt, temp[sz_ad];

// perform turn t (internal value 1-6)

if (t < 1 || t > 6) cb_msg_mt(t, 1); // should never happen
ad_copy(ad, temp); pt = pft[t - 1];

/* the asm below
short i;
for (i = 0; i < sz_ad; i++) ad[i] = temp[pt[i]];
*/
_asm
#if win32
{
        mov   esi, pt
        lea   edi, temp
        mov   edx, ad
        mov   ecx, sz_ad/2      ; write words
ad_tn3: movzx ebx, byte ptr [esi]
        mov   al, [edi+ebx]
        movzx ebx, byte ptr [esi+1]
        mov   ah, [edi+ebx]
        mov   [edx], ax
        add   esi, 2
        add   edx, 2
        loop  ad_tn3
}
#else
{
        mov   si, pt
        lea   di, temp
        mov   dx, ad
        mov   ecx, sz_ad/2      ; write words
ad_tn3: movzx bx, byte ptr [si]
        mov   al, [di+bx]
        movzx bx, byte ptr [si+1]
        mov   ah, [di+bx]
        mov   bx, dx
        mov   [bx], ax
        add   si, 2
        add   dx, 2
        loop  ad_tn3
}
#endif
}

/********************/
ad_move(m) short m;
{
static short pof[12] = // pre-calculated pft row offsets
  { 0,20,40,60,80,100,120,140,160,180,200,220 };
static char pft[12][20] = // 12 lines x 5 sets of 4 circular offsets in ad array
  { 1, 5, 7, 3,50,30,41,23, 0, 2, 8, 6,53,27,44,26,20,47,33,38,
    3, 7, 5, 1,23,41,30,50, 6, 8, 2, 0,26,44,27,53,38,33,47,20,
   19,23,25,21,52, 3,37,14,18,20,26,24,51, 0,38,17,11,53, 6,36,
   21,25,23,19,14,37, 3,52,24,26,20,18,17,38, 0,51,36, 6,53,11,
   37,41,43,39,25, 7,34,16,36,38,44,42,24, 6,33,15,17,26, 8,35,
   39,43,41,37,16,34, 7,25,42,44,38,36,15,33, 6,24,35, 8,26,17,
   10,14,16,12,48,21,39,32, 9,11,17,15,45,18,36,35,29,51,24,42,
   12,16,14,10,32,39,21,48,15,17,11, 9,35,36,18,45,42,24,51,29,
   28,32,34,30,46,12,43, 5,27,29,35,33,47, 9,42, 8, 2,45,15,44,
   30,34,32,28, 5,43,12,46,33,35,29,27, 8,42, 9,47,44,15,45, 2,
   46,50,52,48,28, 1,19,10,45,47,53,51,29, 2,20,11, 9,27, 0,18,
   48,52,50,46,10,19, 1,28,51,53,47,45,11,20, 2,29,18, 0,27, 9 };
short mv;

// perform move m (internal value 1-12)

mv = m; if (m > 6) m -= 6;
if (m < 1 || m > 6) cb_msg_mt(m, 0); // should never happen

/* the asm below
short i, j;  char *pt, *pq, c1, c2;
pt = pft[mv - 1];
for (i = 0; i < 5; i++)
{
c1 = ad[*(pq = pt)];
for (j = 0; j < 3; j++) { pt++; c2 = ad[*pt]; ad[*pt] = c1; c1 = c2; }
ad[*pq] = c1; pt++;
}
*/
_asm
#if win32
{
        movzx esi, word ptr mv
        shl   esi, 1
        movzx esi, word ptr pof[esi-2]
        lea   esi, pft[esi]  ; pft row # mv-1
        mov   edi, ad
        xor   ebx, ebx
        mov   ecx, 5         ; 5 sets of 4 elements
ad_mv3: mov   bl, [esi]      ; rotate the contents of each set
        mov   edx, ebx       ; save ebx for last mov
        mov   al, [edi+ebx]
        mov   bl, [esi+1]
        mov   ah, [edi+ebx]  ; mov+mov is faster than xchg
        mov   [edi+ebx], al
        mov   bl, [esi+2]
        mov   al, [edi+ebx]
        mov   [edi+ebx], ah
        mov   bl, [esi+3]
        mov   ah, [edi+ebx]
        mov   [edi+ebx], al
        mov   [edi+edx], ah
        add   esi, 4
        loop  ad_mv3
}
#else
{
        mov   si, word ptr mv
        shl   si, 1
        mov   si, word ptr pof[si-2]
        lea   si, pft[si]  ; pft row # mv-1
        mov   di, ad
        xor   bx, bx
        mov   ecx, 5       ; 5 sets of 4 elements
ad_mv3: mov   bl, [si]     ; rotate the contents of each set
        mov   dl, bl       ; save bl for last mov
        mov   al, [di+bx]
        mov   bl, [si+1]
        mov   ah, [di+bx]  ; mov+mov is faster than xchg
        mov   [di+bx], al
        mov   bl, [si+2]
        mov   al, [di+bx]
        mov   [di+bx], ah
        mov   bl, [si+3]
        mov   ah, [di+bx]
        mov   [di+bx], al
        mov   bl, dl
        mov   [di+bx], ah
        add   si, 4
        loop  ad_mv3
}
#endif
}

/********************/
ad_acomp()
{
// check anchor pieces, return 0 if correct, 1 if not
// faster than using ad_gcomp with goal ad_anchor

/* the asm below
if (*(uns*)(ad + 9) != 0x0202 || *(uns*)(ad + 12) != 0x0202
  || *(uns*)(ad + 28) != 0x0404 || *(uns*)(ad + 31) != 0x0404
  || *(uns*)(ad + 45) != 0x0606 || *(uns*)(ad + 48) != 0x0606) return 1;
return 0;
*/
_asm
#if win32
{
        mov   esi, ad
        cmp   word ptr [esi+9], 0202h
        jne   ad_ac8
        cmp   word ptr [esi+12], 0202h
        jne   ad_ac8
        cmp   word ptr [esi+28], 0404h
        jne   ad_ac8
        cmp   word ptr [esi+31], 0404h
        jne   ad_ac8
        cmp   word ptr [esi+45], 0606h
        jne   ad_ac8
        cmp   word ptr [esi+48], 0606h
        jne   ad_ac8
        xor   eax, eax       ; return 0
        jmp   ad_ac82
ad_ac8: mov   eax, 1         ; return 1
ad_ac82:
}
#else
{
        mov   si, ad
        cmp   word ptr [si+9], 0202h
        jne   ad_ac8
        cmp   word ptr [si+12], 0202h
        jne   ad_ac8
        cmp   word ptr [si+28], 0404h
        jne   ad_ac8
        cmp   word ptr [si+31], 0404h
        jne   ad_ac8
        cmp   word ptr [si+45], 0606h
        jne   ad_ac8
        cmp   word ptr [si+48], 0606h
        jne   ad_ac8
        xor   ax, ax         ; return 0
        jmp   ad_ac82
ad_ac8: mov   ax, 1          ; return 1
ad_ac82:
}
#endif
}

/********************/
ad_gcomp()
{
// compare ad with ad_goal, return 0 if matching, 1 if not

/* the asm below
short i;  char *dd, *gg;
for (i = 0, dd = ad, gg = ad_goal; i < sz_ad; i++, dd++, gg++)
  if (*gg != 9 && *gg != *dd) return 1;
return 0;
*/
_asm
#if win32
{
        mov   esi, ad
        mov   edi, ad_goal
        mov   ecx, sz_ad/2   ; read words
ad_gc3: mov   ax, [edi]
        cmp   al, 9
        je    ad_gc4
        cmp   al, [esi]
        jne   ad_gc8
ad_gc4: cmp   ah, 9
        je    ad_gc6
        cmp   ah, [esi+1]
        jne   ad_gc8
ad_gc6: add   esi, 2
        add   edi, 2
        loop  ad_gc3
        xor   eax, eax       ; return 0
        jmp   ad_gc82
ad_gc8: mov   eax, 1         ; return 1
ad_gc82:
}
#else
{
        mov   si, ad
        mov   di, ad_goal
        mov   ecx, sz_ad/2   ; read words
ad_gc3: mov   ax, [di]
        cmp   al, 9
        je    ad_gc4
        cmp   al, [si]
        jne   ad_gc8
ad_gc4: cmp   ah, 9
        je    ad_gc6
        cmp   ah, [si+1]
        jne   ad_gc8
ad_gc6: add   si, 2
        add   di, 2
        loop  ad_gc3
        xor   ax, ax         ; return 0
        jmp   ad_gc82
ad_gc8: mov   ax, 1          ; return 1
ad_gc82:
}
#endif
}

/********************/
ad_dcomp()
{
// compare ad with ad_orig, return 0 if same, 1 if not

/* the asm below
short n;  char *s, *d;
for (n = 0, s = ad, d = ad_orig; n < sz_ad; n++, s++, d++)
  if (*s != *d) return 1; return 0;
*/
_asm
#if win32
{
        mov   esi, ad
        mov   edi, offset ad_orig
        mov   ecx, 13      ; compare sz_ad bytes (54 = 13x4 + 2)
ad_dc3: mov   eax, [esi]   ; mov+cmp is faster than repe cmpsd
        cmp   eax, [edi]
        jne   ad_dc6
        add   esi, 4
        add   edi, 4
        loop  ad_dc3
        mov   ax, [esi]    ; 2 more bytes
        cmp   ax, [edi]
        jne   ad_dc6
        xor   eax, eax     ; return 0
        jmp   ad_dc62
ad_dc6: mov   eax, 1       ; return 1
ad_dc62:
}
#else
{
        mov   si, ad
        mov   di, offset ad_orig
        mov   ecx, 13     ; compare sz_ad bytes (54 = 13x4 + 2)
ad_dc3: mov   eax, [si]   ; mov+cmp is faster than repe cmpsd
        cmp   eax, [di]
        jne   ad_dc6
        add   si, 4
        add   di, 4
        loop  ad_dc3
        mov   ax, [si]    ; 2 more bytes
        cmp   ax, [di]
        jne   ad_dc6
        xor   ax, ax      ; return 0
        jmp   ad_dc62
ad_dc6: mov   ax, 1       ; return 1
ad_dc62:
}
#endif
}

/********************/
ad_ecomp()
{
static char edge[9][5] =
// 9 edge pieces 13 14 15 16 23 25 35 36 45: color 1 2, offset 1 2 in ad array, status
  { 1,3, 3,23,0,  1,4, 5,30,0,  1,5, 7,41,0,  1,6, 1,50,0,  2,3,14,21,0,
    2,5,16,39,0,  3,5,25,37,0,  3,6,19,52,0,  4,5,34,43,0 };
short i, j, k, tnum, cnum, fnum, ff;  char *ep, c1, c2, p1, p2, kct;

// compare ad with ad_goal
// if matching and edge pair(s) found, update goal and return 0, else return 1

if (ad_gcomp() != 0) goto s91;

for (i = tnum = cnum = fnum = 0; i < 9; i++)
{
ep = edge[i]; c1 = ep[0]; c2 = ep[1]; p1 = ep[2]; p2 = ep[3];
if (ad_goal[p1] == 9 ^ ad_goal[p2] == 9) { fl_disp("** Error face **"); cb_exit(); }
// should never happen if initial goal was correct (both edge faces were 9 or not)
if (ad_goal[p1] != 9) { ep[4] = 0; continue; } tnum++; // look only at 99 pieces
if (ad[p1] == c1 && ad[p2] == c2) { cnum++; ep[4] = 1; } // auto-corrected
  else if (ad[p1] == c2 && ad[p2] == c1) { fnum++; ep[4] = 2; } // flipped
    else ep[4] = 3; // still wrong
}

if (!(tnum & 1)) { fl_disp("** Error odd **"); cb_exit(); }
// should never happen if initial goal was correct (odd number of 99 pieces)

// if only auto-corrected left (cnum = tnum):
// if initial compare end here, else continue just to identify last pair(s) for display
if ((kct = cnum == tnum) && k_icmp) goto s9;

// search first for correct pairs (after 2 pairs found, allow also flipped pairs)
ff = 0; fl_clr(ad_epair, num_pairs*3);
s5: if (cnum < 2) if (tnum < 6) goto s6; else goto s91; // no pair of correct pieces exists

for (i = 0; i < 9; i++) if (edge[i][4] == 1)
  for (j = i + 1; j < 9; j++) if (edge[j][4] == 1) goto s54; // pair identified
fl_disp("** Error edge **"); cb_exit();  // should never happen
s54: ad_epair[ff][0] = 10 * edge[i][0] + edge[i][1]; // store the pair for display
ad_epair[ff][1] = 10 * edge[j][0] + edge[j][1];

// clear the two 9's in goal array (i j from previous loop)
for (k = 0; k < 2; k++)
{
ep = edge[k == 0 ? i : j];
c1 = ep[0]; c2 = ep[1]; p1 = ep[2]; p2 = ep[3]; ep[4] = 0;
ad_goal[p1] = c1; ad_goal[p2] = c2; // set correct colors
}

ff++; tnum -= 2; if ((cnum -= 2) >= 2) goto s5; // repeat, there is another pair
if (kct) goto s9; // all left were auto-corrected
if (tnum >= 6 || fnum < 2) goto s9; // do not search also for a flipped pair

s6: // search for flipped pairs
s7: if (fnum < 2) goto s91; // no pair of flipped pieces exists
for (i = 0; i < 9; i++) if (edge[i][4] == 2)
  for (j = i + 1; j < 9; j++) if (edge[j][4] == 2) goto s74; // pair identified
fl_disp("** Error flip **"); cb_exit();  // should never happen
s74: ad_epair[ff][0] = 10 * edge[i][0] + edge[i][1]; // store the pair for display
ad_epair[ff][1] = 10 * edge[j][0] + edge[j][1]; ad_epair[ff][2] = 1;

// clear the two 9's in goal array so as the flipped pair looks correct (i j from previous loop)
for (k = 0; k < 2; k++)
{
ep = edge[k == 0 ? i : j];
c1 = ep[0]; c2 = ep[1]; p1 = ep[2]; p2 = ep[3]; ep[4] = 0;
ad_goal[p1] = c2; ad_goal[p2] = c1; // set flipped colors
}

for (i = 0; i < num_fpairs; i++) if (ad_flist[i][2] == 0) goto s78;
fl_disp("** Error list **"); cb_exit(); // should never happen
s78: ad_flist[i][2] = 1; // add to list of flipped pairs
ad_flist[i][0] = ad_epair[ff][0]; ad_flist[i][1] = ad_epair[ff][1];

if ((fnum -= 2) >= 2) { ff++; goto s7; } // repeat, there is another pair

s9: return 0; s91: return 1;
}

/********************/
ad_trials(s, desc) short s; char *desc;
{
static char ast[] = // anchor stop for mstop 1 - 20
  { 0,1,2,3,4,5,6,7,7,7,7,7,8,8,8,9,9,9,9,9,9 };
int ad_dcomp(), ad_acomp(), ad_gcomp(), ad_ecomp();

// search for solution by trying move sequences of increasing length

step = s; // step type can be 0, 1, 2, 3, 9
ycomp = step == 0 ? ad_dcomp : step == 1 ? ad_acomp
  : step == 2 || step == 3 ? ad_gcomp : ad_ecomp;
cb_disp_start(); cb_disp_goal(step == 0 ? ad_orig : ad_goal);

k_fin = 0; k_icmp = 1; cb_incseq();
if (ycomp() == 0) // initial compare to see if step already solved
{
if (step != 9) { cb_disp_lno(desc); fl_disp("---"); fl_nl(); }
if (step == 1 || step == 2) { cb_rec_add(30 + cstep[step]); cb_rec_add(0); }
k_fin = 1; goto s9;
}
k_icmp = 0;

cb_disp_lno(desc); ad_copy(ad, ad_save);
for (im = 1; im <= (step == 1 ? ast[mstop] : mstop); im++)
{
if (im > 1 && (im - 1) % 5 == 0) fl_sp(); fl_ch('.');
if (step == 9 && im & 1) continue; // skip odd im
elvl = (edif = im - 14) <= 0 ? 3 : 3 + edif;
ad_loop(0); // start recursion
if (k_fin) { fl_sp(); cb_rec_add(20 + cstep[step]); cb_disp_moves();
  cb_disp_end(); goto s9; }
}
fl_nl();

s9: ;
}

/********************/
ad_loop(lvl) short lvl;
{
static char mvo[6][12] = // 6 options for the order of moves tried at each level
  { 1,2,3,4,5,6,7,8,9,10,11,12,  1,2,5,6,3,4,7,8,11,12,9,10,  3,4,1,2,5,6,9,10,7,8,11,12,
    3,4,5,6,1,2,9,10,11,12,7,8,  5,6,1,2,3,4,11,12,7,8,9,10,  5,6,3,4,1,2,11,12,9,10,7,8 };
short i, mr;  char *scur, *snext, *mp, m, m1, m2, km;

// try all moves at the current nesting level (called recursively)

mr = step == 1 ? 12 : 6; snext = (scur = ad_save + lvl * 64) + 64;
m1 = move[lvl - 1]; m2 = move[lvl - 2]; km = !(m1 & 1) || lvl > 1 && m1 == m2;

for (i = 1, mp = mvo[moveord]; i <= mr; i++, mp++)
{
m = move[lvl] = *mp;
// skip move combinations that do nothing or have same effect as others
if (lvl > 0) if (m == crev[m1] || m == m1 && km) continue;
if (!(lvl > 0 && i == 1)) ad_copy(scur, ad); ad_move(m);
if (lvl + 1 == im) { cb_incseq(); if (ycomp() == 0) { k_fin = 1; break; } }
  else { ad_copy(ad, snext); ad_loop(lvl + 1); }
if (k_fin) break;
}

if (lvl == elvl) cb_esc();
}

/********************/
ad_bk_trials()
{
short i, n;  char mv1, mv2;

// try back sequences from specified position to attain the required wrong pieces

ad_copy(ad, ad_save);
for (i = 0; i < 10 && (im = bklth[i]) > 0; i++)
{
fl_nl(); fl_disp("Sequence length: "); fl_num(im); fl_nl();
n = bkstnum == 0 ? 5 : 6; elvl = (edif = im - 15) <= 0 ? n : n + edif;
cb_time(0); bkcount = bkseqs = 0;
bktot = 2376; // calculate bktot = highest expected bkcount
if (bkstnum > 0)
  {
  n = bkstnum - 1; mv1 = bkstmove[n]; mv2 = n == 0 ? 0 : bkstmove[n - 1];
  bktot = !(mv1 & 1) || mv1 == mv2 ? 1584 : 1940; while (n--) bktot /= 4.44796;
  }
if ((n = edif) > 0) while (n--) bktot *= 4.44796;
ad_bk_loop(0); // start recursion
}
}

/********************/
ad_bk_loop(lvl) short lvl;
{
short i, n;  char *scur, *snext, m, m1, m2, km;

// try all moves at the current nesting level in a back sequence (called recursively)

snext = (scur = ad_save + lvl * 64) + 64;
m1 = move[lvl - 1]; m2 = move[lvl - 2]; km = !(m1 & 1) || lvl > 1 && m1 == m2;

for (move[lvl] = 1; (m = move[lvl]) <= 6; move[lvl]++)
{
if (bkstnum > 0 && lvl == bkstnum) // show only sequences starting with bkstmove
  for (i = 0; i < lvl; i++) if (move[i] != bkstmove[i]) goto s6;
// skip move combinations that do nothing or have same effect as others
if (lvl > 0) if (m == crev[m1] || m == m1 && km) continue;
if (!(lvl > 0 && m == 1)) ad_copy(scur, ad); ad_move(m);
if (lvl + 1 == im) cb_bk_comp();
  else { ad_copy(ad, snext); ad_bk_loop(lvl + 1); }
s6: ;
}

if (lvl == elvl) { bkcount++; cb_bk_erase(); cb_esc(); cb_bk_chg(); }
if (lvl == 0) cb_bk_erase();
}

/********************/

/*** fc functions ***/

/********************/
fc_custom()
{
switch (custom)
{
case 1: // solve random position
  cb_scramble(0); fc_solve(0); break;

case 2: // solve several random positions
  cu_positions(); break;

case 3: // solve benchmark position
  fc_copy(fc_init0, fc); fc_solve(0); break;

case 4: // list back sequences
  cu_backseq(); fc_bk_trials(); break;

case 5: // repeat script
  cu_script(); break;
}
}

/********************/
fc_solve(st) short st;
{
static char sl[] = // number of moves for position modification 1 - 9
  { 2,4,4,4,4,4,4,4,4 };
short i;

// start solution process (st is the solution type)

stype = st; k_solve = 1; cb_pos_save(); cb_disp_init();
fc_chk(1); // only user-specified cube can be invalid

if (stype == 0) // solution type s
{
cb_begin(); lno = -1; fc_solution();
for (i = 0; i < 9; i++) { if (k_fin) break; // k_fin=1 if ended successfully
  cb_end(i); cb_begin(); cb_pos_restore(); cb_modpos(sl[i]); fc_solution(); }
cb_end(i);
}
if (stype == 1) // solution type sd
  { cb_begin(); lno = -1; fc_trials(0, "Solution: "); cb_end(-1); }
if (stype == 2) // solution type sg
  { fc_gchk(1); cb_begin(); lno = -1; fc_trials(3, "Goal solution: "); cb_end(-1); }

if (!k_fin) cb_pos_restore(); k_solve = 0; stype = 0;
}

/********************/
fc_solution()
{
fc_trials(0, "Solution: ");
}

/********************/
fc_reset(kg, ke, kd) short kg, ke, kd;
{
short i, j, k;

// reset cube to standard orientation (piece 246 correct)

if (kg) cb_inp_xchg();
for (i = 0; i < 4; i++) // try all 24 possible orientations
{
for (j = 0; j < (i == 0 ? 4 : i == 2 ? 0 : 1); j++)
  {
  for (k = 0; k < 4; k++) // check piece in location 246
    { if (fc[4] == 2 && fc[13] == 4 && fc[20] == 6) goto s8; fc_turn(5); }
  if (i == 0) fc_turn(1);
  }
fc_turn(3);
}
// ends in initial orientation
// invalid cube/goal, can happen only if fc_reset called before fc_chk/fc_gchk
if (kg) cb_inp_xchg(); cb_msg_noreset(kg); if (ke) cb_xrun(); goto s9;

s8: if (kg) cb_inp_xchg(); if (kd) cb_msg_reset(kg);
s9: ;
}

/********************/
fc_chk_colors(rr) char rr[6][4];
{
short i, j, r, c;  char tot[6];

// check colors validity

r = 0; fl_clr(tot, 6);
for (i = 0; i < 6; i++)
  for (j = 0; j < 4; j++)
    if (!((c = rr[i][j]) >= 1 && c <= 6 || c == 9 && rr == fc_goal)
      || c != 9 && ++tot[c - 1] > 4) { r = 1; goto s9; }
s9: return r;
}

/********************/
fc_chk_corners(rr) char rr[6][4];
{
short k, r, psum, qn, c;  char temp[sz_fc], kg;

// check corner pieces validity and parity (ignore 9's if goal)

fc_copy(fc, temp); if (rr != fc) fc_copy(rr, fc); r = psum = qn = kg = 0;
while (qn < 8)
{ // bring the 8 corner pieces to location 136
// turns sequence matches the flip order of pieces in fc_corner table (0 - 23)
if ((c = 100 * fc[0] + 10 * fc[9] + fc[23]) == 999) { kg = 1; goto s6; }
for (k = 0; k < 24; k++) if (c == fc_corner[k]) goto s4; r = 4; goto s8;
s4: psum += k % 3;
s6: qn++; fc_turn(1); if (qn == 4 || qn == 8) { fc_turn(3); fc_turn(3); }
}
if (!kg && psum % 3) r = 5;

s8: fc_copy(temp, fc); return r;
}

/********************/
fc_chk_dup(rr) char rr[6][9];
{
short i, j, r, c, wcd[8];  char *dd;

// check pieces duplication (after the other checks), no fc swap parity

r = 0; dd = rr;

for (i = 0; i < 8; i++)
{
wcd[i] = c = cb_col_f3(dd[fc_cf[i][0]], dd[fc_cf[i][1]], dd[fc_cf[i][2]]);
for (j = 0; j < i; j ++) if (c == wcd[j] && c != 999) { r = 12; goto s9; }
}

s9: return r;
}

/********************/
fc_chk(ke) short ke;
{
short r;

// check fc array

// check colors, corners, duplication
if (r = fc_chk_colors(fc)) goto s6;
if (r = fc_chk_corners(fc)) goto s6;
if (r = fc_chk_dup(fc)) goto s6;

// check orientation (after the other checks)
if (!(fc[4] == 2 && fc[13] == 4 && fc[20] == 6)) // location 246
  { cb_inp_msg("Anchor wrong, cube incorrectly oriented"); goto s8; }

if (!ke) { fl_disp("Cube OK"); fl_nl(); }
goto s9;

s6: cb_inp_msg("Invalid cube"); cb_msg_chk(r);
s8: fl_nl(); if (ke) cb_error();
s9: ;
}

/********************/
fc_gchk(ke) short ke;
{
short i, j, k, r, c;  char *gg;

// check fc_goal array

if (r = fc_chk_colors(fc_goal)) goto s6; // check colors

gg = fc_goal; // check each piece's faces (all or none are 9)
for (i = 0; i < 8; i++)
  if ((c = (gg[fc_cf[i][0]] == 9) + (gg[fc_cf[i][1]] == 9)
    + (gg[fc_cf[i][2]] == 9)) == 1 || c == 2) { r = 9; goto s6; }

// check corners, duplication (after faces check)
if (r = fc_chk_corners(fc_goal)) goto s6;
if (r = fc_chk_dup(fc_goal)) goto s6;

if (k_gen) goto s4;
// check anchor, orientation (after the other checks)
for (i = 0; i < 8; i++) // check if piece 246 exists
{
for (j = k = 0; j < 3; j++)
  if ((c = gg[fc_cf[i][j]]) == 2 || c == 4 || c == 6) k++;
if (k == 3) goto s36;
}
cb_inp_msg("Goal anchor missing"); goto s8;

s36: if (!(gg[4] == 2 && gg[13] == 4 && gg[20] == 6)) // location 246
  { cb_inp_msg("Goal incorrectly oriented"); goto s8; }

s4: if (!ke) cb_msg_goal(); goto s9;

s6: cb_inp_msg("Invalid goal"); cb_msg_chk(r);
s8: fl_nl(); if (ke) cb_error();
s9: ;
}

/********************/
fc_gen()
{
short i, j, k, n, r, psum, qn, en, cn, c, c1, c2, c3;  char tp[8];

// generate valid random position from goal position

k_gen = 1; cb_pos_save(); fc_gchk(1); cb_gen_pos(tp, &en, &cn);
// cb_gen_pos replaced the 999 pieces randomly with real ones (now ex-999)

// check and correct corner pieces validity and parity (see fc_chk_corners)
// if parity wrong and not seen in fc_gchk, at least one ex-999 piece exists
s45: psum = qn = 0;
while (qn < 8)
{
while (1)
  {
  c = 100 * fc[0] + 10 * fc[9] + fc[23];
  for (k = 0; k < 24; k++) if (c == fc_corner[k]) goto s46;
  c = fc[0]; fc[0] = fc[9]; fc[9] = c; // invalid, change piece orientation
  }
s46: psum += k % 3;
qn++; fc_turn(1); if (qn == 4 || qn == 8) { fc_turn(3); fc_turn(3); }
}
if (psum % 3) // parity wrong
{
if (cn == 0) { r = 5; goto s6; } // should never happen
for (i = 0; ; i++) if (tp[i] == 3) // flip first ex-999 piece found
  {
  j = fc_ci[i]; c1 = fc_cf[j][0]; c2 = fc_cf[j][1]; c3 = fc_cf[j][2];
  c = fc[c1]; fc[c1] = fc[c2]; fc[c2] = fc[c3]; fc[c3] = c;
  goto s45; // repeat check, may need a second flip
  }
}

fc_reset(0, 0, 0); // no orientation check was done in fc_gchk
goto s8;

// error means goal is invalid and this could not be seen in fc_gchk
s6: cb_inp_msg("Invalid goal"); cb_msg_chk(r); fl_nl(); cb_error();
s8: k_gen = 0;
}

/********************/
fc_copy(s, d) char *s, *d;
{
// copy position (sz_fc bytes) from s to d

/* the asm below
fl_move(s, d, sz_fc);
*/
_asm
#if win32
{                         ; sz_fc 24 = 2x12
        mov   esi, s
        mov   edi, d
        mov   ecx, 2      ; 2x3 mov pairs is faster than 6 rep movsd
fc_cp3: mov   eax, [esi]
        mov   [edi], eax
        mov   eax, [esi+4]
        mov   [edi+4], eax
        mov   eax, [esi+8]
        mov   [edi+8], eax
        add   esi, 12
        add   edi, 12
        loop  fc_cp3
}
#else
{                         ; sz_fc 24 = 2x12
        mov   si, s
        mov   di, d
        mov   ecx, 2      ; 2x3 mov pairs is faster than 6 rep movsd
fc_cp3: mov   eax, [si]
        mov   [di], eax
        mov   eax, [si+4]
        mov   [di+4], eax
        mov   eax, [si+8]
        mov   [di+8], eax
        add   si, 12
        add   di, 12
        loop  fc_cp3
}
#endif
}

/********************/
fc_turn(t) short t;
{
static char pft[6][6][4] = // 6 lines x 6x4 "from" offsets in fc array
  { 2, 0, 3, 1,   5, 7, 4, 6,  16,17,18,19,  23,22,21,20,  15,14,13,12,   8, 9,10,11,
    1, 3, 0, 2,   6, 4, 7, 5,  20,21,22,23,  19,18,17,16,   8, 9,10,11,  15,14,13,12,
   22,20,23,21,  18,16,19,17,  10, 8,11, 9,  13,15,12,14,   2, 0, 3, 1,   6, 4, 7, 5,
   17,19,16,18,  21,23,20,22,   9,11, 8,10,  14,12,15,13,   5, 7, 4, 6,   1, 3, 0, 2,
    8, 9,10,11,  12,13,14,15,   4, 5, 6, 7,   0, 1, 2, 3,  18,16,19,17,  21,23,20,22,
   12,13,14,15,   8, 9,10,11,   0, 1, 2, 3,   4, 5, 6, 7,  17,19,16,18,  22,20,23,21 };
char *pt, temp[sz_fc];

// perform turn t (internal value 1-6)

if (t < 1 || t > 6) cb_msg_mt(t, 1); // should never happen
fc_copy(fc, temp); pt = pft[t - 1];

/* the asm below
short i;
for (i = 0; i < sz_fc; i++) fc[i] = temp[pt[i]];
*/
_asm
#if win32
{
        mov   esi, pt
        lea   edi, temp
        mov   edx, fc
        mov   ecx, sz_fc/2      ; write words
fc_tn3: movzx ebx, byte ptr [esi]
        mov   al, [edi+ebx]
        movzx ebx, byte ptr [esi+1]
        mov   ah, [edi+ebx]
        mov   [edx], ax
        add   esi, 2
        add   edx, 2
        loop  fc_tn3
}
#else
{
        mov   si, pt
        lea   di, temp
        mov   dx, fc
        mov   ecx, sz_fc/2      ; write words
fc_tn3: movzx bx, byte ptr [si]
        mov   al, [di+bx]
        movzx bx, byte ptr [si+1]
        mov   ah, [di+bx]
        mov   bx, dx
        mov   [bx], ax
        add   si, 2
        add   dx, 2
        loop  fc_tn3
}
#endif
}

/********************/
fc_move(m) short m;
{
static short pof[12] = // pre-calculated pft row offsets
  { 0,12,24,36,48,60,72,84,96,108,120,132 };
static char pft[12][12] = // 12 lines x 3 sets of 4 circular offsets in fc array
  { 0, 1, 3, 2,23,12,19,11, 9,21,14,17,
    2, 3, 1, 0,11,19,12,23,17,14,21, 9,
    8, 9,11,10,22, 0,17, 7, 5,23, 2,16,
   10,11, 9, 8, 7,17, 0,22,16, 2,23, 5,
   16,17,19,18,10, 2,14, 6, 7,11, 3,15,
   18,19,17,16, 6,14, 2,10,15, 3,11, 7,
    4, 5, 7, 6,20, 8,16,15,13,22,10,18,
    6, 7, 5, 4,15,16, 8,20,18,10,22,13,
   12,13,15,14,21, 4,18, 3, 1,20, 6,19,
   14,15,13,12, 3,18, 4,21,19, 6,20, 1,
   20,21,23,22,13, 1, 9, 5, 4,12, 0, 8,
   22,23,21,20, 5, 9, 1,13, 8, 0,12, 4 };
short mv;

// perform move m (internal value 1-12)

mv = m; if (m > 6) m -= 6;
if (m < 1 || m > 6) cb_msg_mt(m, 0); // should never happen

/* the asm below
short i, j;  char *pt, *pq, c1, c2;
pt = pft[mv - 1];
for (i = 0; i < 3; i++)
{
c1 = fc[*(pq = pt)];
for (j = 0; j < 3; j++) { pt++; c2 = fc[*pt]; fc[*pt] = c1; c1 = c2; }
fc[*pq] = c1; pt++;
}
*/
_asm
#if win32
{
        movzx esi, word ptr mv
        shl   esi, 1
        movzx esi, word ptr pof[esi-2]
        lea   esi, pft[esi]  ; pft row # mv-1
        mov   edi, fc
        xor   ebx, ebx
        mov   ecx, 3         ; 3 sets of 4 elements
fc_mv3: mov   bl, [esi]      ; rotate the contents of each set
        mov   edx, ebx       ; save ebx for last mov
        mov   al, [edi+ebx]
        mov   bl, [esi+1]
        mov   ah, [edi+ebx]  ; mov+mov is faster than xchg
        mov   [edi+ebx], al
        mov   bl, [esi+2]
        mov   al, [edi+ebx]
        mov   [edi+ebx], ah
        mov   bl, [esi+3]
        mov   ah, [edi+ebx]
        mov   [edi+ebx], al
        mov   [edi+edx], ah
        add   esi, 4
        loop  fc_mv3
}
#else
{
        mov   si, word ptr mv
        shl   si, 1
        mov   si, word ptr pof[si-2]
        lea   si, pft[si]  ; pft row # mv-1
        mov   di, fc
        xor   bx, bx
        mov   ecx, 3       ; 3 sets of 4 elements
fc_mv3: mov   bl, [si]     ; rotate the contents of each set
        mov   dl, bl       ; save bl for last mov
        mov   al, [di+bx]
        mov   bl, [si+1]
        mov   ah, [di+bx]  ; mov+mov is faster than xchg
        mov   [di+bx], al
        mov   bl, [si+2]
        mov   al, [di+bx]
        mov   [di+bx], ah
        mov   bl, [si+3]
        mov   ah, [di+bx]
        mov   [di+bx], al
        mov   bl, dl
        mov   [di+bx], ah
        add   si, 4
        loop  fc_mv3
}
#endif
}

/********************/
fc_gcomp()
{
// compare fc with fc_goal, return 0 if matching, 1 if not

/* the asm below
short i;  char *dd, *gg;
for (i = 0, dd = fc, gg = fc_goal; i < sz_fc; i++, dd++, gg++)
  if (*gg != 9 && *gg != *dd) return 1;
return 0;
*/
_asm
#if win32
{
        mov   esi, fc
        mov   edi, fc_goal
        mov   ecx, sz_fc/2   ; read words
fc_gc3: mov   ax, [edi]
        cmp   al, 9
        je    fc_gc4
        cmp   al, [esi]
        jne   fc_gc8
fc_gc4: cmp   ah, 9
        je    fc_gc6
        cmp   ah, [esi+1]
        jne   fc_gc8
fc_gc6: add   esi, 2
        add   edi, 2
        loop  fc_gc3
        xor   eax, eax       ; return 0
        jmp   fc_gc82
fc_gc8: mov   eax, 1         ; return 1
fc_gc82:
}
#else
{
        mov   si, fc
        mov   di, fc_goal
        mov   ecx, sz_fc/2   ; read words
fc_gc3: mov   ax, [di]
        cmp   al, 9
        je    fc_gc4
        cmp   al, [si]
        jne   fc_gc8
fc_gc4: cmp   ah, 9
        je    fc_gc6
        cmp   ah, [si+1]
        jne   fc_gc8
fc_gc6: add   si, 2
        add   di, 2
        loop  fc_gc3
        xor   ax, ax         ; return 0
        jmp   fc_gc82
fc_gc8: mov   ax, 1          ; return 1
fc_gc82:
}
#endif
}

/********************/
fc_dcomp()
{
// direct compare: compare fc with fc_orig, return 0 if same, 1 if not

/* the asm below
short n;  char *s, *d;
for (n = 0, s = fc, d = fc_orig; n < sz_fc; n++, s++, d++)
  if (*s != *d) return 1; return 0;
*/
_asm
#if win32
{
        mov   esi, fc
        mov   edi, offset fc_orig
        mov   ecx, 6       ; compare sz_fc bytes (24 = 6x4)
fc_dc3: mov   eax, [esi]   ; mov+cmp is faster than repe cmpsd
        cmp   eax, [edi]
        jne   fc_dc6
        add   esi, 4
        add   edi, 4
        loop  fc_dc3
        xor   eax, eax     ; return 0
        jmp   fc_dc62
fc_dc6: mov   eax, 1       ; return 1
fc_dc62:
}
#else
{
        mov   si, fc
        mov   di, offset fc_orig
        mov   ecx, 6      ; compare sz_fc bytes (24 = 6x4)
fc_dc3: mov   eax, [si]   ; mov+cmp is faster than repe cmpsd
        cmp   eax, [di]
        jne   fc_dc6
        add   si, 4
        add   di, 4
        loop  fc_dc3
        xor   ax, ax      ; return 0
        jmp   fc_dc62
fc_dc6: mov   ax, 1       ; return 1
fc_dc62:
}
#endif
}

/********************/
fc_trials(s, desc) short s; char *desc;
{
int fc_dcomp(), fc_gcomp();

// search for solution by trying move sequences of increasing length

step = s; // step type can be 0, 3
ycomp = step == 0 ? fc_dcomp : fc_gcomp;
cb_disp_start(); cb_disp_goal(step == 0 ? fc_orig : fc_goal); cb_disp_lno(desc);

k_fin = 0; cb_incseq(); // initial compare to see if already solved
if (ycomp() == 0) { fl_disp("---"); fl_nl(); k_fin = 1; goto s9; }

fc_copy(fc, fc_save);
for (im = 1; im <= mstop; im++)
{
if (im > 1 && (im - 1) % 5 == 0) fl_sp(); fl_ch('.');
elvl = (edif = im - 14) <= 0 ? 3 : 3 + edif;
fc_loop(0); // start recursion
if (k_fin) { fl_sp(); cb_rec_add(20 + cstep[step]); cb_disp_moves();
  cb_disp_end(); goto s9; }
}
fl_nl();

s9: ;
}

/********************/
fc_loop(lvl) short lvl;
{
static char mvo[6][6] = // 6 options for the order of moves tried at each level
  { 1,2,3,4,5,6,  1,2,5,6,3,4,  3,4,1,2,5,6,  3,4,5,6,1,2,  5,6,1,2,3,4,  5,6,3,4,1,2 };
short i;  char *scur, *snext, *mp, m, m1, m2, km;

// try all moves at the current nesting level (called recursively)

snext = (scur = fc_save + lvl * 64) + 64;
m1 = move[lvl - 1]; m2 = move[lvl - 2]; km = !(m1 & 1) || lvl > 1 && m1 == m2;

for (i = 1, mp = mvo[moveord]; i <= 6; i++, mp++)
{
m = move[lvl] = *mp;
// skip move combinations that do nothing or have same effect as others
if (lvl > 0) if (m == crev[m1] || m == m1 && km) continue;
if (!(lvl > 0 && i == 1)) fc_copy(scur, fc); fc_move(m);
if (lvl + 1 == im) { cb_incseq(); if (ycomp() == 0) { k_fin = 1; break; } }
  else { fc_copy(fc, snext); fc_loop(lvl + 1); }
if (k_fin) break;
}

if (lvl == elvl) cb_esc();
}

/********************/
fc_bk_trials()
{
short i, n;  char mv1, mv2;

// try back sequences from specified position to attain the required wrong pieces

fc_copy(fc, fc_save);
for (i = 0; i < 10 && (im = bklth[i]) > 0; i++)
{
fl_nl(); fl_disp("Sequence length: "); fl_num(im); fl_nl();
n = bkstnum == 0 ? 5 : 6; elvl = (edif = im - 15) <= 0 ? n : n + edif;
cb_time(0); bkcount = bkseqs = 0;
bktot = 2376; // calculate bktot = highest expected bkcount
if (bkstnum > 0)
  {
  n = bkstnum - 1; mv1 = bkstmove[n]; mv2 = n == 0 ? 0 : bkstmove[n - 1];
  bktot = !(mv1 & 1) || mv1 == mv2 ? 1584 : 1940; while (n--) bktot /= 4.44796;
  }
if ((n = edif) > 0) while (n--) bktot *= 4.44796;
fc_bk_loop(0); // start recursion
}
}

/********************/
fc_bk_loop(lvl) short lvl;
{
short i, n;  char *scur, *snext, m, m1, m2, km;

// try all moves at the current nesting level in a back sequence (called recursively)

snext = (scur = fc_save + lvl * 64) + 64;
m1 = move[lvl - 1]; m2 = move[lvl - 2]; km = !(m1 & 1) || lvl > 1 && m1 == m2;

for (move[lvl] = 1; (m = move[lvl]) <= 6; move[lvl]++)
{
if (bkstnum > 0 && lvl == bkstnum) // show only sequences starting with bkstmove
  for (i = 0; i < lvl; i++) if (move[i] != bkstmove[i]) goto s6;
// skip move combinations that do nothing or have same effect as others
if (lvl > 0) if (m == crev[m1] || m == m1 && km) continue;
if (!(lvl > 0 && m == 1)) fc_copy(scur, fc); fc_move(m);
if (lvl + 1 == im) cb_bk_comp();
  else { fc_copy(fc, snext); fc_bk_loop(lvl + 1); }
s6: ;
}

if (lvl == elvl) { bkcount++; cb_bk_erase(); cb_esc(); cb_bk_chg(); }
if (lvl == 0) cb_bk_erase();
}

/********************/

/*** py functions ***/

/********************/
py_custom()
{
switch (custom)
{
case 1: // solve random position
  cb_scramble(0); py_solve(0); break;

case 2: // solve several random positions
  cu_positions(); break;

case 3: // solve benchmark position
  py_copy(py_init0, py); py_solve(0); break;

case 4: // list back sequences
  cu_backseq(); py_bk_trials(); break;

case 5: // repeat script
  cu_script(); break;
}
}

/********************/
py_solve(st) short st;
{
static char sl[] = // number of moves for position modification 1 - 9
  { 2,4,4,4,4,4,4,4,4 };
short i;

// start solution process (st is the solution type)

stype = st; k_solve = 1; cb_pos_save(); cb_disp_init();
py_chk(1); // only user-specified pyramid can be invalid

if (stype == 0) // solution type s
{
cb_begin(); lno = -1; py_solution();
for (i = 0; i < 9; i++) { if (k_fin) break; // k_fin=1 if ended successfully
  cb_end(i); cb_begin(); cb_pos_restore(); cb_modpos(sl[i]); py_solution(); }
cb_end(i);
}
if (stype == 1) // solution type sd
  { cb_begin(); lno = -1; py_trials(0, "Solution: "); cb_end(-1); }
if (stype == 2) // solution type sg
  { py_gchk(1); cb_begin(); lno = -1; py_trials(3, "Goal solution: "); cb_end(-1); }

if (!k_fin) cb_pos_restore(); k_solve = 0; stype = 0;
}

/********************/
py_solution()
{
py_trials(0, "Solution: ");
}

/********************/
py_reset(kg, ke, kd) short kg, ke, kd;
{
short i, j, k;

// reset pyramid to standard orientation (axial 123 correct)

if (kg) cb_inp_xchg();
for (i = 0; i < 3; i++) // try all 12 possible orientations
{
for (j = 0; j < (i == 0 ? 3 : i == 1 ? 1 : 0); j++)
  {
  for (k = 0; k < 3; k++) // check piece in location 123
    { if (py[1] == 1 && py[7] == 2 && py[13] == 3) goto s8; py_turn(4); }
  if (i == 0) py_turn(2);
  }
py_turn(1);
}
// ends in initial orientation
// invalid pyramid/goal, can happen only if py_reset called before py_chk/py_gchk
if (kg) cb_inp_xchg(); cb_msg_noreset(kg); if (ke) cb_xrun(); goto s9;

s8: if (kg) cb_inp_xchg(); if (kd) cb_msg_reset(kg);
s9: ;
}

/********************/
py_chk_colors(rr) char rr[4][6];
{
short i, j, r, c;  char tot[4];

// check colors validity

r = 0; fl_clr(tot, 4);
for (i = 0; i < 4; i++)
  for (j = 0; j < 6; j++)
    if (!((c = rr[i][j]) >= 1 && c <= 4 || c == 9 && rr == py_goal)
      || c != 9 && ++tot[c - 1] > 6) { r = 1; goto s9; }
s9: return r;
}

/********************/
py_chk_edges(rr) char rr[4][6];
{
short k, r, psum, qn, c;  char temp[sz_py], kg;

// check edge pieces validity and parity (ignore 9's if goal)

py_copy(py, temp); if (rr != py) py_copy(rr, py); r = psum = qn = kg = 0;
while (qn < 6)
{ // bring the 6 edge pieces to location 12
// turns sequence matches the flip order of pieces in py_edge table
if ((c = 10 * py[2] + py[6]) == 99) { kg = 1; goto s6; }
for (k = 0; k < 12; k++) if (c == py_edge[k]) goto s4; r = 2; goto s8;
s4: psum += k % 2;
s6: qn++; if (qn <= 2) py_turn(1);
if (qn == 3) { py_turn(1); py_turn(8); } if (qn == 4) py_turn(2);
if (qn == 5) py_turn(4); if (qn == 6) { py_turn(1); py_turn(4); }
}
if (!kg && psum & 1) r = 3;

s8: py_copy(temp, py); return r;
}

/********************/
py_chk_axials(rr) char rr[4][6];
{
short k, r, qn, c;  char temp[sz_py];

// check axial pieces validity (ignore 9's if goal), no axial parity

py_copy(py, temp); if (rr != py) py_copy(rr, py); r = qn = 0;
while (qn < 4)
{ // bring the 4 axial pieces to location 123
// turns sequence matches the flip order of pieces in py_axial table (0 - 11)
if ((c = 100 * py[1] + 10 * py[7] + py[13]) == 999) goto s4;
for (k = 0; k < 12; k++) if (c == py_axial[k]) goto s4; r = 6; goto s8;
s4: qn++; py_turn(1); if (qn == 3) py_turn(3); if (qn == 4) py_turn(4);
}

s8: py_copy(temp, py); return r;
}

/********************/
py_chk_swap(rr) char rr[4][6];
{
short i, j, r, c, c1, c2, esum, asum, wcl[10], wcd[6];  char *dd, kne, kna;

// check pieces duplication and swap parity (after the other checks)

r = kne = kna = 0; fl_clr(wcl, 10*2); dd = rr;

for (i = 0; i < 6; i++) // set the 6 edge pieces in wcl, check for duplicates
{
c = cb_col_f2(dd[py_ef[i][0]], dd[py_ef[i][1]]);
wcl[((char*)py_cn)[py_ef[i][0]] - 1] = wcd[i] = c;
for (j = 0; j < i; j ++) if (c == wcd[j] && c != 99) { r = 11; goto s9; }
if (c == 99) kne = 1;
}

for (i = 0; i < 4; i++) // set the 4 axial pieces in wcl, check for duplicates
{
c = cb_col_f3(dd[py_cf[i][0]], dd[py_cf[i][1]], dd[py_cf[i][2]]);
wcl[((char*)py_cn)[py_cf[i][0]] - 1] = wcd[i] = c;
for (j = 0; j < i; j ++) if (c == wcd[j] && c != 999) { r = 13; goto s9; }
if (c == 999) kna = 1;
}

if (kne && kna) goto s9; // check goal swap parity only if no 9's
for (i = esum = asum = 0; i < 10; i++) // compare wcl with py_cl
{
if ((c1 = wcl[i]) == (c2 = py_cl[i])) continue;
for (j = i + 1; j < 10; j++) // edges and axials have separate swap parity
  if (wcl[j] == c2) { wcl[j] = c1; if (c1 <= 99) esum++; else asum++; break; }
}
if (esum & 1 && !kne) r = 15; else if (asum & 1 && !kna) r = 16;

s9: return r;
}

/********************/
py_chk(ke) short ke;
{
short r;

// check py array

if (r = py_chk_colors(py)) goto s6; // check colors

// no orientation check, py need not be correctly oriented

// check edges, corners, duplication, swaps
if (r = py_chk_edges(py)) goto s6;
if (r = py_chk_axials(py)) goto s6;
if (r = py_chk_swap(py)) goto s6;

if (!ke) { fl_disp("Pyramid OK"); fl_nl(); }
goto s9;

s6: cb_inp_msg("Invalid pyramid"); cb_msg_chk(r);
fl_nl(); if (ke) cb_error();
s9: ;
}

/********************/
py_gchk(ke) short ke;
{
short i, r, c;  char *gg;

// check py_goal array

if (r = py_chk_colors(py_goal)) goto s6; // check colors

// no orientation check, py goal need not be correctly oriented

gg = py_goal; // check each piece's faces (all or none are 9)
for (i = 0; i < 6; i++) // edges
  if (gg[py_ef[i][0]] == 9 ^ gg[py_ef[i][1]] == 9) { r = 8; goto s6; }
for (i = 0; i < 4; i++) // axials
  if ((c = (gg[py_cf[i][0]] == 9) + (gg[py_cf[i][1]] == 9)
    + (gg[py_cf[i][2]] == 9)) == 1 || c == 2) { r = 10; goto s6; }

// check edges, axials, duplication, swaps (after faces check)
if (r = py_chk_edges(py_goal)) goto s6;
if (r = py_chk_axials(py_goal)) goto s6;
if (r = py_chk_swap(py_goal)) goto s6;

if (!ke) cb_msg_goal(); goto s9;

s6: cb_inp_msg("Invalid goal"); cb_msg_chk(r);
fl_nl(); if (ke) cb_error();
s9: ;
}

/********************/
py_gen()
{
short i, j, k, n, r, psum, qn, en, cn, c, c1, c2, c3;  char tp[10];

// generate valid random position from goal position

k_gen = 1; cb_pos_save(); py_gchk(1); cb_gen_pos(tp, &en, &cn);
// cb_gen_pos replaced the 99/999 pieces randomly with real ones (now ex-99/999)

// check and correct swap parity (edges and axials have separate parity)
if (py_chk_swap(py) == 15) // edge pieces wrong
 if (en >= 2) // ex-99 pieces
{
for (i = n = 0; ; i++) // swap first two ex-99 pieces found
  if (tp[i] == 2) if (n == 0) { j = py_ci[i]; n++; } else { k = py_ci[i]; break; }
c1 = py_ef[j][0]; c2 = py_ef[k][0]; c = py[c1]; py[c1] = py[c2]; py[c2] = c;
c1 = py_ef[j][1]; c2 = py_ef[k][1]; c = py[c1]; py[c1] = py[c2]; py[c2] = c;
}
 else { r = 15; goto s6; }

if (py_chk_swap(py) == 16) // axial pieces wrong (any wrong edge pieces were corrected above)
 if (cn >= 2) // ex-999 pieces
{
for (i = n = 0; ; i++) // swap first two ex-999 pieces found
  if (tp[i] == 3) if (n == 0) { j = py_ci[i]; n++; } else { k = py_ci[i]; break; }
c1 = py_cf[j][0]; c2 = py_cf[k][0]; c = py[c1]; py[c1] = py[c2]; py[c2] = c;
c1 = py_cf[j][1]; c2 = py_cf[k][1]; c = py[c1]; py[c1] = py[c2]; py[c2] = c;
c1 = py_cf[j][2]; c2 = py_cf[k][2]; c = py[c1]; py[c1] = py[c2]; py[c2] = c;
}
 else { r = 16; goto s6; }

// check and correct edge pieces validity and parity (see py_chk_edges)
// if parity wrong and not seen in py_gchk, at least one ex-99 piece exists
psum = qn = 0;
while (qn < 6)
{
c = 10 * py[2] + py[6];
for (k = 0; k < 12; k++) if (c == py_edge[k]) goto s42;
r = 2; goto s6; // should never happen, validity confirmed in py_gchk
s42: psum += k % 2;
qn++; if (qn <= 2) py_turn(1);
if (qn == 3) { py_turn(1); py_turn(8); } if (qn == 4) py_turn(2);
if (qn == 5) py_turn(4); if (qn == 6) { py_turn(1); py_turn(4); }
}
if (psum & 1) // parity wrong
{
if (en == 0) { r = 3; goto s6; } // should never happen
for (i = 0; ; i++) if (tp[i] == 2) // flip first ex-99 piece found
  {
  j = py_ci[i]; c1 = py_ef[j][0]; c2 = py_ef[j][1];
  c = py[c1]; py[c1] = py[c2]; py[c2] = c; break;
  }
}

// check and correct axial pieces validity (see ad_chk_corners)
qn = 0;
while (qn < 4)
{
while (1)
  {
  c = 100 * py[1] + 10 * py[7] + py[13];
  for (k = 0; k < 12; k++) if (c == py_axial[k]) goto s46;
  c = py[1]; py[1] = py[7]; py[7] = c; // invalid, change piece orientation
  }
s46: qn++; py_turn(1); if (qn == 3) py_turn(3); if (qn == 4) py_turn(4);
}

goto s8;

// error means goal is invalid and this could not be seen in py_gchk
s6: cb_inp_msg("Invalid goal"); cb_msg_chk(r); fl_nl(); cb_error();
s8: k_gen = 0;
}

/********************/
py_copy(s, d) char *s, *d;
{
// copy position (sz_py bytes) from s to d

/* the asm below
fl_move(s, d, sz_py);
*/
_asm
#if win32
{                         ; sz_py 24 = 2x12
        mov   esi, s
        mov   edi, d
        mov   ecx, 2      ; 2x3 mov pairs is faster than 6 rep movsd
py_cp3: mov   eax, [esi]
        mov   [edi], eax
        mov   eax, [esi+4]
        mov   [edi+4], eax
        mov   eax, [esi+8]
        mov   [edi+8], eax
        add   esi, 12
        add   edi, 12
        loop  py_cp3
}
#else
{                         ; sz_py 24 = 2x12
        mov   si, s
        mov   di, d
        mov   ecx, 2      ; 2x3 mov pairs is faster than 6 rep movsd
py_cp3: mov   eax, [si]
        mov   [di], eax
        mov   eax, [si+4]
        mov   [di+4], eax
        mov   eax, [si+8]
        mov   [di+8], eax
        add   si, 12
        add   di, 12
        loop  py_cp3
}
#endif
}

/********************/
py_turn(t) short t;
{
static char pft[8][4][6] = // 8 lines x 4x6 "from" offsets in py array
  {  4, 3, 0, 5, 2, 1,  14,17,16,13,12,15,  22,21,18,23,20,19,   6, 7, 8, 9,10,11,
    18,19,20,21,22,23,  10, 9, 6,11, 8, 7,   2, 5, 4, 1, 0, 3,  16,15,12,17,14,13,
     8,11,10, 7, 6, 9,  20,23,22,19,18,21,  16,15,12,17,14,13,   2, 5, 4, 1, 0, 3,
    12,13,14,15,16,17,   0, 1, 2, 3, 4, 5,   6, 7, 8, 9,10,11,  22,21,18,23,20,19,
     2, 5, 4, 1, 0, 3,  18,19,20,21,22,23,  10, 9, 6,11, 8, 7,  14,17,16,13,12,15,
    16,15,12,17,14,13,   8,11,10, 7, 6, 9,  20,23,22,19,18,21,   0, 1, 2, 3, 4, 5,
    22,21,18,23,20,19,   4, 3, 0, 5, 2, 1,  14,17,16,13,12,15,  10, 9, 6,11, 8, 7,
     6, 7, 8, 9,10,11,  12,13,14,15,16,17,   0, 1, 2, 3, 4, 5,  20,23,22,19,18,21 };
char *pt, temp[sz_py];

// perform turn t (internal value 1-8)

if (t < 1 || t > 8) cb_msg_mt(t, 1); // should never happen
py_copy(py, temp); pt = pft[t - 1];

/* the asm below
short i;
for (i = 0; i < sz_py; i++) py[i] = temp[pt[i]];
*/
_asm
#if win32
{
        mov   esi, pt
        lea   edi, temp
        mov   edx, py
        mov   ecx, sz_py/2      ; write words
py_tn3: movzx ebx, byte ptr [esi]
        mov   al, [edi+ebx]
        movzx ebx, byte ptr [esi+1]
        mov   ah, [edi+ebx]
        mov   [edx], ax
        add   esi, 2
        add   edx, 2
        loop  py_tn3
}
#else
{
        mov   si, pt
        lea   di, temp
        mov   dx, py
        mov   ecx, sz_py/2      ; write words
py_tn3: movzx bx, byte ptr [si]
        mov   al, [di+bx]
        movzx bx, byte ptr [si+1]
        mov   ah, [di+bx]
        mov   bx, dx
        mov   [bx], ax
        add   si, 2
        add   dx, 2
        loop  py_tn3
}
#endif
}

/********************/
py_move(m) short m;
{
static short pof[8] = // pre-calculated pft row offsets
  { 0,15,30,45,60,75,90,105 };
static char pft[8][15] = // 8 lines x 5 sets of 3 circular offsets in py array
  { 4, 0, 2, 3, 1, 5,14, 6,18,17, 7,19,13, 9,21,
   19, 1,15,20, 2,12,23, 5,13,10, 6, 8, 9, 7,11,
    8, 0,22,11, 1,21, 7, 3,23,16,12,14,15,13,17,
   15, 3, 9,16, 4,10,17, 5,11,22,18,20,21,19,23,
    2, 0, 4, 5, 1, 3,18, 6,14,19, 7,17,21, 9,13,
   15, 1,19,12, 2,20,13, 5,23, 8, 6,10,11, 7, 9,
   22, 0, 8,21, 1,11,23, 3, 7,14,12,16,17,13,15,
    9, 3,15,10, 4,16,11, 5,17,20,18,22,23,19,21 };

// perform move m (internal value 1-8)

if (m < 1 || m > 8) cb_msg_mt(m, 0); // should never happen

/* the asm below
short i, j;  char *pt, *pq, c1, c2;
pt = pft[m - 1];
for (i = 0; i < 5; i++)
{
c1 = py[*(pq = pt)];
for (j = 0; j < 2; j++) { pt++; c2 = py[*pt]; py[*pt] = c1; c1 = c2; }
py[*pq] = c1; pt++;
}
*/
_asm
#if win32
{
        movzx esi, word ptr m
        shl   esi, 1
        movzx esi, word ptr pof[esi-2]
        lea   esi, pft[esi]  ; pft row # m-1
        mov   edi, py
        xor   ebx, ebx
        mov   ecx, 5         ; 5 sets of 3 elements
py_mv3: mov   bl, [esi]      ; rotate the contents of each set
        mov   edx, ebx       ; save ebx for last mov
        mov   al, [edi+ebx]
        mov   bl, [esi+1]
        mov   ah, [edi+ebx]  ; mov+mov is faster than xchg
        mov   [edi+ebx], al
        mov   bl, [esi+2]
        mov   al, [edi+ebx]
        mov   [edi+ebx], ah
        mov   [edi+edx], al
        add   esi, 3
        loop  py_mv3
}
#else
{
        mov   si, word ptr m
        shl   si, 1
        mov   si, word ptr pof[si-2]
        lea   si, pft[si]  ; pft row # m-1
        mov   di, py
        xor   bx, bx
        mov   ecx, 5       ; 5 sets of 3 elements
py_mv3: mov   bl, [si]     ; rotate the contents of each set
        mov   dl, bl       ; save bl for last mov
        mov   al, [di+bx]
        mov   bl, [si+1]
        mov   ah, [di+bx]  ; mov+mov is faster than xchg
        mov   [di+bx], al
        mov   bl, [si+2]
        mov   al, [di+bx]
        mov   [di+bx], ah
        mov   bl, dl
        mov   [di+bx], al
        add   si, 3
        loop  py_mv3
}
#endif
}

/********************/
py_gcomp()
{
short i, j, k, s, r;  char *dd, *gg;

// compare py with py_goal (any orientation), return 0 if matching, 1 if not

for (i = 0; i < 3; i++) // try all 12 possible orientations
{
for (j = 0; j < (i == 0 ? 3 : i == 1 ? 1 : 0); j++)
  {
  for (k = 0; k < 3; k++)
    {
    for (s = 0, dd = py, gg = py_goal; s < sz_py; s++, dd++, gg++)
      if (*gg != 9 && *gg != *dd) goto s4;
    r = 0; goto s9;
    s4: py_turn(4);
    }
  if (i == 0) py_turn(2);
  }
py_turn(1);
}
r = 1; // if no match found, ends in initial orientation

s9: return r;
}

/********************/
py_dcomp()
{
// direct compare: return 0 if solved (any orientation), 1 if not

/* the asm below
short i, j;  char c;
for (i = 0; i < 4; i++) { c = py[6 * i + 0];
  for (j = 1; j < 6; j++) if (py[6 * i + j] != c) return 1; }
return 0;
*/
_asm
#if win32
{
        mov   esi, py
        mov   ecx, 4       ; 4 faces
py_dc3: mov   al, [esi]    ; compare piece 1 with 2
        mov   ah, [esi+1]
        cmp   ah, al
        jne   py_dc6
        cmp   ax, [esi+2]  ; compare pieces 3-4 with 1-2
        jne   py_dc6
        cmp   ax, [esi+4]  ; compare pieces 5-6 with 1-2
        jne   py_dc6
        add   esi, 6
        loop  py_dc3
        xor   eax, eax     ; return 0
        jmp   py_dc62
py_dc6: mov   eax, 1       ; return 1
py_dc62:
#
}
#else
{
        mov   si, py
        mov   ecx, 4       ; 4 faces
py_dc3: mov   al, [si]     ; compare piece 1 with 2
        mov   ah, [si+1]
        cmp   ah, al
        jne   py_dc6
        cmp   ax, [si+2]   ; compare pieces 3-4 with 1-2
        jne   py_dc6
        cmp   ax, [si+4]   ; compare pieces 5-6 with 1-2
        jne   py_dc6
        add   si, 6
        loop  py_dc3
        xor   ax, ax       ; return 0
        jmp   py_dc62
py_dc6: mov   ax, 1        ; return 1
py_dc62:
#
}
#endif
}

/********************/
py_trials(s, desc) short s; char *desc;
{
int py_dcomp(), py_gcomp();

// search for solution by trying move sequences of increasing length

step = s; // step type can be 0, 3
ycomp = step == 0 ? py_dcomp : py_gcomp;
cb_disp_start(); cb_disp_goal(step == 0 ? py_orig : py_goal); cb_disp_lno(desc);

k_fin = 0; cb_incseq(); // initial compare to see if already solved
if (ycomp() == 0) { fl_disp("---"); fl_nl(); k_fin = 1; goto s8; }

py_copy(py, py_save);
for (im = 1; im <= mstop; im++)
{
if (im > 1 && (im - 1) % 5 == 0) fl_sp(); fl_ch('.');
elvl = (edif = im - 14) <= 0 ? 3 : 3 + edif;
py_loop(0); // start recursion
if (k_fin) { fl_sp(); cb_rec_add(20 + cstep[step]); cb_disp_moves();
  cb_disp_end(); goto s8; }
}
fl_nl(); goto s9;

s8: if (step == 0 && fl_ncomp(py, py_orig, sz_py) != 0)
      { cb_disp_goal(py_orig); py_reset(0, 0, 1); cb_disp_end(); }
s9: ;
}

/********************/
py_loop(lvl) short lvl;
{
char *scur, *snext, m, m1;

// try all moves at the current nesting level (called recursively)

snext = (scur = py_save + lvl * 64) + 64; m1 = move[lvl - 1];

for (move[lvl] = 1; (m = move[lvl]) <= 8; move[lvl]++)
{
// skip move combinations that do nothing or have same effect as others
if (lvl > 0) if (m == m1 || m == yrev[m1]) continue;
if (!(lvl > 0 && m == 1)) py_copy(scur, py); py_move(m);
if (lvl + 1 == im) { cb_incseq(); if (ycomp() == 0) { k_fin = 1; break; } }
  else { py_copy(py, snext); py_loop(lvl + 1); }
if (k_fin) break;
}

if (lvl == elvl) cb_esc();
}

/********************/
py_bk_trials()
{
short i, n;

// try back sequences from specified position to attain the required wrong pieces

py_copy(py, py_save);
for (i = 0; i < 10 && (im = bklth[i]) > 0; i++)
{
fl_nl(); fl_disp("Sequence length: "); fl_num(im); fl_nl();
n = bkstnum == 0 ? 5 : 6; elvl = (edif = im - 13) <= 0 ? n : n + edif;
cb_time(0); bkcount = bkseqs = 0;
bktot = 10368; // calculate bktot = highest expected bkcount
if (bkstnum > 0) { n = bkstnum - 1; bktot = 7776; while (n--) bktot /= 6; }
if ((n = edif) > 0) while (n--) bktot *= 6;
py_bk_loop(0); // start recursion
}
}

/********************/
py_bk_loop(lvl) short lvl;
{
short i, n;  char *scur, *snext, m, m1, temp[sz_py];

// try all moves at the current nesting level in a back sequence (called recursively)

snext = (scur = py_save + lvl * 64) + 64; m1 = move[lvl - 1];

for (move[lvl] = 1; (m = move[lvl]) <= 8; move[lvl]++)
{
if (bkstnum > 0 && lvl == bkstnum) // show only sequences starting with bkstmove
  for (i = 0; i < lvl; i++) if (move[i] != bkstmove[i]) goto s6;
// skip move combinations that do nothing or have same effect as others
if (lvl > 0) if (m == m1 || m == yrev[m1]) continue;
if (!(lvl > 0 && m == 1)) py_copy(scur, py); py_move(m);
if (lvl + 1 == im)
   { py_copy(py, temp); py_reset(0, 0, 0); cb_bk_comp(); py_copy(temp, py); }
  else { py_copy(py, snext); py_bk_loop(lvl + 1); }
s6: ;
}

if (lvl == elvl) { bkcount++; cb_bk_erase(); cb_esc(); cb_bk_chg(); }
if (lvl == 0) cb_bk_erase();
}

/********************/

/*** cu functions ***/

/********************/
cu_accept()
{
char ch;

fl_disp("Press Enter to continue, X to cancel: ");
while (1) { if ((ch = fl_up(ds_chin())) == 'X') cb_exit();
  if (ch == '\r') { fl_nl(); break; } }
}

/********************/
cu_rdchar()
{
char ch;

// accept one character from keyboard

fl_kbstr(); cb_disp_wrd(kbuf + 1);
ch = 0; if (*kbuf == 1) ch = fl_up(*(kbuf + 1));
return ch; // return upper case if one character entered, else 0
}

/********************/
cu_rdnum(min, max) short min, max;
{
short n;

// accept unsigned integer in range min-max from keyboard

fl_kbstr(); kptr = kbuf + 1;
if ((n = cb_kb_num(0)) >= 0) if (n < min || n > max) n = -1;
return n; // return number, or -1 if invalid or not 0/min - 29999/max, or -2 if missing
}

/********************/
cu_bklength()
{
short k, n;

// accept sequence lengths of back sequences from keyboard

fl_kbstr(); kptr = kbuf + 1; fl_clr(bklth, 10); k = 0;
while (1)
{
if ((n = cb_kb_num(1)) < 0 || n < 1 || n > 20 || k > 9) goto s91;
bklth[k++] = n; while (*kptr == ' ') kptr++; if (*kptr == 0) break;
}
if (k == 0) goto s91;
cb_disp_wrd(kbuf + 1);

return 0; s91: return 1;
}

/********************/
cu_bkstart()
{
short k, n;  char ch, cl, kr, ks;

// accept starting moves of back sequences from keyboard

fl_kbstr(); fl_clr(bkstmove, 4);
for (kptr = kbuf + 1, k = kr = 0; ch = *kptr; kptr++)
{
if (ch == ' ') continue; if (k == 4) goto s91;
ks = (cl = *(kptr + 1)) == ' ' || cl == 0;
if (ch == '-') { if (kr || ks) goto s91; kr = 1; continue; }
if ((n = ch - '0') < 1 || !ks) goto s91;
if (m_cube) { if (!(n == 1 || n == 3 || n == 5)) goto s91; if (kr) n++; }
if (m_pyramid) { if (n > 4) goto s91; if (kr) n += 4; }
bkstmove[k++] = n; kr = 0;
}
bkstnum = k; cb_disp_wrd(kbuf + 1);

return 0; s91: return 1;
}

/********************/
cu_invalid()
{
fl_disp("Invalid"); fl_nl();
}

/********************/
cu_positions()
{
short k, n;

// custom operation: solve several random positions

while (1)
{
fl_disp("Number of positions (1 - 25000): ");
if ((k = cu_rdnum(1, 25000)) >= 0) { cb_num_wrd(k); break; }
cu_invalid();
}
cu_accept(); fl_nl();

for (n = 1; n <= k; n++) { fl_disp("#"); fl_num(n); fl_spaces(2);
  cb_scramble(0); x_solve(0); }
}

/********************/
cu_backseq()
{
static char *pt[] = { "", "corners", "", "axials" };
static char mw[] = { 0,16,7,9 }; // max wrong pieces (ad fc py)
short n;  char ch;

// custom operation: list back sequences

while (1)
{
fl_disp("Number of wrong pieces (0 - "); fl_num(n = mw[mdl]); fl_disp("): ");
if ((bkwrong = cu_rdnum(0, n)) >= 0) { cb_num_wrd(bkwrong); break; }
cu_invalid();
}

if (bkwrong > 0)
{
fl_disp("Type of wrong pieces (1 - 3)"); fl_nl();
if (mdl == m_fc) { fl_disp("corners only"); bktype = 2; fl_nl(); }
 else while (1)
  {
  fl_disp("1 for edges only, 2 for "); fl_disp(pt[mdl]); fl_disp(" only, 3 for both: ");
  if ((bktype = cu_rdnum(1, 3)) >= 0) { cb_num_wrd(bktype); break; }
  cu_invalid();
  }

while (1)
  {
  fl_disp("1 for flipped in place only, 2 for wrong location only, 3 for both: ");
  if ((bkflip = cu_rdnum(1, 3)) >= 0) { cb_num_wrd(bkflip); break; }
  cu_invalid();
  }
}

while (1)
{
fl_disp("Sequence lengths (up to 10 lengths of 1 - 20): ");
if (cu_bklength() == 0) break; cu_invalid();
}

while (1)
{
fl_disp("List only sequences starting with these (up to 4) moves: ");
if (cu_bkstart() == 0) break; cu_invalid();
}

fl_disp("R to show also the reverse moves: ");
bkrev = 0; if ((ch = cu_rdchar()) == 'R') bkrev = 1;

fl_disp("S to start with script, else starts with solved position: ");
ch = cu_rdchar(); cu_accept();

if (ch == 'S')
{
fl_open_inp(); cb_inp_start(); cb_inp_read(); fl_close_inp();
x_chk(1); if (mdl == m_ad) ad_chk_anchor(ad); x_copy(x_dd, w_pp);
}
  else x_copy(x_oo, w_pp);
}

/********************/
cu_script()
{
short k, n, d;

// custom operation: repeat script

while (1)
{
fl_disp("Number of repetitions (1 - 25000): ");
if ((k = cu_rdnum(1, 25000)) >= 0) { cb_num_wrd(k); break; }
cu_invalid();
}

while (1)
{
fl_disp("Display cycle number every n cycles (1/10/100/1000, 0 for none): ");
if ((d = cu_rdnum(0, 1000)) == -2) d = 0;
if (d == 0 || d == 1 || d == 10 || d == 100 || d == 1000) { cb_num_wrd(d); break; }
cu_invalid();
}
cu_accept(); fl_nl(); fl_open_inp();

for (n = 1; n <= k; n++)
{
if (d > 0 && n % d == 0) { fl_disp("#"); fl_num(n); fl_nl(); }
fl_init_inp(); cb_inp_start(); cb_inp_read();
}

fl_close_inp();
}

/********************/

/*** cb functions ***/

/********************/
cb_help_run()
{
fl_disp("Digicube v1.2, the programmable virtual Rubik's cube"); fl_nl(); fl_nl();
fl_disp("RUN OPTIONS (example: m2 s15 d)"); fl_nl();
fl_disp("M1, M2, M3: Specify Model (cube 3x, cube 2x, pyramid 3x; default: M1)"); fl_nl();
fl_disp("D: Display initial position"); fl_nl();
fl_disp("D1, D2: Display also intermediate positions (2 levels of detail)"); fl_nl();
fl_disp("J, U, A: Display initial position as rows, pieces, schematic"); fl_nl();
fl_disp("R: Show also Reverse moves"); fl_nl();
fl_disp("W: Show also Widely spaced moves"); fl_nl();
fl_disp("S1-S999: Start reading input file at label #1-#999 (default: at beginning)"); fl_nl();
fl_disp("I: Specify Input file or Interactive run (default: file diginp)"); fl_nl();
fl_disp("O: Specify Output file (default: digout)"); fl_nl();
fl_disp("N: Create New output file (default: add to existing file)"); fl_nl();
fl_disp("K: Cancel writing to output file"); fl_nl();
fl_nl();
fl_disp("H: Help (display this list)"); fl_nl();
fl_disp("L: List operations"); fl_nl();
fl_disp("C: Show Custom operations"); fl_nl();
fl_disp("C1-C99: Perform Custom operation 1-99"); fl_nl();
fl_disp("G1-G999: Get run options at label #1-#999 in options file"); fl_nl();
fl_disp("X: End program (while running, stop current operation)"); fl_nl();
fl_disp("Z: License information"); fl_nl();
}

/********************/
cb_help_op()
{
fl_disp("OPERATIONS"); fl_nl();
fl_disp("P, PG: Specify Position, Goal position"); fl_nl();
fl_disp("P1-P6, PG1-PG6: Specify side 1-6 for Position, Goal position"); fl_nl();
fl_disp("E: Exchange position with goal position"); fl_nl();
fl_disp("V, VG: Verify position, Goal position"); fl_nl();
fl_disp("R, RG: Reset position, Goal position to standard orientation"); fl_nl();
fl_disp("S: Solve position"); fl_nl();
fl_disp("SA, SC, SE: Solve position for Anchor, Corners, Edges"); fl_nl();
fl_disp("SD, SG: Solve position Directly, for Goal"); fl_nl();
fl_disp("M: Choose order of trial Moves in solutions (default: M135)"); fl_nl();
fl_disp("K1-K20: Stop after sequences of 1-20 moves (default: M1 K15, M2 K13, M3 K11)"); fl_nl();
fl_disp("D, J, U, A: Display position as list, rows, pieces, schematic"); fl_nl();
fl_disp("DG, JG, UG, AG: Display Goal position as list, rows, pieces, schematic"); fl_nl();
fl_disp("L, *: Display Label, comment"); fl_nl();
fl_disp("Q, Q0: Display, clear count of moves, turns, cycles"); fl_nl();
fl_disp("QC: Display current cycle"); fl_nl();
fl_disp("T1-T99, F1-F99: Copy position To, From memory 1-99"); fl_nl();
fl_disp("TG1-TG99, FG1-FG99: Copy Goal position To, From memory 1-99"); fl_nl();
fl_disp("C1-C99: Compare position with memory 1-99"); fl_nl();
fl_disp("CC1-CC99: Compare Constantly position with memory 1-99"); fl_nl();
fl_disp("CC, CC0: Display, clear CC list"); fl_nl();
fl_disp("Press a key for more: "); ds_chin(); fl_nl();
fl_disp("G1-G99: Compare position with memory 1-99 as Goal"); fl_nl();
fl_disp("GG1-GG99: Compare constantly position with memory 1-99 as Goal"); fl_nl();
fl_disp("GG, GG0: Display, clear GG list"); fl_nl();
fl_disp("ON, OF: Switch On/Off display of input moves and turns (default: OF)"); fl_nl();
fl_disp("N3, N6: Use 3 or 6 faces in random moves (default: N3)"); fl_nl();
fl_disp("NS, NR: Save/Restore random state"); fl_nl();
fl_disp("Y, YP, YT: Scramble position using moves and turns, Pieces, Template"); fl_nl();
fl_disp("I: Begin or end Interactive mode"); fl_nl();
fl_disp("^: Set colors of one piece"); fl_nl();
fl_disp("/, \\: Flip one piece cw or ccw"); fl_nl();
fl_disp("+, -: Set/reset current run options"); fl_nl();
fl_disp("[n ]: Repeat group of elements n times"); fl_nl();
fl_disp("Z: End script in custom operation"); fl_nl();
fl_disp("X: End current run"); fl_nl();
}

/********************/
cb_help_cus()
{
fl_disp("CUSTOM OPERATIONS"); fl_nl();
fl_disp("C1. Solve a random position"); fl_nl();
fl_disp("C2. Solve several random positions"); fl_nl();
fl_disp("C3. Solve benchmark position"); fl_nl();
fl_disp("C4. List back sequences"); fl_nl();
fl_disp("C5. Repeat a script"); fl_nl();
}

/********************/
cb_help_lic()
{
fl_disp("LICENSE"); fl_nl();
fl_disp("Digicube v1.2 (C) Copyright 2018, 2022 Andrei Sorin"); fl_nl();
fl_disp("This program comes with ABSOLUTELY NO WARRANTY."); fl_nl();
fl_disp("This is free software. You may also redistribute it, under certain conditions."); fl_nl();
fl_disp("For details see the GNU General Public License ver. 3 in a file included with"); fl_nl();
fl_disp("  this software, or see http://www.gnu.org/licenses/"); fl_nl();
}

/********************/
cb_exit()
{
if (!k_end) fl_nl(); k_wrf = 0; fl_close_inp();
if (k_out) { cb_reminder(); ds_close(wrhdl); }
if (k_ent) longjmp(jmpbuf[0], 1); exit();
}

/********************/
cb_xrun()
{
char ch;

fl_disp("X to end run, Enter to continue: ");
while (1) { if ((ch = fl_up(ds_chin())) == 'X') cb_exit();
  if (ch == '\r') { fl_nl(); break; } }
}

/********************/
cb_esc()
{
static short count = 0;

// called periodically to check if X pressed to stop current operation
// also updates here the timer variables (no need to do it in each call)
if (++count > 100) { count = 0; cb_tmchk(); }
while (ds_kbchk()) if (fl_up(ds_chin()) == 'X') { fl_nl(); cb_error(); }
}

/********************/
cb_error()
{
if (!k_read) { k_end = 1; cb_exit(); } // custom operation and not in cb_inp_read
if (!k_int) cb_xrun();
if (k_solve) { cb_pos_restore(); cb_goal_restore(); k_solve = 0; stype = 0; }
if (k_gen) { cb_pos_restore(); k_gen = 0; }
longjmp(jmpbuf[1], 1);
}

/********************/
cb_init()
{
int ad_chk(), ad_gchk(), ad_solve(), ad_reset(), ad_move(), ad_turn(),
  ad_copy(), ad_gen();
int fc_chk(), fc_gchk(), fc_solve(), fc_reset(), fc_move(), fc_turn(),
  fc_copy(), fc_gen();
int py_chk(), py_gchk(), py_solve(), py_reset(), py_move(), py_turn(),
  py_copy(), py_gen();
short i;  char *cp;

// initialize pointers and arrays

if (mdl == m_ad)
{
ad_save = (ad_goal = (ad = memcom) + 64) + 64;
x_sz = sz_ad; x_fsz = 9; x_npcs = 27;
x_dd = ad; x_oo = ad_orig; x_gg = ad_goal; x_nn = ad_generic;
x_cn = ad_cn; x_cl = ad_cl; x_cp = ad_cp; x_ci = ad_ci; x_ef = ad_ef; x_cf = ad_cf;
x_chk = ad_chk; x_gchk = ad_gchk; x_solve = ad_solve; x_reset = ad_reset;
x_move = ad_move; x_turn = ad_turn; x_copy = ad_copy; x_gen = ad_gen;
ad_copy(ad_anchor, ad_goal);
}

if (mdl == m_fc)
{
fc_save = (fc_goal = (fc = memcom) + 64) + 64;
x_sz = sz_fc; x_fsz = 4; x_npcs = 8;
x_dd = fc; x_oo = fc_orig; x_gg = fc_goal; x_nn = fc_generic;
x_cn = fc_cn; x_cl = fc_cl; x_cp = fc_cp; x_ci = fc_ci; x_cf = fc_cf;
x_chk = fc_chk; x_gchk = fc_gchk; x_solve = fc_solve; x_reset = fc_reset;
x_move = fc_move; x_turn = fc_turn; x_copy = fc_copy; x_gen = fc_gen;
fc_copy(fc_anchor, fc_goal);
}

if (mdl == m_py)
{
py_save = (py_goal = (py = memcom) + 64) + 64;
x_sz = sz_py; x_fsz = 6; x_npcs = 10;
x_dd = py; x_oo = py_orig; x_gg = py_goal; x_nn = py_generic;
x_cn = py_cn; x_cl = py_cl; x_cp = py_cp; x_ci = py_ci; x_ef = py_ef; x_cf = py_cf;
x_chk = py_chk; x_gchk = py_gchk; x_solve = py_solve; x_reset = py_reset;
x_move = py_move; x_turn = py_turn; x_copy = py_copy; x_gen = py_gen;
py_copy(py_aaxials, py_goal);
}

x_copy(x_oo, x_dd); for (i = 0; i < 98; i++) x_copy(x_oo, mempos + i * x_sz);
x_copy(x_nn, mempos + 98 * x_sz);
if (mdl == m_ad) { x_copy(ad_anchor, mempos + 97 * x_sz);
  x_copy(ad_acorners, mempos + 96 * x_sz); }
if (mdl == m_fc) x_copy(fc_anchor, mempos + 97 * x_sz);
if (mdl == m_py) x_copy(py_aaxials, mempos + 97 * x_sz);
fl_clr(cmem, 99); fl_clr(gmem, 99); cb_inp_clr(); mstop = dstop[mdl];
}

/********************/
cb_pos_save()
{
x_copy(x_dd, initpos);
}

/********************/
cb_pos_restore()
{
x_copy(initpos, x_dd);
}

/********************/
cb_goal_save()
{
if (mdl == m_ad) ad_copy(ad_goal, initgoal);
}

/********************/
cb_goal_restore()
{
if (mdl == m_ad) ad_copy(initgoal, ad_goal);
}

/********************/
cb_reminder()
{
#define mbyte 1048576
long flsz;

ofst = 0;
if ((flsz = ds_lseek(wrhdl, 2, &ofst)) > mbyte)
{
fl_disp("Reminder: "); fl_disp(wrfile); fl_disp(" is over ");
fl_num((uns)(flsz / mbyte)); fl_disp(" MB"); fl_nl();
}
}

/********************/
cb_begin()
{
movetot = lno = 0; fl_clr(seqtot, 8); recn = recseq = 0; cb_rec_add(0);
cb_time(0);
}

/********************/
cb_end(st) short st;
{
long t;

if (k_fin) { fl_disp("Moves: "); fl_num(movetot); }
  else { fl_disp("** Stopped"); if (st >= 0) { fl_sp(); fl_num(st + 1); } }
fl_spaces(4); fl_disp("Time: "); t = cb_time(1) + 5; // round display to .1 sec
fl_cnum(t / 100); fl_ch('.'); fl_num((uns)(t % 100 / 10)); fl_disp(" sec");
fl_spaces(4); fl_disp("Sequences tried: "); fl_qnum(seqtot); fl_nl(); fl_nl();
if (k_fin && recseq > 0)
  { if (k_rev) cb_rev_disp(); if (k_wide) cb_wide_disp(); }
}

/********************/
cb_rdopt()
{
short r, cn;  uns num;  char *cp, ks;

// find label #optlabel in options file

r = num = 0;
while (1)
{
if ((cn = fl_rdch()) < 0)
  { fl_disp("Options label #"); fl_num(optlabel); fl_disp(" not found"); cb_exit(); }
if (cn == 0) goto s8; if (eol(cn)) { k_eol = 1; continue; }
if (!(k_eol && cn == '#')) { k_eol = 0; continue; }
while (1) { cn = fl_nextch(); if (!dig(cn)) break; if (num > 99) { num = 0; break; }
  fl_rdch(); num = num * 10 + cn - '0'; }
k_eol = 0; if (num == optlabel) break; num = 0;
}

*kbuf = 0; cp = kbuf + 1; ks = 1; // copy options list to keyboard buffer
while (1)
{
if (cp == kbuf + sz_kbuf - 2) { r = 1; goto s9; }
if ((cn = fl_rdch()) < 0 || eol(cn)) break; if (cn == 0) goto s8;
if (ks && spc(cn)) continue; ks = 0; *cp++ = cn; (*kbuf)++;
}
*cp = 0; goto s9;

s8: fl_disp("Error reading options file"); cb_exit();
s9: return r; // return 0, or 1 if options list too long
}

/********************/
cb_inp_start()
{
short cn;  uns num;

// find label #stlabel in input file

if (stlabel == 0) goto s9; num = 0;
while (1)
{
if ((cn = fl_rdch()) < 0)
  { fl_disp("Label #"); fl_num(stlabel); fl_disp(" not found"); cb_exit(); }
if (cn == 0) { fl_disp("Error reading input file"); cb_exit(); }
if (eol(cn)) { k_eol = 1; continue; }
if (!(k_eol && cn == '#')) { k_eol = 0; continue; }
while (1) { cn = fl_nextch(); if (!dig(cn)) break; if (num > 99) { num = 0; break; }
  fl_rdch(); num = num * 10 + cn - '0'; }
k_eol = 0; if (num == stlabel) break; num = 0;
}

s9: ;
}

/********************/
cb_inp_read()
{
static char com[] = "EYRVSMKONFTCGQDJUA+-^/\\";
static char *emsg[] = // script error messages
  { "", "Unexpected end of input file", "Error reading input file",  // 1 2
    "Invalid data", "Invalid operation", "Missing \")\"",            // 3 4 5
    "Missing quotation mark", "Invalid for current model",           // 6 7
    "Invalid location", "Invalid colors", "Invalid 9's",             // 8 9 10
    "Invalid repeat count", "Invalid run option", "Invalid memory",  // 11 12 13
    "Incomplete operation", "Misplaced \"[\" or \"]\"",              // 14 15
    "Missing \"]\"", "Operation invalid in loop",                    // 16 17
    "Loop limit exceeded", "Invalid move", "Invalid turn" };         // 18 19 20
static char *imsg[] = // interactive error messages
  { "", "Incomplete operation", "Invalid data", "Invalid operation", // 1 2 3
    "Invalid for current model", "Invalid location",                 // 4 5
    "Invalid colors", "Invalid 9's", "Invalid run option",           // 6 7 8
    "Invalid move", "Invalid turn", "Invalid repeat count",          // 9 10 11
    "Too many moves/turns", "Missing \"]\"", "Invalid memory" };     // 12 13 14
static uns esnum = 0;
short i, k, r, s, cn, cv, mt;  char *cp, kt, kr, kc;

// process script and interactive elements

k_read = 1; k_rep = k_rex = cycle = 0;
while (1)
{
setjmp(jmpbuf[1]); if (++esnum == 0) cb_esc();
if (k_rex) // read stored loop elements
  {
  if (repn == replast) // end cycle
    {
    if (--repnum == 0) k_rex = cycle = 0; else { repn = 0; cycle++; ccount++; }
    continue;
    }
  cn = rval[repn++]; kt = cn >> 7; kr = (cn & 0x7f) >> 6; // see "store" below
  if ((cn &= 0x3f) > 9) cn += 32;
  switch (cn)
    {
    case 'Y': i_scr = rval[repn++]; break;
    case 'R': case 'V': i_goal = rval[repn++]; break;
    case 'S': i_stype = rval[repn++]; break;
    case 'M': i_mord = *(uns*)&rval[repn]; repn += 2; break;
    case 'K': i_stop = rval[repn++]; break;
    case 'O': i_wil = rval[repn++]; break;
    case 'N': i_ran = rval[repn++]; break;
    case 'F': case 'T': case 'C': case 'G': i_mem = rval[repn++]; break;
    case 'D': case 'J': case 'U': case 'A': i_goal = rval[repn++]; // continues at l-q
    case 'L': case 'Q': fl_copy(rval + repn, i_lstr);
      repn += fl_lth(i_lstr) + 1; break;
    case '+': case '-': i_opt = rval[repn++]; i_optn = rval[repn++]; break;
    case '^': i_loc = *(uns*)&rval[repn]; i_col = *(uns*)&rval[repn + 2];
      repn += 4; break;
    case '/': case '\\': i_loc = *(uns*)&rval[repn]; repn += 2; break;
    }
  }
 else // read elements in input file
  {
  if (k_irun) cn = 'I'; // interactive run, never returns here
    else if ((cn = cb_inp_val(&kt, &kr)) == -1)
      if (k_rep) { r = 16; goto s8; } else { cb_inp_wil(); k_end = 1; cb_exit(); }
  if (cn < 0) { r = -cn; goto s8; }
  if (cn == '[') { k_rep = 1; repn = 0; continue; }
  if (cn == ']')
    { k_rep = 0; k_rex = 1; replast = repn; repn = 0; cycle = 1; ccount++; continue; }
  if (k_rep) // store loop elements: 0 - 9 = move/turn, 10 - 63 = char-32, bits 6 7 = kr kt
    {
    if (fl_offset(cn, "XZP") >= 0) { r = 17; goto s8; }
    if (repn == sz_rep) { r = 18; goto s8; }
    if (cn <= 9) if (cn > 6) { r = kt ? 20 : 19; goto s8; }
        else if (m_pyramid && cn > 4) { r = 7; goto s8; }
    rval[repn++] = cn <= 9 ? kt << 7 | kr << 6 | cn : cn - 32;
    switch (cn)
      {
      case 'R': case 'V': if (repn > sz_rep - 1) { r = 18; goto s8; }
        rval[repn++] = i_goal; break;
      case 'Y': case 'S': case 'O': case 'N':
      case 'F': case 'T': case 'C': case 'G':
        if (repn > sz_rep - 1) { r = 18; goto s8; }
        rval[repn++] = cn == 'Y' ? i_scr : cn == 'S' ? i_stype
          : cn == 'O' ? i_wil : cn == 'N' ? i_ran : i_mem;
        break;
      case 'M': if (repn > sz_rep - 2) { r = 18; goto s8; }
        *(uns*)&rval[repn] = i_mord; repn += 2; break;
      case 'K': if (repn > sz_rep - 1) { r = 18; goto s8; }
        rval[repn++] = i_stop; break;
      case 'D': case 'J': case 'U': case 'A':
        if (repn > sz_rep - 1) { r = 18; goto s8; }
        rval[repn++] = i_goal; // continues at l-q
      case 'L': case 'Q':
        if (repn > sz_rep - (s = fl_lth(i_lstr) + 1)) { r = 18; goto s8; }
        fl_copy(i_lstr, rval + repn); repn += s; break;
      case '+': case '-': if (repn > sz_rep - 2) { r = 18; goto s8; }
        rval[repn++] = i_opt; rval[repn++] = i_optn; break;
      case '^': if (repn > sz_rep - 4) { r = 18; goto s8; }
        *(uns*)&rval[repn] = i_loc; *(uns*)&rval[repn + 2] = i_col; repn += 4; break;
      case '/': case '\\': if (repn > sz_rep - 2) { r = 18; goto s8; }
        *(uns*)&rval[repn] = i_loc; repn += 2; break;
      }
    continue;
    }
  }

// perform script elements
if (fl_offset(cn, "XYRVSDJUAI") >= 0) cb_inp_wil();
if (fl_offset(cn, com) >= 0) { cb_inp_com(cn); continue; }
switch (cn)
  {
  case 'X': k_end = 1; cb_exit();
  case 'Z': if (custom > 0) { cb_inp_wil(); k_wil = k_read = 0; goto s9; } continue;
  case 'P': if ((k = i_face) > 0) k = (i_face - 1) * x_fsz;
    for (i = 0; i < (i_face == 0 ? x_sz : x_fsz); i++)
      {
      if ((cv = cb_inp_val(&kt, &kr)) == -1) { r = 1; goto s8; }
      if (cv == -2) { r = 2; goto s8; }
      if (!(cv >= 1 && cv <= 6 || cv == 9) || kt || kr) { r = 3; goto s8; }
      if (m_pyramid && (cv == 5 || cv == 6)) { r = 7; goto s8; }
      if (i_goal == 0) x_dd[k + i] = cv; else x_gg[k + i] = cv;
      }
    if (i_goal == 0) cb_inp_match(); continue;
  case 'L': cb_inp_msg(i_lstr); fl_nl(); continue;
  case 'I': fl_disp("Begin interactive"); fl_nl();
    k_int = 1; cb_inp_iprompt();
    while (1)
      {
      setjmp(jmpbuf[1]); fl_disp("> ");
      if ((cv = cb_kb_val()) == 0) { cb_inp_iprompt(); continue; }
      if (cv > 0) // perform interactive elements
       if (fl_offset(cv, com) >= 0) cb_inp_com(cv);
        else switch (cv)
        {
        case 'X': break;
        case 'P': k = (i_face - 1) * x_fsz;
          for (i = 0; i < x_fsz; i++)
            if (i_goal == 0) x_dd[k + i] = i_fpos[i]; else x_gg[k + i] = i_fpos[i];
          if (i_goal == 0) cb_inp_match(); break;
        case 'L': case '*': break;
        case 'I': break;
        case 1: kc = i_rnum == 0; // moves/turns: i_rnum = repeat count, 0 if no repeat
          while (kc || i_rnum-- > 0)
            {
            if (++esnum == 0) cb_esc(); if (kc) kc = 0; else ccount++;
            for (i = 0; i < i_mnum; i++) // i_mnum = number of moves/turns entered
              { // bits 0-5 = move/turn, bits 6 7 = kr kt
              mt = i_mov[i]; kt = mt >> 7; kr = (mt & 0x7f) >> 6; mt &= 0x3f;
              cb_inp_move(&mt, &kt, &kr); cb_inp_wdisp(mt, kt, kr); cb_inp_match();
              }
            }
          cb_inp_wil(); break;
        }
      if (cv < 0)
        {
        fl_disp(imsg[-cv]); // display up to 12 characters preceding error (moves/turns)
        if (ekptr != 0) { if ((cp = ekptr - 12) < kbuf + 1) cp = kbuf + 1;
          fl_disp(": "); while (cp <= ekptr) fl_ch(*cp++); }
        fl_nl(); continue; // interactive loop
        }
      if (cv == 'I' || cv == 'X')
        {
        fl_disp("End interactive"); fl_nl();
        if (k_irun || cv == 'X') { k_end = 1; cb_exit(); }
        wilpos = 0; k_int = 0; break; // interactive loop
        }
      }
    continue; // main loop
  }

// move or turn
if ((r = cb_inp_move(&cn, &kt, &kr)) == 1) { r = kt ? 20 : 19; goto s8; }
if (r == 2) { r = 7; goto s8; }
cb_inp_wdisp(cn, kt, kr); cb_inp_match();
}
// never exits main loop here

s8: cb_inp_msg(emsg[r]); fl_disp(": "); fl_disp(i_estr); cb_exit(); // error
s9: ; // only custom operations exit here
}

/********************/
cb_inp_com(cn) short cn;
{
// perform common script and interactive elements

switch(cn)
{
case 'E': cb_inp_xchg(); cb_inp_match(); break;
case 'Y': cb_scramble(i_scr); cb_inp_match(); break;
case 'R': x_reset(i_goal, !k_int, 1); cb_inp_match(); break;
case 'V': if (i_goal == 0) x_chk(0); else x_gchk(0); break;
case 'S': x_solve(i_stype); cb_inp_match(); break;
case 'M': moveord = i_mord; break;
case 'K': cb_inp_stop(); break;
case 'O': cb_inp_winp(); break;
case 'N': cb_inp_ran(); break;
case 'F': case 'T': cb_inp_mcopy(cn); if (cn == 'F') cb_inp_match(); break;
case 'C': case 'G': cb_inp_cmem(cn); break;
case 'Q': cb_inp_count(); break;
case 'D': case 'J': case 'U': case 'A': cb_inp_pdisp(cn, i_lstr); break;
case '+': case '-': cb_inp_opt(); break;
case '^': cb_set_piece(1); cb_inp_match(); break;
case '/': case '\\': cb_flip_piece(cn == '/' ? 1 : 2); cb_inp_match(); break;
}
}

/********************/
cb_inp_val(at, ar) char *at, *ar;
{
static char ops[] = "XZPEYRVSMKONFTCGLQDJUAI+-^/\\";
short i, r, n, cn, cv, num;  char *cpd, kn, ks, kp, kt, kr;

// read next script element in input file

num = r = kn = ks = kp = kt = kr = 0;

while (1)
{
if ((cn = fl_rdch()) < 0) { r = -1; break; } if (cn == 0) { r = -2; break; }
if (eol(cn)) { cb_inp_sav(' '); if (kp || kr || kn) break; k_eol = 1; continue; }
if (k_eol && cn == '*')
  { while (1) if ((cn = fl_rdch()) < 0 || eol(cn)) break; continue; }
k_eol = 0; cb_inp_sav(cn); cn = fl_up(cn);
if (kr && !(kn || dig(cn))) { r = -3; break; }

if (fl_offset(cn, ops) >= 0)
  {
  cv = fl_nextch(); if (kp) { if (cn == '-') goto s6; break; }
  if (fl_offset(cn, "MONFTCG+-") >= 0)
    if (spc(cv) || eol(cv) || cv <= 0) { r = -14; break; }
  switch (cn)
    {
    case 'P': i_goal = 0;
      if (fl_up(cv) == 'G') { i_goal = 1; cb_inp_sav(fl_rdch()); cv = fl_nextch(); }
      if (!dig(cv)) { i_face = 0; break; } cb_inp_sav(fl_rdch());
      if ((i_face = cv - '0') == 0 || i_face > 6) { r = -4; break; }
      if (m_pyramid && i_face > 4) { r = -7; break; } cv = fl_nextch();
      if (cv < 0 || !(spc(cv) || eol(cv))) { cb_inp_sav(cv); r = -4; } break;
    case 'Y': i_scr = 0;
      if ((n = fl_offset(fl_up(cv), "PT")) >= 0)
        { i_scr = n + 1; cb_inp_sav(fl_rdch()); }
      break;
    case 'R': case 'V': i_goal = 0;
      if (fl_up(cv) == 'G') { i_goal = 1; cb_inp_sav(fl_rdch()); } break;
    case 'S': i_stype = 0;
      if ((n = fl_offset(fl_up(cv), "DGACE")) >= 0)
        { cb_inp_sav(fl_rdch()); i_stype = n + 1; }
      if (mdl != m_ad && i_stype >= 3 && stype <= 5) r = -7; break;
    case 'M': if (!m_cube) { r = -7; break; }
      if ((n = cb_inp_num(999)) <= 0 || cb_inp_mord(n)) r = -4; break;
    case 'K': if (!dig(cv)) { i_stop = 0; break; }
      if ((i_stop = cb_inp_num(max_moves)) < 1) r = -4; break;
    case 'O': if ((n = fl_offset(fl_up(cv), "NF")) < 0)
        { cb_inp_sav(cv); r = -4; break; }
      cb_inp_sav(fl_rdch()); i_wil = n; break;
    case 'N': if ((n = fl_offset(fl_up(cv), "SR36")) < 0)
        { cb_inp_sav(cv); r = -4; break; }
      cb_inp_sav(fl_rdch()); if (!m_cube && n >= 2) { r = -7; break; }
      i_ran = n; break;
    case 'F': case 'T': i_goal = 0;
      if (fl_up(cv) == 'G') { i_goal = 1; cb_inp_sav(fl_rdch()); }
      if ((i_mem = cb_inp_num(99)) < 1) r = -13; if (i_goal == 1) i_mem += 100;
      break;
    case 'C': case 'G': if (fl_up(cv) == cn)
        {
        cb_inp_sav(fl_rdch());
        if ((cv = fl_nextch()) == '0') { cb_inp_sav(fl_rdch()); i_mem = 0; break; }
        if (!dig(cv)) { i_mem = 100; break; }
        if ((i_mem = cb_inp_num(99)) < 1) r = -13; else i_mem += 100; break;
        }
      if ((i_mem = cb_inp_num(99)) < 1) r = -13; break;
    case 'L': case 'Q': case 'D': case 'J': case 'U': case 'A':
      if (cn == 'Q' && (n = fl_offset(fl_up(cv), "0C")) >= 0)
        { cb_inp_sav(fl_rdch()); *i_lstr = n + 1; *(i_lstr + 1) = 0; break; }
      if (fl_offset(cn, "DJUA") >= 0) { i_goal = 0; if (fl_up(cv) == 'G')
          { i_goal = 1; cb_inp_sav(fl_rdch()); cv = fl_nextch(); } }
      *(cpd = i_lstr) = 0; if (cv != '"') break; cb_inp_sav(fl_rdch());
      while (1)
        {
        if ((cv = fl_nextch()) < 0 || eol(cv)) { r = -6; break; }
        if (cv == 0) { r = -2; break; }
        cb_inp_sav(fl_rdch()); if (cv == '"') break;
        if (cpd < i_lstr + (cn == 'L' || cn == 'Q' ? sz_lstr - 1 : sz_dstr))
          { *cpd++ = cv; *cpd = 0; }
        }
      break;
    case '+': case '-': if ((n = fl_offset(fl_up(cv), "MDJUARWK")) < 0)
        { if (cn == '-') goto s6; cb_inp_sav(cv); r = -12; break; }
      cb_inp_sav(fl_rdch()); i_opt = fl_up(cv);
      if (cn == '-') { i_optn = 0; break; } if (n >= 2) { i_optn = 1; break; }
      if (i_opt == 'M') { if ((i_optn = cb_inp_num(3)) < 1) r = -12; break; }
      cv = fl_nextch(); if (!dig(cv)) { i_optn = 1; break; } // i_opt='d'
      if ((i_optn = cb_inp_num(2) + 1) < 2) r = -12; break;
    case '^': if ((i_loc = cb_inp_num(999)) <= 0) { r = -8; break; }
      if (fl_nextch() == '=') { cb_inp_sav(fl_rdch());
         if ((i_col = cb_inp_num(999)) <= 0) { r = -9; break; } }
        else i_col = i_loc;
      if ((r = cb_set_piece(0)) > 0) r = r == 1 ? -8 : r == 2 ? -9 : -10; break;
    case '/': case '\\':
      if ((i_loc = cb_inp_num(999)) <= 9 || cb_flip_piece(0)) r = -8; break;
    }
  if (r < 0) break; // main loop
  cv = fl_nextch(); // trap operation joined to following element
  if (spc(cv) || eol(cv) || cv <= 0 || fl_offset(cv, "[](") >= 0) r = cn;
    else { cb_inp_sav(cv); r = -4; }
  break; // main loop
  }

s6: if (spc(cn)) { if (kn) if (kp) ks = 1; else break; continue; }
if (cn == '-') { if (kn || kr) { r = -3; break; } kr = 1; continue; }
if (cn == '(') { if (kp) { r = -3; break; } kp = kt = 1; continue; }
if (cn == ')') { if (!kp || !kn) r = -3; kp = 0; break; }
if (cn == '[')
  {
  if (k_rep) { r = -15; break; } repnum = kn = 0;
  while (1)
    {
    if ((cv = fl_nextch()) == ']') { cb_inp_sav(cv); break; }
    if ((cv = fl_rdch()) < 0) { r = -1; break; } if (cv == 0) { r = -2; break; }
    if (eol(cv)) { if (kn) break; k_eol = 1; continue; }
    if (k_eol && cv == '*')
      { while (1) if ((cv = fl_rdch()) < 0 || eol(cv)) break; continue; }
    k_eol = 0; cb_inp_sav(cv);
    if (spc(cv)) { if (kn) break; continue; }
    if (!dig(cv) || repnum > 99999999) { r = -11; break; }
    repnum = repnum * 10 + cv - '0'; kn = 1;
    }
  if (r <= 0 && r >= -2) r = repnum == 0 ? -11 : cn; break; // main loop
  }
if (cn == ']') { r = !k_rep ? -15 : cn; break; }

// here must be number<=9 (move, turn, color)
if (ks || !dig(cn) || num > 0) { r = -3; break; }
num = num * 10 + cn - '0'; kn = 1;
if (fl_offset(cv = fl_nextch(), "[](") >= 0) break;
// trap number joined to following element
if (fl_offset(fl_up(cv), ops) >= 0) { cb_inp_sav(cv); r = -3; break; }
}
// here r=0, operation character, '[', ']', or error code <0

if (kp) r = -5; if (kr && (!kn || num == 0)) r = -3;
if (r == 0 || (r == -1 || r == -2) && kn) r = num;

*at = kt; *ar = kr;
return r;
// return 0-9 (move/turn/color) and set *at *ar (turn & reverse flags),
// or operation character, or "[", "]", or error code: -1 eof, -2 read error,
// -3 invalid data, -4 invalid operation, -5 missing ")", -6 missing quotation mark,
// -7 invalid for current model, -8 invalid location, -9 invalid colors,
// -10 invalid 9's, -11 invalid repeat count, -12 invalid run option,
// -13 invalid memory, -14 incomplete operation, -15 misplaced "[" or "]"
}

/********************/
cb_inp_sav(cn) short cn;
{
// save last few characters read for error messages, compressing spaces

if (cn == '\t') cn = ' ';
if (!(eol(cn) || cn <= 0 || cn == ' ' && *(i_ept - 1) == ' '))
  if (i_ept < i_estr + sz_estr - 1) { *i_ept = cn; *++i_ept = 0; }
    else { fl_move(i_estr + 1, i_estr, sz_estr - 2); *(i_ept - 1) = cn; }
}

/********************/
cb_inp_num(max) short max;
{
short cn, num;  char kn;

// read unsigned integer 0 - max (or 999) in input file

num = kn = 0; if (max > 999) max = 999;
while (1) { cn = fl_nextch(); if (!dig(cn)) break; cb_inp_sav(fl_rdch());
  if (num <= max) num = num * 10 + cn - '0'; kn = 1; }
if (!kn) cb_inp_sav(cn); if (num > max || !kn) num = -1;
return num; // return number, or -1 if >max or missing
// if missing, next character is added to error message string
}

/********************/
cb_inp_winp()
{
// perform operations on, of

if (i_wil == 0 && !k_wil) { k_wil = 1; wilpos = 0;
  fl_disp("Disp on"); if (!k_int) wilpos += 8; else fl_nl(); }
if (i_wil == 1 && k_wil)
  { if (wilpos > 0) fl_sp(); fl_disp("Disp off"); fl_nl(); k_wil = 0; }
}

/********************/
cb_inp_wil()
{
if (k_wil && wilpos > 0) { fl_nl(); wilpos = 0; }
}

/********************/
cb_inp_msg(str) char *str;
{
cb_inp_wil(); fl_disp(str);
}

/********************/
cb_inp_cmsg(str) char *str;
{
cb_inp_wil(); if (fl_lth(str) > 0) { fl_disp(str); fl_disp(": "); }
}

/********************/
cb_inp_wdisp(n, kt, kr) short n, kt, kr;
{
// display input move or turn

if (k_wil)
{
fl_sp(); if (kt) fl_ch('('); if (kr) fl_ch('-'); fl_num(n); if (kt) fl_ch(')');
wilpos += 2; if (kt) wilpos += 2; if (kr) wilpos++;
if (wilpos >= 73) { fl_nl(); wilpos = 0; }
}
if (kt) tcount++; else mcount++;
}

/********************/
cb_inp_stop()
{
// perform operation k

mstop = i_stop == 0 ? dstop[mdl] : i_stop;
}

/********************/
cb_inp_ran()
{
// perform operations ns, nr, n3, n6

if (i_ran == 0) cb_ran_save(); else if (i_ran == 1) cb_ran_restore();
  else k_six = i_ran - 2;
}

/********************/
cb_inp_count()
{
static char msg[] = "Clear count";

// perform operations q q0 qc

if (*i_lstr == 1) // q0
   { cb_inp_clr(); if (!k_int) cb_inp_msg(msg); else fl_disp(msg); fl_nl(); }
  else if (*i_lstr == 2) { cb_inp_msg("#"); fl_lnum(cycle); fl_nl(); } // qc
    else { if (!k_int) cb_inp_cmsg(i_lstr); cb_inp_cdisp(1); } // q
}

/********************/
cb_inp_clr()
{
mcount = tcount = ccount = 0;
}

/********************/
cb_inp_cdisp(kc) short kc;
{
char kz, kdm, kdt, kdc;

// display count of moves, turns, cycles if not 0, always if kc=1

kz = mcount == 0 && tcount == 0 && ccount == 0;
if (!kc && !kz) fl_disp(": ");
kdm = kc && kz || mcount; kdt = kc && kz || tcount; kdc = kc && kz || ccount;
if (kdm) fl_lns(mcount, " move");
if (kdt) { if (kdm) fl_disp(", "); fl_lns(tcount, " turn"); }
if (kdc) { if (kdm || kdt) fl_disp(", "); fl_lns(ccount, " cycle"); }
fl_nl();
}

/********************/
cb_inp_mcopy(ft) short ft;
{
short mno;  char *cp, *mp;

// copy position to/from memory

if (i_mem <= 99) { cp = x_dd; mno = i_mem; }
  else { cp = x_gg; mno = i_mem - 100; }
mp = mempos + (mno - 1) * x_sz;
if (ft == 'F') x_copy(mp, cp); else x_copy(cp, mp);
}

/********************/
cb_inp_cmem(cg) short cg;
{
short i, r, mm, pos;  char *tmem;

// perform operations c cc (if cg='C') or g gg (if cg='G')
// i_mem= 1-99 for c/g 1-99, 101-199 for cc/gg 1-99, 0 for cc0/gg0, 100 for cc/gg

r = 0; tmem = cg == 'C' ? cmem : gmem;

if (i_mem == 0)
{
fl_clr(tmem, 99);
cb_inp_wil(); fl_disp(cg == 'C' ? "Clear CC list" : "Clear GG list");
fl_nl(); goto s9;
}

if (i_mem == 100)
{
cb_inp_wil(); fl_disp(cg == 'C' ? "CC:" : "GG:"); pos = 3;
for (i = 0; i < 99 && (mm = tmem[i]) > 0; i++) { fl_sp(); fl_num(mm);
  pos += mm <= 9 ? 2 : 3; if (pos > 60) { fl_nl(); pos = 0; } }
if (pos > 0) fl_nl(); goto s9;
}

if (i_mem < 100) { cb_inp_comp(cg, i_mem); goto s9; } // i_mem=1-99

for (i = 0; i < 99; i++) // i_mem=101-199
  if (tmem[i] == i_mem - 100) goto s9;
    else if (tmem[i] == 0) { tmem[i] = i_mem - 100; goto s9; }

s9: ;
}

/********************/
cb_inp_match()
{
short i, j, mm;  char *tmem, cg;

// check for memory match in cc & gg lists

for (i = 0; i <= 1; i++)
{
if (i == 0) { tmem = cmem; cg = 'C'; } else { tmem = gmem; cg = 'G'; }
for (j = 0; j < 99 && (mm = tmem[j]) > 0; j++)
  if (cb_inp_comp(cg, mm) == 0) break;
}
}

/********************/
cb_inp_comp(cg, mno) short cg, mno;
{
short i, r;  char *mp, *dd;

// compare current position with memory mno, return 0 if match, 1 if not
// cg='C' for operations c cc, 'G' for operations g gg

r = 1; mp = mempos + (mno - 1) * x_sz;
if (cg == 'C') if (fl_ncomp(x_dd, mp, x_sz) == 0) goto s6; else goto s9;
for (i = 0, dd = x_dd; i < x_sz; i++, dd++, mp++) // cg='G'
  if (*mp != 9 && *mp != *dd) goto s9;
s6: r = 0; cb_inp_msg(cg == 'C' ? "Match mem " : "Match goal mem ");
fl_num(mno); cb_inp_cdisp(0);

s9: return r;
}

/********************/
cb_inp_move(an, at, ar) short *an; char *at, *ar;
{
#define n246 (n == 2 || n == 4 || n == 6)
short n, r;  char kt, kr;

// convert input move/turn value to internal value and perform it
// if 0, generate random internal value, perform it, return input move/turn value

n = *an; kt = *at; kr = *ar; // number, turn, reverse
r = 0; if (n > 6) { r = 1; goto s9; }

if (m_cube)
  if (n == 0)
    {
    *an = n = cb_ran(0, kt, kt ? 0 : k_six); *ar = 0;
    if (n246) { (*an)--; *ar = 1; }
    if (n > 6) if (n & 1) *an -= 5; else { *an -= 6; *ar = 1; }
    }
   else if (n246) if (kt) n -= (kr ? 1 : 0); else n += (kr ? 6 : 5);
          else if (kr) n++;
if (m_pyramid)
  if (n == 0)
    { *an = n = cb_ran(0, kt, 0); *ar = 0; if (n > 4) { *an -= 4; *ar = 1; } }
   else { if (n > 4) { r = 2; goto s9; } if (kr) n += 4; }

if (kt) x_turn(n); else x_move(n);

s9: return r; // return 0 and if random set *an, *ar (move/turn, reverse flag),
// or error code: 1 invalid, 2 invalid for current model
}

/********************/
cb_inp_xchg()
{
char temp[sz_max];

// exchange position with goal position
x_copy(x_dd, temp); x_copy(x_gg, x_dd); x_copy(temp, x_gg);
}

/********************/
cb_inp_mord(m) short m;
{
static short mord[6] = { 135,153,315,351,513,531 }; // valid move orders
short i;

// if m is a valid order of trial moves return 0, else return 1
for (i = 0; i < 6; i++) if (m == mord[i]) { i_mord = i; return 0; } return 1;
}

/********************/
cb_inp_opt()
{
// set/reset run option during run

switch (i_opt)
{
case 'M': if ((mdl = i_optn) == 0) mdl = 1; cb_init();
  if (!k_int) cb_inp_wil(); moveord = k_wil = k_six = 0;
  fl_disp("Initialize M"); fl_num(mdl); fl_nl(); break;
case 'D': disp = i_optn; break;
case 'J': k_rows = i_optn; break;
case 'U': k_pcs = i_optn; break;
case 'A': k_schem = i_optn; break;
case 'R': k_rev = i_optn; break;
case 'W': k_wide = i_optn; break;
case 'K': k_nowrf = i_optn; break;
}
}

/********************/
cb_inp_pdisp(fmt, t) short fmt; char *t;
{
ptf disp;
int cb_disp_list(), cb_disp_rows(), cb_disp_pieces(), cb_disp_schem();

disp = fmt == 'D' ? cb_disp_list : fmt == 'J' ? cb_disp_rows
  : fmt == 'U' ? cb_disp_pieces : cb_disp_schem;
disp(9, t, i_goal == 0 ? x_dd : x_gg);
}

/********************/
cb_inp_iprompt()
{
k_wrf = 0;
fl_disp("Enter operation, moves, turns, I to end interactive");
k_wrf = 1; fl_nl();
}

/********************/
cb_kb_val()
{
short i, n, r, s, sz, cn, cv;  char ch, kc, kt, kr, kb;

// read operation/moves/turns from keyboard buffer in interactive mode

r = 0; ekptr = 0; fl_kbstr(); if ((sz = kbuf[0]) == 0) { cn = 0; goto s9; }
kptr = kbuf; cn = fl_up(*++kptr); cv = fl_up(*++kptr);
if (fl_offset(cn, "XEI") >= 0)
  { if (sz > 1) r = cv == ' ' ? -2 : -3; goto s8; }
if (fl_offset(cn, "PMONFTCG+-") >= 0)
  if (sz == 1 || cv == ' ') { r = -1; goto s8; }

switch (cn)
{
case 'P': i_goal = 0;
  if (cv == 'G') { i_goal = 1; cv = *++kptr;
    if (sz == 2 || cv == ' ') { r = -1; break; } }
  if ((n = cv - '0') <= 0 || n > 6) { r = -3; break; }
  if (m_pyramid && n > 4) { r = -4; break; } i_face = n;
  if ((cv = *++kptr) != 0 && cv != ' ') { r = -3; break; }
  for (i = 0; i < x_fsz; i++)
    {
    while (*kptr == ' ') kptr++;
    if ((n = cb_kb_num(i == x_fsz - 1 ? 0 : 1)) <= 0) { r = -2; break; }
    if (!(n >= 1 && n <= 6 || n == 9)) { r = -2; break; }
    if (m_pyramid && (n == 5 || n == 6)) { r = -4; break; }
    i_fpos[i] = n;
    }
  break;
case 'Y': if (sz == 1) { i_scr = 0; break; } n = fl_offset(cv, "PT");
  if (sz == 2) { if (n >= 0) i_scr = n + 1; else r = -3; break; }
  r = cv == ' ' || n >= 0 && *(kptr + 1) == ' ' ? -2 : -3; break;
case 'D': case 'J': case 'U': case 'A': *i_lstr = 0; // continues at r-v
case 'R': case 'V': if (sz == 1) { i_goal = 0; break; }
  if (sz == 2) { if (cv == 'G') i_goal = 1; else r = -3; break; }
  r = cv == ' ' || cv == 'G' && *(kptr + 1) == ' ' ? -2 : -3; break;
case 'S': if (sz == 1) { i_stype = 0; break; } n = fl_offset(cv, "DGACE");
  if (n < 0) { r = cv == ' ' ? -2 : -3; break; } i_stype = n + 1;
  if (sz > 2) { r = *(kptr + 1) == ' ' ? -2 : -3; break; }
  if (mdl != m_ad && i_stype >= 3 && stype <= 5) r = -4; break;
case 'M': if (!m_cube) { r = -4; break; }
  if ((n = cb_kb_num(1)) <= 0 || cb_inp_mord(n)) { r = -3; break; }
  if (*kptr != 0) r = *kptr == ' ' ? -2 : -3; break;
case 'K': if (sz == 1) { i_stop = 0; break; } if (*kptr == ' ') { r = -2; break; }
  if ((i_stop = cb_kb_num(1)) < 1 || i_stop > max_moves) { r = -3; break; }
  if (*kptr != 0) r = *kptr == ' ' ? -2 : -3; break;
case 'O': if ((n = fl_offset(cv, "NF")) < 0) { r = -3; break; }
  if (sz == 2) i_wil = n; else r = *(kptr + 1) == ' ' ? -2 : -3; break;
case 'N': if ((n = fl_offset(cv, "SR36")) < 0) { r = -3; break; }
  if (!m_cube && n >= 2) { r = -4; break; }
  if (sz == 2) i_ran = n; else r = *(kptr + 1) == ' ' ? -2 : -3; break;
case 'F': case 'T': i_goal = 0; if (cv == 'G') { i_goal = 1; kptr++; }
  if ((i_mem = cb_kb_num(1)) < 1 || i_mem > 99) { r = -14; break; }
  if (i_goal == 1) i_mem += 100;
  if (*kptr != 0) r = *kptr == ' ' ? -2 : -14; break;
case 'C': case 'G': kc = 0;
  if (cv == cn)
    {
    kc = 1; if (sz == 2) { i_mem = 100; break; }
    if (*++kptr == '0')
      { if (sz == 3) i_mem = 0; else r = *(kptr + 1) == ' ' ? -2 : -3; break; }
    }
  if ((i_mem = cb_kb_num(1)) < 1 || i_mem > 99) { r = -14; break; }
  if (kc) i_mem += 100; if (*kptr != 0) r = *kptr == ' ' ? -2 : -14; break;
case 'L': case '*': break;
case 'Q': if (sz == 1) { *i_lstr = 0; break; } n = fl_offset(cv, "0C");
  if (sz == 2)
    { if (n >= 0) { *i_lstr = n + 1; *(i_lstr + 1) = 0; } else r = -3; break; }
  r = cv == ' ' || n >= 0 && *(kptr + 1) == ' ' ? -2 : -3; break;
case '+': case '-': if ((n = fl_offset(cv, "MDJUARWK")) < 0)
    { r = cn == '-' && dig(cv) ? 1 : -8; break; } // r=1 if it could be a move
  i_opt = cv; cv = *++kptr;
  if (cn == '-' || n >= 2) i_optn = cn == '-' ? 0 : 1;
    else if (i_opt == 'M')
        { if ((i_optn = cb_kb_num(1)) < 1 || i_optn > 3) { r = -8; break; } }
      else if (!dig(cv)) i_optn = 1; // i_opt='d'
        else if ((i_optn = cb_kb_num(1) + 1) < 2 || i_optn > 3) { r = -8; break; }
  if (*kptr != 0) r = *kptr == ' ' ? -2 : -8; break;
case '^': if ((i_loc = cb_kb_num(1)) <= 0 || i_loc > 999) { r = -5; break; }
  if (*kptr == '=')
    {
    kptr++; if ((i_col = cb_kb_num(1)) <= 0 || i_col > 999) { r = -6; break; }
    if (*kptr != 0) { r = *kptr == ' ' ? -2 : -6; break; }
    }
   else if (*kptr == 0) i_col = i_loc;
     else { r = *kptr == ' ' ? -2 : -5; break; }
  if ((r = cb_set_piece(0)) > 0) r = r == 1 ? -5 : r == 2 ? -6 : -7; break;
case '/': case '\\': if ((i_loc = cb_kb_num(1)) <= 9 || i_loc > 999
    || cb_flip_piece(0)) { r = -5; break; }
  if (*kptr != 0) r = *kptr == ' ' ? -2 : -5; break;
default: r = 1; break; // must be moves or turns
}
if (r <= 0) goto s8; r = 0; // switch ends with r=1 if default or if "-" & digit

// moves/turns
if (!(dig(cn) || fl_offset(cn, "-([") >= 0)) { r = -3; goto s8; }
i_rnum = kb = 0; kptr = kbuf; // i_rnum = repeat count, or 0 if no '['
if (cn == '[')
{
kptr++; while (*++kptr == ' ');
while (1)
  {
  if ((cv = *kptr) == ' ' || cv == ']') { kptr--; kb = 1; break; }
  if (cv == 0) { r = -13; goto s8; }
  if (!dig(cv) || i_rnum > 99999999) { r = -11; goto s8; }
  i_rnum = i_rnum * 10 + cv - '0'; kptr++;
  }
if (i_rnum == 0) { r = -11; goto s8; }
}

// moves/turns are read thru state machine logic, state s is the parsing point
s = n = kt = kr = i_mnum = 0; // i_mnum = number of moves/turns entered
while (1)
{
ch = *++kptr;
switch (s)
  {
  case 0: if (ch == ' ') continue; if (ch == 0 || ch == ']') break;
    if (dig(ch)) { n = ch - '0'; s = 1; continue; }
    if (ch == '(') { kt = 1; s = 3; continue; }
    if (ch == '-') { kr = 1; s = 5; continue; } r = -2; goto s78;
  case 1: if (dig(ch)) { if (n == 0) { n = ch - '0'; continue; } r = -9; goto s78; }
    if (ch == ' ' || ch == '(' || ch == 0 || ch == ']') break; r = -9; goto s78;
  case 2: if (dig(ch)) { if (n == 0) { n = ch - '0'; continue; } r = -10; goto s78; }
    if (ch == ' ') { s = 4; continue; } if (ch == ')') break; r = -10; goto s78;
  case 3: if (ch == ' ') continue; if (dig(ch)) { n = ch - '0'; s = 2; continue; }
    if (ch == '-') { kr = 1; s = 6; continue; } r = -10; goto s78;
  case 4: if (ch == ' ') continue; if (ch == ')') break; r = -10; goto s78;
  case 5: case 6: if (dig(ch)) { n = ch - '0'; s -= 4; continue; } r = -2; goto s78;
  case 7: if (ch == ' ') continue; if (ch == 0) break; r = -2; goto s78;
  }
if (s != 0 && s != 7) // add the move or turn to list
  {
  if (i_mnum == max_imov) { r = -12; goto s78; }
  if (n == 0 && kr) { r = -2; goto s78; }
  if (n > 6) { r = kt ? -10 : -9; goto s78; }
  if (m_pyramid && n > 4) { r = -4; goto s78; }
  i_mov[i_mnum++] = n | kt << 7 | kr << 6; // bits 0-5 = move/turn, bits 6 7 = kr kt
  }
if (ch == 0) break;
if (ch == '(') { n = kr = 0; kt = 1; s = 3; continue; }
if (ch == ']') { if (kb) { kb = 0; s = 7; continue; } r = -2; goto s78; }
s = n = kt = kr = 0;
}
if (kb) { r = -13; goto s8; }
cn = 1;

s78: // set ekptr to display characters preceding error
ekptr = kptr; if (ch == ' ' || ch == 0) ekptr--;
s8: cb_disp_wrd(kbuf + 1); if (r < 0) cn = r;
s9: return cn;
// return 0 if empty buffer, or 1 if moves/turns, or operation character,
// or error code: -1 incomplete operation, -2 invalid data,
// -3 invalid operation, -4 invalid for current model, -5 invalid location,
// -6 invalid colors, -7 invalid 9's, -8 invalid run option, -9 invalid move,
// -10 invalid turn, -11 invalid repeat count, -12 too many moves/turns,
// -13 missing "]", -14 invalid memory
}

/********************/
cb_kb_num(kt) short kt;
{
short n;  char ch, kn;

// read unsigned integer from keyboard buffer starting at kptr, advance kptr
// kt=0 if number must end at 0 (end buffer), kt=1 if may end at any non-digit

for (n = kn = 0; ch = *kptr, dig(ch) && n <= 2999; kptr++)
  { n = 10 * n + ch - '0'; kn = 1; }
if (dig(ch) || !kt && ch > 0) n = -1; else if (!kn) n = -2;
return n; // return 0 - 29999, or -1 if greater or invalid, or -2 if missing
}

/********************/
cb_rec_add(n) short n;
{
// add move or step type to rmove[] (see cb_wide_disp)

if (recn > 0 && n == 0) recseq++; // number of sequences
if (recn < max_rec) rmove[recn++] = n; else recseq = 99; // too many moves
reclast = recn - 1;
}

/********************/
cb_rec_line(ln) short ln;
{
if (recseq > 1) { fl_num(ln); fl_ch('.'); }
}

/********************/
cb_rev_disp()
{
short n, ln;

fl_disp("Reverse moves"); if (recseq == 99) fl_disp(" (incomplete)");
// see cb_wide_disp for rmove[] values
for (recn = reclast, ln = 0; recn > 0; recn--)
{
if ((n = rmove[recn]) == 0)
  if (rmove[recn - 1] >= 30); // skip lines with "---"
    else { fl_nl(); cb_rec_line(++ln); fl_sp(); }
 else if (n < 20) { fl_sp(); cb_disp_mr(n); }
}
fl_nl(); fl_nl();
}

/********************/
cb_wide_disp()
{
static char *stp[] = // step titles (codes 0 - 6)
  { "Solution:", "Modify position:", "Anchor:", "Corners:", // 0 1 2 3
    "Edges:", "Flip:", "Goal solution:" };                  // 4 5 6
short n, ln;

// rmove[] values: 0 end sequence, 1-12 move, 20-26 normal step, 30-36 solved step "---"

fl_disp("Widely spaced moves"); if (recseq == 99) fl_disp(" (incomplete)");
for (recn = 0, ln = 0; recn < reclast; recn++)
{
if ((n = rmove[recn]) == 0) { fl_nl(); fl_nl(); cb_rec_line(++ln); continue; }
if (n >= 30) { fl_sp(); fl_disp(stp[n - 30]); fl_disp(" ---"); }
  else if (n >= 20) { fl_sp(); fl_disp(stp[n - 20]); }
    else { fl_spaces(2); cb_disp_m(n); }
}
fl_nl(); fl_nl();
}

/********************/
cb_modpos(n) short n;
{
short i, m;

cb_disp_lno("Modify position: "); cb_rec_add(21);
for (i = 0; i < n; i++)
{
m = cb_ran(0, 0, step == 1 ? 1 : 0); // use 6 sides if stopped in anchor step
x_move(m); fl_sp(); cb_disp_m(m); movetot++; cb_rec_add(m);
}
fl_nl(); cb_rec_add(0);
}

/********************/
cb_set_piece(km) short km;
{
short i, j, k, n, r, nloc, ncol, *ep, *cp, epn, cpn, cps;  char c, cd[3];

// set i_col colors or 9's in location i_loc
// km=0 check i_loc i_col validity, km=1 modify colors (i_loc i_col assumed valid)

// calculate number of digits (i_loc & i_col are 1 - 999)
nloc = i_loc <= 9 ? 1 : i_loc <= 99 ? 2 : 3;
ncol = i_col <= 9 ? 1 : i_col <= 99 ? 2 : 3;
for (i = 0, n = i_col; i < ncol; i++) // separate i_col digits in cd
  { c = n % 10; n /= 10; cd[2] = cd[1]; cd[1] = cd[0]; cd[0] = c; }
r = 0;

for (i = 0; i < x_npcs; i++)
{
if (x_cl[i] != i_loc) continue; if (!km) goto s2;
// order of piece number references in x_cn matches order of i_col digits
for (j = k = 0; j < x_sz; j++)
  if (x_cn[j] == i + 1) { x_dd[j] = cd[k++]; if (k == nloc) break; }
goto s9;
}
r = 1; goto s9;

s2: if (nloc != ncol) { r = 2; goto s9; }
for (i = n = 0; i < ncol; i++) if (cd[i] == 9) n++;
if (!(n == 0 || n == ncol)) { r = 3; goto s9; }

if (mdl == m_ad) { ep = ad_edge; epn = 24; cp = ad_corner; cpn = cps = 24; }
if (mdl == m_fc) { cp = fc_corner; cpn = cps = 24; }
if (mdl == m_py) { ep = py_edge; epn = 12; cp = py_axial; cpn = cps = 12; }
if (m_cube && (i_loc == 136 || i_loc == 145 || i_loc == 235 || i_loc == 246)
  || m_pyramid && (i_loc == 123 || i_loc == 134)) cps = 0;

if (ncol == 1 && mdl == m_ad) { if (i_col == 9 || i_col <= 6) goto s9; goto s6; }
if (ncol == 2 && (mdl == m_ad || mdl == m_py)) { if (i_col == 99) goto s9;
  for (i = 0; i < epn; i ++) if (i_col == ep[i]) goto s9; goto s6; }
if (ncol == 3) { if (i_col == 999) goto s9;
  for (i = cps; i < cps + cpn; i ++) if (i_col == cp[i]) goto s9; goto s6; }
s6: r = 2; // also reaches s6 if the preceding if's are false

s9: return r; // return 0 or error code:
// 1 invalid i_loc, 2 i_col has invalid colors, 3 i_col has invalid 9's
}

/********************/
cb_flip_piece(fl) short fl;
{
short i, j, k, nloc;  char c, c1, c2, c3, cd[3];

// flip piece in location i_loc
// fl=0 check i_loc validity, fl=1 or 2 flip cw or ccw (i_loc assumed valid)

nloc = i_loc <= 99 ? 2 : 3; // calculate number of digits (i_loc is 10 - 999)

// order of piece number references in x_cn matches order of i_col digits
for (i = 0; i < x_npcs; i++) if (x_cl[i] == i_loc)
{
for (j = k = 0; j < x_sz; j++)
  if (x_cn[j] == i + 1) { cd[k++] = j; if (k == nloc) break; }
if (fl > 0) goto s4; goto s9;
}
goto s91;

s4: c1 = cd[0]; c2 = cd[1]; c3 = cd[2];
if (nloc == 2) { c = x_dd[c1]; x_dd[c1] = x_dd[c2]; x_dd[c2] = c; } // flip edge piece
 else // nloc = 3, flip corner/axial piece cw or ccw
{
if (m_cube && (i_loc == 135 || i_loc == 146 || i_loc == 236 || i_loc == 245)
  || m_pyramid && (i_loc == 123 || i_loc == 134)) fl = 3 - fl; // reverse cw/ccw
for (i = 0; i < fl; i++)
  { c = x_dd[c3]; x_dd[c3] = x_dd[c2]; x_dd[c2] = x_dd[c1]; x_dd[c1] = c; }
}

s9: return 0; s91: return 1; // return 0, or 1 if invalid i_loc
}

/********************/
cb_disp_init()
{
static char ti[] = "Initial";

cb_disp_list(1, ti, x_dd);
cb_disp_rows(0, ti, x_dd);
cb_disp_pieces(0, ti, x_dd);
cb_disp_schem(0, ti, x_dd);
}

/********************/
cb_disp_start()
{
cb_disp_list(2, "Start", x_dd);
}

/********************/
cb_disp_end()
{
cb_disp_list(3, "End", x_dd);
}

/********************/
cb_disp_goal(gg) char *gg;
{
cb_disp_list(4, "Goal", gg);
}

/********************/
cb_disp_t(t, st, as, ad) char *t, *st; short *as, *ad;
{
short s, d;

if ((s = fl_lth(t)) > sz_dstr) s = sz_dstr; fl_move(t, st, s);
if (s > 0) st[s++] = ':'; st[s] = 0;
if (s == 0) s = d = 1; else { s = sz_dstr + 2 - s; d = sz_dstr + 2; }
*as = s; *ad = d;
}

/********************/
cb_disp_list(v, t, dd) short v; char *t, *dd;
{
short i, j, s, d;  char *cp, ks, st[sz_dstr+2];

// display title and list of colors: for any d option if v=0, always if v=9,
// depending on d option if v=1/2/3/4 (str="Initial"/"Start"/"End"/"Goal")

if (!(disp == 1 && v == 1 || disp == 2 && (v == 1 || v == 3) || disp == 3
  || disp && v == 0 || v == 9)) goto s9;
ks = v >= 1 && v <= 4;
cb_disp_t(t, st, &s, &d);

if (v == 1 || v == 3) x_copy(dd, w_pp); // avoid identical consecutive displays
if (v == 2 && fl_ncomp(dd, w_pp, x_sz) == 0) goto s9;

if (mdl == m_ad && ks && k_dnl) fl_nl(); // insert blank line between consecutive displays
if (!ks) fl_nl(); fl_disp(st); fl_spaces(s);

if (mdl == m_ad) for (i = 0; i < 6; i++)
{
if (i > 0 && i != 3) fl_spaces(2);
for (j = 0; j < 9; j++) { fl_sp(); fl_num(dd[9 * i + j]); }
if (i == 2) { fl_nl(); fl_spaces(d); }
}

if (mdl == m_fc) for (i = 0; i < 6; i++)
{
if (i > 0) fl_spaces(2);
for (j = 0; j < 4; j++) { fl_sp(); fl_num(dd[4 * i + j]); }
}

if (mdl == m_py) for (i = 0; i < 4; i++)
{
if (i > 0) fl_spaces(2);
for (j = 0; j < 6; j++) { fl_sp(); fl_num(dd[6 * i + j]); }
}

fl_nl(); if (!ks) fl_nl(); k_dnl = 1;

s9: ;
}

/********************/
cb_disp_rows(v, t, dd) short v; char *t, *dd;
{
static char ad_sa[] = // spacing codes and offsets in ad array
  { 108,45,46,47,100,
    108,48,49,50,100,
    108,51,52,53,100,100,
    101, 9,10,11,101,18,19,20,101,0,1,2,101,27,28,29,100,
    101,12,13,14,101,21,22,23,101,3,4,5,101,30,31,32,100,
    101,15,16,17,101,24,25,26,101,6,7,8,101,33,34,35,100,100,
    108,36,37,38,100,
    108,39,40,41,100,
    108,42,43,44,100,99 };
static char fc_sa[] = // spacing codes and offsets in fc array
  { 106,20,21,100,
    106,22,23,100,100,
    101,4,5,101, 8, 9,101,0,1,101,12,13,100,
    101,6,7,101,10,11,101,2,3,101,14,15,100,100,
    106,16,17,100,
    106,18,19,100,99 };
static char py_sa[] = // spacing codes and offsets in py array
  { 101,0,1,2,101,6, 7, 8,101,12,13,14,102,18,19,20,100,
    101,3,4,5,101,9,10,11,101,15,16,17,102,21,22,23,100,99 };
char *cp, ca;

// display title and rows of colors: for run option j if v=0, always if v=9

if (!(k_rows && v == 0 || v == 9)) goto s9;
if (fl_lth(t) > 0) { fl_disp(t); fl_ch(':'); } fl_nl();

if (mdl == m_ad) cp = ad_sa;
if (mdl == m_fc) cp = fc_sa;
if (mdl == m_py) cp = py_sa;

while (1)
{
if ((ca = *cp++) == 99) break; if (ca < 100) { fl_sp(); fl_num(dd[ca]); continue; }
if (ca > 100) fl_spaces(ca - 100); else fl_nl();
}
fl_nl();

s9: ;
}

/********************/
cb_disp_pieces(v, t, dd) short v; char *t, *dd;
{
short i, j, k, s, d, nd, np, loc, col;  char st[sz_dstr+2];

// display title and list of pieces: for run option u if v=0, always if v=9

if (!(k_pcs && v == 0 || v == 9)) goto s9;
cb_disp_t(t, st, &s, &d); fl_nl(); fl_disp(st);

for (k = 3; k > 0; k--)
{
fl_spaces(k == 3 ? s : d);
for (i = np = 0; i < x_npcs; i++)
  {
  loc = x_cl[x_cp[i] - 1]; nd = loc == 0 ? 0 : loc <= 9 ? 1 : loc <= 99 ? 2 : 3;
  if (nd < k) continue; if (nd > k) break;
  for (j = col = 0; j < x_sz; j++) if (x_cn[j] == x_cp[i]) col = 10 * col + dd[j];
  fl_sp(); np++; fl_ch('^'); fl_num(loc); fl_ch('='); fl_num(col);
  if (mdl == m_ad && (k == 2 && (np == 4 || np == 8) || k == 3 && np == 4)
    || mdl == m_fc && k == 3 && np == 4) { fl_nl(); fl_spaces(d); }
  }
if (np > 0) fl_nl();
}
fl_nl();

s9: ;
}

/********************/
cb_disp_schem(v, t, dd) short v; char *t, *dd;
{
static char ad_cv[] = // 10 rows: schematic data and offsets in ad array
  { 14,14, 0, 3, 3,  45,46,47,
    14,14, 0, 3, 3,  48,49,50,
    14,14, 0, 3, 3,  51,52,53,
     2, 2,36,12,12,   9,10,11,18,19,20,0,1,2,27,28,29,
     2, 2, 0,12,12,  12,13,14,21,22,23,3,4,5,30,31,32,
     2, 2, 0,12,12,  15,16,17,24,25,26,6,7,8,33,34,35,
    14, 2,36, 3,12,  36,37,38,
    14,14, 0, 3, 3,  39,40,41,
    14,14, 0, 3, 3,  42,43,44,
    14,14, 0, 3, 3 };
static char fc_cv[] = // 7 rows: schematic data and offsets in fc array
  { 10,10, 0,2,2,  20,21,
    10,10, 0,2,2,  22,23,
     2, 2,24,8,8,   4, 5, 8, 9,0,1,12,13,
     2, 2, 0,8,8,   6, 7,10,11,2,3,14,15,
    10, 2,24,2,8,  16,17,
    10,10, 0,2,2,  18,19,
    10,10, 0,2,2 };
static char py_cv[] = // 3 rows: schematic data and offsets in py array
  { 2,12,12,  0,1,2,6, 7, 8,12,13,14,18,19,20,
    2, 0,12,  3,4,5,9,10,11,15,16,17,21,22,23,
    2,12,12 };
short i, k, num_ln, sz_ln, nsp, nspp, nsep, nfa, nfaa;
char ln[55], *pln, *pcv, c;

// display title and schematic: for run option a if v=0, always if v=9

if (!(k_schem && v == 0 || v == 9)) goto s9;

if (mdl == m_ad) { pcv = ad_cv; num_ln = 10; sz_ln = 52; }
if (mdl == m_fc) { pcv = fc_cv; num_ln = 7; sz_ln = 36; }
if (mdl == m_py) { pcv = py_cv; num_ln = 3; sz_ln = 55; }

if (fl_lth(t) > 0) { fl_disp(t); fl_ch(':'); } fl_nl();

if (m_cube) for (k = 0; k < num_ln; k++)
{
nsp = *pcv++; nspp = *pcv++; nsep = *pcv++; nfa = *pcv++; nfaa = *pcv++;
fl_fill(ln, ' ', sz_ln);
for (i = 0, pln = ln + nspp; i < nfaa; i++) { fl_copy("+---", pln); pln += 4; }
if (nsep) ln[nspp + nsep] = '|';
fl_copy("+", pln); fl_disp(ln); fl_nl(); if (k == num_ln - 1) continue;
fl_fill(ln, ' ', sz_ln);
for (i = 0, pln = ln + nsp; i < nfa; i++) { c = dd[*pcv++];
  *pln = '|'; pln += 2; *pln = c <= 6 ? '0' + c : '-'; pln += 2; }
fl_copy("|", pln); fl_disp(ln); fl_nl();
}

if (m_pyramid) for (k = 0; k < num_ln; k++)
{
nsp = *pcv++; nsep = *pcv++; nfa = *pcv++;
fl_fill(ln, ' ', sz_ln);
for (i = 0, pln = ln + nsp; i < nfa; i++) { fl_copy("+---", pln); pln += 4; }
if (nsep) ln[nsp + nsep] = ln[nsp + nsep * 2] = '|';
fl_copy("+", pln); fl_move(ln + 38, ln + 41, 14); ln[39] = ln[40] = ' ';
fl_disp(ln); fl_nl(); if (k == num_ln - 1) continue;
fl_fill(ln, ' ', sz_ln);
for (i = 0, pln = ln + nsp; i < nfa; i++) { c = dd[*pcv++];
  *pln = '|'; pln += 2; *pln = c <= 4 ? '0' + c : '-'; pln += 2; }
fl_copy("|", pln); fl_move(ln + 38, ln + 41, 14); ln[39] = ln[40] = ' ';
fl_disp(ln); fl_nl();
}

fl_nl();

s9: ;
}

/********************/
cb_disp_lno(str) char *str;
{
if (lno >= 0) { fl_num(++lno); fl_disp(". "); } fl_disp(str);
}

/********************/
cb_disp_moves()
{
short i, m;  char kp;

// display the moves sequence of the current solution step

for (i = 0; i < im; i++)
  { if (i > 0) fl_sp(); m = move[i]; cb_disp_m(m); movetot++; cb_rec_add(m); }

if (step == 9) for (i = 0; i < num_pairs; i++)
 if (ad_epair[i][0] > 0)
{
fl_disp(i == 0 ? "  >> " : ", "); if ((kp = ad_epair[i][2])) fl_ch('(');
fl_num(ad_epair[i][0]); fl_sp(); fl_num(ad_epair[i][1]); if (kp) fl_ch(')');
}

fl_nl(); cb_rec_add(0);
}

/********************/
cb_disp_m(m) short m;
{
// display move (internal value)

if (m_cube)
  if (m <= 6) if (m & 1) fl_num(m); else { fl_ch('-'); fl_num(m - 1); }
    else { m -= 6; if (m & 1) fl_num(m + 1); else { fl_ch('-'); fl_num(m); } }
if (m_pyramid) if (m <= 4) fl_num(m); else { fl_ch('-'); fl_num(m - 4); }
}

/********************/
cb_disp_mr(m) short m;
{
// display reverse move (internal value)
if (m_cube) cb_disp_m(crev[m]); if (m_pyramid) cb_disp_m(yrev[m]);
}

/********************/
cb_msg_goal()
{
short i;  char kp;

// goal position with 9's cannot be fully checked, so message is different
for (i = kp = 0; i < x_sz; i++) if (x_gg[i] == 9) { kp = 1; break; }
fl_disp(kp ? "Goal appears OK" : "Goal OK"); fl_nl();
}

/********************/
cb_msg_reset(kg) short kg;
{
fl_disp(!kg ? "Reset position" : "Reset goal position"); fl_nl();
}

/********************/
cb_msg_noreset(kg) short kg;
{
fl_disp(!kg ? (m_cube ? "Invalid cube" : "Invalid pyramid") : "Invalid goal");
fl_disp(", cannot reset position"); fl_nl();
}

/********************/
cb_msg_chk(r) short r;
{
static char *msg[] = // position check error messages
  { "", "colors wrong", "edges wrong", "edges flipped wrongly",    // 1 2 3
    "corners wrong", "corners flipped wrongly", "axials wrong",    // 4 5 6
    "centers wrong", "edges have invalid 9's",                     // 7 8
    "corners have invalid 9's", "axials have invalid 9's",         // 9 10
    "edges duplicated", "corners duplicated", "axials duplicated", // 11 12 13
    "pieces swapped wrongly", "edges swapped wrongly",             // 14 15
    "axials swapped wrongly" };                                    // 16

fl_disp(", some "); fl_disp(msg[r]);
}

/********************/
cb_msg_mt(m, mt) short m, mt;
{
fl_nl(); fl_disp(mt ? "** Invalid turn: " : "** Invalid move: "); fl_num(m);
cb_exit();
}

/********************/
cb_col_f2(c1, c2) short c1, c2;
{
short c;

if (c1 > c2) { c = c1; c1 = c2; c2 = c; }
return 10 * c1 + c2;
}

/********************/
cb_col_f3(c1, c2, c3) short c1, c2, c3;
{
short c;

if (c1 > c2) { c = c1; c1 = c2; c2 = c; }
if (c2 > c3) { c = c2; c2 = c3; c3 = c; }
if (c1 > c2) { c = c1; c1 = c2; c2 = c; }
return 100 * c1 + 10 * c2 + c3;
}

/********************/
cb_scramble(scr) short scr;
{
short n;  char temp[sz_max];

// skip a variable number of cb_frand calls, to prevent series of identical
// random numbers and hence series of identical positions in different runs
n = ds_tvar(); while (n--) cb_frand(1);

if (scr == 0)
{
x_copy(x_oo, x_dd); n = cb_frand(1) * 20 + 190; // n=190-210 (210 very rarely)
while (n--) { x_turn(cb_ran(1, 1, 0)); x_move(cb_ran(1, 0, 1)); }
x_reset(0, 0, 0); fl_disp("Scramble"); fl_nl();
}

if (scr == 1)
{
x_copy(x_gg, temp); x_copy(x_nn, x_gg); x_gen(); x_copy(temp, x_gg);
fl_disp("Scramble pieces"); fl_nl();
}

if (scr == 2) { x_gen(); fl_disp("Scramble template"); fl_nl(); }
}

/********************/
cb_gen_pos(tp, aen, acn) char tp[]; short *aen, *acn;
{
short i, j, k, n, tn, c, c1, c2, c3, nd, ep, cp, mcol, col[27];
char *pp, cd[3], ord[27];

// generate random unverified position from goal position

if (mdl == m_ad) { ep = 12; cp = 8; }
if (mdl == m_fc) { ep = 0; cp = 8; }
if (mdl == m_py) { ep = 6; cp = 4; }

for (i = 0; i < x_npcs; i++) ord[i] = i;
for (i = 0; i < 200; i++) // scramble piece reading order
{
if ((j = cb_frand(1) * x_npcs) > x_npcs - 1) j = x_npcs - 1;
if ((k = cb_frand(1) * x_npcs) > x_npcs - 1) k = x_npcs - 1;
n = ord[j]; ord[j] = ord[k]; ord[k] = n;
}

x_copy(x_gg, x_dd); fl_clr(col, x_npcs*2); fl_clr(tp, x_npcs); *aen = *acn = 0;

// set actual and 99/999 pieces in tables col & tp (only first x_npc elements are used)
if (mdl == m_ad || mdl == m_py)
for (i = 0; i < ep; i++) // edge pieces
{
pp = x_ef + 2 * i; c = cb_col_f2(x_dd[pp[0]], x_dd[pp[1]]);
col[j = x_cn[pp[0]] - 1] = c; if (c == 99) { tp[j] = 2; (*aen)++; }
}
for (i = 0; i < cp; i++) // corner/axial pieces
{
pp = x_cf + 3 * i; c = cb_col_f3(x_dd[pp[0]], x_dd[pp[1]], x_dd[pp[2]]);
col[j = x_cn[pp[0]] - 1] = c; if (c == 999) { tp[j] = 3; (*acn)++; }
}

// replace the 99/999 pieces randomly with the missing colors
for (i = tn = 0; i < x_npcs && tn < *aen + *acn; i++)
{
if ((mcol = x_cl[ord[i]]) <= 9) continue; // skip center pieces (ad)
for (j = 0; j < x_npcs; j++) if (mcol == col[j]) goto s6; // skip actual pieces
nd = mcol <= 99 ? 2 : 3;
// flip piece colors randomly
for (j = 0, n = mcol; j < nd; j++)
  { c = n % 10; n /= 10; c3 = c2; c2 = c1; c1 = c; }
if ((k = cb_frand(1) * 2) > 0) { c = c1; c1 = c2; c2 = c; }
if (nd == 2) n = 10 * c1 + c2;
 else
  {
  if ((k = cb_frand(1) * 3) > 2) k = 2;
  for (j = 0; j < k; j++) { c = c1; c1 = c2; c2 = c3; c3 = c; }
  n = 100 * c1 + 10 * c2 + c3;
  }
for (j = 0; ; j++) // set the piece colors in first available 99/999 location
  if (nd == 2 && col[j] == 99 || nd == 3 && col[j] == 999)
    { col[j] = n; tn++; break; }
s6: ;
}

for (i = 0; i < x_npcs; i++) if ((nd = tp[i]) > 0) // update position array
{
for (j = 0, n = col[i]; j < nd; j++)
  { c = n % 10; n /= 10; cd[2] = cd[1]; cd[1] = cd[0]; cd[0] = c; }
// order of piece number references in x_cn matches order of color digits
for (j = k = 0; j < x_sz; j++)
  if (x_cn[j] == i + 1) { x_dd[j] = cd[k++]; if (k == nd) break; }
}
}

/********************/
cb_bk_comp()
{
short i, j, tot, c, p1, p2, p3;  char *cp, kf;

// check if current back sequence creates the required wrong pieces

/* the asm below
char *cn, *dd, *pp;
tot = 0; fl_clr(bkpcs, x_npcs); // only first x_npc bytes are used
for (i = 0, cn = x_cn, dd = x_dd, pp = w_pp; i < x_sz; i++, cn++, dd++, pp++)
// a piece may be wrong in 2 or 3 faces, but increment tot & bkpcs only once
  if (*dd != *pp) if (bkpcs[c = *cn - 1] == 0)
    { if (++tot > bkwrong) goto s9; bkpcs[c]++; }
*/
_asm
#if win32
{                                  ; no need to preserve ebx esi edi thru this asm
        mov   edi, offset bkpcs
        mov   ecx, 7               ; clear bkpcs 28 bytes for all models,
        xor   eax, eax             ;   but only first x_npc bytes are used
        mov   es, word ptr s_ds
        cld
    rep stosd
        mov   esi, x_dd
        mov   edi, w_pp
        mov   edx, x_cn
        movzx ecx, word ptr x_sz   ; es, cld, ax from previous loop (ax=0)
bkc_l3: cmpsb
        je    bkc_l6
        movzx ebx, byte ptr [edx]  ; piece number (1 - x_npc)
        lea   ebx, bkpcs[ebx-1]
        cmp   byte ptr [ebx], 0    ; bkpcs element
        jne   bkc_l6               ; already incremented
        inc   ax                   ; tot
        cmp   ax, bkwrong
        jg    s9
        inc   byte ptr [ebx]
bkc_l6: inc   edx
        loop  bkc_l3
        mov   tot, ax
}
#else
{                                 ; no need to preserve bx si di thru this asm
        mov   di, offset bkpcs
        mov   ecx, 7              ; clear bkpcs 28 bytes for all models,
        xor   eax, eax            ;   but only first x_npc bytes are used
        mov   es, word ptr s_ds
        cld
    rep stosd
        mov   si, x_dd
        mov   di, w_pp
        mov   dx, x_cn
        movzx ecx, word ptr x_sz  ; es, cld, ax from previous loop (ax=0)
bkc_l3: cmpsb
        je    bkc_l6
        mov   bx, dx
        movzx bx, byte ptr [bx]   ; piece number (1 - x_npc)
        lea   bx, bkpcs[bx-1]
        cmp   byte ptr [bx], 0    ; bkpcs element
        jne   bkc_l6              ; already incremented
        inc   ax                  ; tot
        cmp   ax, bkwrong
        jg    s9
        inc   byte ptr [bx]
bkc_l6: inc   dx
        loop  bkc_l3
        mov   tot, ax
}
#endif

if (tot < bkwrong) goto s9; // include only tot=bkwrong (tot>bkwrong already excluded)
if (bkwrong == 0) { cb_bk_disp(0); goto s9; } // bkwrong=0: do-nothing sequence

// bkwrong>0
if (bktype != 3 || bkflip != 3)
for (i = 0; i < x_npcs; i++) if (bkpcs[i] != 0) // only edges/corners/axials can be wrong
{
c = x_cl[i]; // c<=99 for edges, c>99 for corners or axials
if (bktype != 3)
  if (c <= 99 && bktype == 2 || c > 99 && bktype == 1) goto s9;
if (bkflip == 3) continue;
if (c <= 99) // check if edge flipped
  {
  cp = x_ef + 2 * x_ci[i]; p1 = x_dd[cp[0]]; p2 = x_dd[cp[1]];
  kf = c == 10 * p2 + p1;
  }
 else // check if corner/axial flipped, either cw or ccw
  {
  cp = x_cf + 3 * x_ci[i]; p1 = x_dd[cp[0]]; p2 = x_dd[cp[1]]; p3 = x_dd[cp[2]];
  kf = c == 100 * p2 + 10 * p3 + p1 || c == 100 * p3 + 10 * p1 + p2;
  }
if (kf && bkflip == 2 || !kf && bkflip == 1) goto s9;
}

cb_bk_disp(1); cb_disp_list(0, "", x_dd);

s9: ;
}

/********************/
cb_bk_disp(kw) short kw;
{
short i, j, k, col;

// display current back sequence and wrong pieces (kw=0 for do-nothing sequence)

bkseqs++; cb_bk_erase();
fl_sp(); for (i = 0; i < im; i++) { fl_sp(); cb_disp_m(move[i]); }

if (kw)
{
fl_disp("   ");
for (i = 0; i < x_npcs; i++) if (bkpcs[(j = x_cp[i]) - 1] > 0)
  {
  for (k = col = 0; k < x_sz; k++) if (x_cn[k] == j) col = 10 * col + x_dd[k];
  fl_sp(); fl_ch('^'); fl_num(x_cl[j - 1]); fl_ch('='); fl_num(col);
  }
}

if (bkrev) { fl_nl(); fl_disp("  Reverse:");
  for (i = im - 1; i >= 0; i--) { fl_sp(); cb_disp_mr(move[i]); } }
fl_nl(); cb_bk_chg(); // cb_bk_chg in case lines displayed too fast for the regular calls
}

/********************/
cb_bk_erase()
{
// erase possible cb_bk_chg display
k_wrf = 0; fl_spaces(sp55); fl_back(sp55); k_wrf = 1;
}

/********************/
cb_bk_chg()
{
#define max_time 35996400
#define shour 3600
#define sminute 60
dbl dt;  long t1, t2;  short n, h1, h2, m1, m2;

// calculate and display changing percent and time elapsed/left

k_wrf = 0;
if ((n = bkcount * 100 / bktot) > 100) n = 100; // bkcount & bktot must be >0
fl_num(n); fl_disp("% "); fl_cnum(bkseqs); fl_disp(" seqs");
t1 = cb_time(1) / 100; // max 247 days (5928 hours)
h1 = t1 / shour; m1 = (t1 % shour) / sminute;
dt = (dbl)t1 * (bktot - bkcount) / bkcount;
t2 = dt < max_time ? dt : max_time; // max 9999 hours
h2 = t2 / shour; m2 = (t2 % shour) / sminute;
if (t1 + t2 > sminute) // display times only if total > 1 minute
{
fl_disp("  elapsed/left H:M ");
fl_num(h1); fl_ch(':'); fl_znum(m1, 2); fl_ch('/');
fl_num(h2); fl_ch(':'); fl_znum(m2, 2);
}

fl_back(sp55); k_wrf = 1;
}

/********************/
long cb_time(r) short r;
{
static long t0;
long dif;

// r=0 initialize timer, else return hundredths of second since initialized
// elapsed time is correct to 247 days, then restarts from 0

if (r == 0) { t0 = ds_time(); tm_hour = ds_hour(); tm_day = dif = 0; }
  else { cb_tmchk(); dif = (long)tm_day * 8640000 + ds_time() - t0; }
return dif;
}

/********************/
cb_tmchk()
{
short hh;

// update elapsed time variables, check if passed midnight
// limit elapsed time to 247 days to allow for cb_time calculations

if ((hh = ds_hour()) < tm_hour) { if (tm_day == 247) tm_day = 0; tm_day++; }
tm_hour = hh;
}

/********************/
dbl cb_frand(rg) short rg;
{
static long p = 2147483647;
long seed, xalo, leftlo, fhi, xhi, k;  short i;  dbl d;

// return next random number, using generator rg 0 or 1
// adapted from DeSmet C library

seed = ran_seed[rg];
xhi = seed >> 16; xalo = (seed - (xhi << 16)) * 16807;
leftlo = xalo >> 16; fhi = xhi * 16807 + leftlo; k = fhi >> 15;
seed = (((xalo - (leftlo << 16)) - p) + ((fhi - (k << 15)) << 16)) + k;
if (seed < 0) seed = -seed; i = seed; if (i < 0) i = -i;
d = (dbl)i / 32767.0; if (d < 0) d = 0; else if (d > 1) d = 1;
ran_seed[rg] = seed;
return d; // return 0 - 1 (exactly 1.000... only every ~32,000 calls on average)
}

/********************/
cb_frand_init()
{
short n;

// initialize the two random number generators

ran_seed[0] = ran_seed[1] = 7613941;
n = ds_tvar() + 500; while (n--) { cb_frand(0); cb_frand(1); }
fl_clr(ran_prev, 2*2*2); fl_clr(sav_prev, 2*2*2);
}

/********************/
cb_ran_save()
{
// save random state of generator 0
sav_seed[0] = ran_seed[0]; fl_move(ran_prev[0], sav_prev[0], 2*2);
}

/********************/
cb_ran_restore()
{
// restore random state of generator 0
ran_seed[0] = sav_seed[0]; fl_move(sav_prev[0], ran_prev[0], 2*2);
}

/********************/
cb_ran(rg, mt, ks) short rg, mt, ks;
{
char m, m1, m2, mp, max;

// return next random move/turn, using generator rg 0/1 for regular/scramble
// mt 0/1 for move/turn, ks 0/1 for 3/6 sides (if cube)
// truncate random value and restrict it to range 0 - max-1
// cb_frand returns 0 - 1, but 1.000... (which makes m == max) very rarely
// skip if reverse of previous number or if same as previous two (one for py),
//   to prevent useless moves/turns when all in a sequence are random

if (m_cube) { max = ks ? 12 : 6;
  m1 = ran_prev[rg][mt][0]; m2 = ran_prev[rg][mt][1]; mp = crev[m1]; }
if (m_pyramid) { max = 8; m1 = m2 = ran_prev[rg][mt][0]; mp = yrev[m1]; }
while (1) { m = cb_frand(rg) * max;
  if (m != max && ++m != mp && !(m1 == m2 && m == m1)) break; }
ran_prev[rg][mt][1] = ran_prev[rg][mt][0]; ran_prev[rg][mt][0] = m;
return m;
}

/********************/
cb_incseq()
{
_asm
{ ; increment quad-word integer seqtot
    add   dword ptr seqtot, 1
    adc   dword ptr seqtot[4], 0
}
}

/********************/
cb_ch_wrd(ch) char ch;
{
k_wrd = 0; fl_ch(ch); fl_nl(); k_wrd = 1;
}

/********************/
cb_disp_wrd(s) char *s;
{
k_wrd = 0; fl_disp(s); fl_nl(); k_wrd = 1;
}

/********************/
cb_num_wrd(n) uns n;
{
k_wrd = 0; fl_num(n); fl_nl(); k_wrd = 1;
}

/********************/

/*** fl functions ***/

/********************/
fl_clr(s, n) char *s; short n;
{
while (n-- > 0) *s++ = 0; // fill n bytes with 0
}

/********************/
fl_fill(s, ch, n) char *s, ch; short n;
{
while (n-- > 0) *s++ = ch; // fill n bytes with ch
}

/********************/
fl_move(s, d, n) char *s, *d; short n;
{
if (n > 0) // copy n bytes (source and destination areas may overlap)
  if (s > d || s + n <= d) while (n--) *d++ = *s++;
    else { s += n - 1; d += n - 1; while (n--) *d-- = *s--; }
}

/********************/
fl_copy(s, d) char *s, *d;
{
while (*d++ = *s++); // copy string including terminating 0 (no overlap)
}

/********************/
fl_ncomp(s, d, n) char *s, *d; short n;
{
// compare n bytes of two strings, return 0 if same, 1 if not
while (n-- > 0) if (*s++ != *d++) return 1; return 0;
}

/********************/
fl_offset(ch, s) char ch, *s;
{
char *cp, c;

// return offset of byte in 0-terminated string, or -1 if not found
for (cp = s; (c = *cp) != 0 && c != ch; cp++);
return c > 0 ? cp - s : -1;
}

/********************/
fl_lth(st) char *st;
{
short s;
for (s = 0; *st++; s++); return s; // length of 0-terminated string
}

/********************/
fl_up(ch) char ch;
{
return ch >= 'a' && ch <= 'z' ? ch - 32 : ch; // return upper case if letter
}

/********************/
fl_low(ch) char ch;
{
return ch >= 'A' && ch <= 'Z' ? ch + 32 : ch; // return lower case if letter
}

/********************/
fl_lowstr(s) char *s;
{
char ch;
while (ch = *s) *s++ = fl_low(ch); // convert 0-terminated string to lower case
}

/********************/
fl_ch(ch) char ch;
{
if (k_wrd) ds_chout(ch); // display ch
if (k_wrf && !k_nowrf) ds_write(wrhdl, 1, &ch); // write ch to file
}

/********************/
fl_disp(s) char *s;
{
char ch;
while (ch = *s++) fl_ch(ch); // display 0-terminated string
}

/********************/
fl_nl()
{
static short count = 0;

// cb_esc needed in case line display rate exceeds the cb_esc check in loops
if (++count > 100) { count = 0; cb_esc(); }

// write end of line, ensure the line is added now to the output file
fl_disp("\r\n"); k_dnl = 0;
if (k_wrf && !k_nowrf) { ds_close(wrhdl); ds_open(wrfile, &wrhdl);
  ofst = 0; ds_lseek(wrhdl, 2, &ofst); }
}

/********************/
fl_sp()
{
fl_ch(' ');
}

/********************/
fl_spaces(n) short n;
{
while (n-- > 0) fl_ch(' '); // n may be <0
}

/********************/
fl_back(n) short n;
{
while (n-- > 0) fl_ch('\b');
}

/********************/
fl_num(n) uns n;
{
uns r;
if (r = n / 10) fl_num(r); fl_ch('0' + n % 10); // display unsigned integer
}

/********************/
fl_znum(n, w) uns n, w;
{
#define szd 6
short i;  char c[szd];

// display unsigned integer with leading 0's in field of w characters (max 6)

if (w < 1) w = 1; if (w > szd) w = szd; i = szd - 1;
while (n) { c[i--] = '0' + (n % 10); n /= 10; } while (i >= 0) c[i--] = '0';
for (i = szd - w; i < szd; i++) fl_ch(c[i]);
}

/********************/
fl_lnum(n) long n;
{
long r;

// display positive long
if (n < 0) n = 0; if (r = n / 10) fl_lnum(r); fl_ch('0' + (char)(n % 10));
}

/********************/
fl_cnum(n) long n;
{
#define szd 13
short i, j;  char c[szd];

// display positive long with commas

if (n < 0) n = 0; i = szd - 1; j = 0;
while (n) { if (j++ == 3) { c[i--] = ','; j = 1; }
  c[i--] = '0' + (char)(n % 10); n /= 10; }
if (i == szd - 1) c[i--] = '0'; while (++i < 13) fl_ch(c[i]);
}

/********************/
fl_lns(n, str) long n; char *str;
{
fl_cnum(n); fl_disp(str); if (n != 1) fl_ch('s');
}

/********************/
fl_qnum(pq) char *pq;
{
#define szd 25
long q1, q2;  short i, j, k;  char c[szd];

// display quad-word positive integer up to 18 digits with commas

q1 = 1000000000; // divisor for 9-digit remainder
q2 = 400000000; // ensures q2 after div is < 2^31 (max displayed is ~1.7 10^18)

_asm
{                      ; no need to preserve esi thru this asm
#if win32
    mov   esi, pq
    mov   eax, [esi]
    mov   edx, [esi+4]
#else
    mov   si, pq
    mov   eax, [si]
    mov   edx, [si+4]
#endif
    cmp   edx, q2
    ja    s9           ; too large, display nothing
    div   dword ptr q1
    mov   q2, eax      ; 9 lowest digits in q1, rest in q2
    mov   q1, edx
}

i = szd - 1; j = 0; k = 9;
// if q2>0 display 9 digits for q1 even if it has fewer
while (q2 == 0 ? q1 : k--) { if (j++ == 3) { c[i--] = ','; j = 1; }
  c[i--] = '0' + (char)(q1 % 10); q1 /= 10; }
while (q2) { if (j++ == 3) { c[i--] = ','; j = 1; }
  c[i--] = '0' + (char)(q2 % 10); q2 /= 10; }
if (i == szd - 1) c[i--] = '0'; while (++i < szd) fl_ch(c[i]);

s9: ;
}

/********************/
fl_getnum(cp) char *cp;
{
short n;  char ch, kn;

// read 0-terminated unsigned integer string starting at cp

for (n = kn = 0; (ch = *cp) >= '0' && ch <= '9' && n <= 2999; cp++)
  { n = 10 * n + ch - '0'; kn = 1; }
if (ch > 0) n = -1; else if (!kn) n = -2;
return n; // return 0 - 29999, or -1 if greater or invalid, or -2 if missing
}

/********************/
fl_rdch()
{
// read next character in input file
// return character, or 0 if error, or -1 if eof

if (--rdsz < 0)
  if ((rdsz = ds_read(rdhdl, sz_rdbuf, rdbuf)) <= 0) return rdsz;
    else { rdptr = rdbuf; rdsz--; }
return *rdptr++;
}

/********************/
fl_nextch()
{
short r;

// read next character in input file without advancing pointer
// return character, or 0 if error, or -1 if eof

if (rdsz <= 0)
  if ((r = ds_read(rdhdl, 1, rdbuf + sz_rdbuf - 1)) <= 0) return r;
    else { rdptr = rdbuf + sz_rdbuf - 1; rdsz = 1; }
return *rdptr;
}

/********************/
fl_open_inp()
{
if (ds_open(rdfile, &rdhdl))
  { ds_strout("Cannot open "); ds_strout(rdfile); cb_exit(); }
k_inp = 1; rdsz = 0; *(i_ept = i_estr) = 0; k_eol = 1;
}

/********************/
fl_close_inp()
{
if (k_inp) ds_close(rdhdl); k_inp = 0;
}

/********************/
fl_init_inp()
{
// reinitialize input file
ofst = 0; ds_lseek(rdhdl, 0, &ofst);
rdsz = 0; *(i_ept = i_estr) = 0; k_eol = 1;
}

/********************/
fl_open_out(kn) short kn;
{
// if kn=0 and output file found, open it at eof, else create new one

if (kn || ds_open(wrfile, &wrhdl))
  if (ds_create(wrfile, &wrhdl))
    { ds_strout("Cannot create "); ds_strout(wrfile); goto s91; }
   else { ds_close(wrhdl); ds_open(wrfile, &wrhdl); } // just to set the flags in ds_open
 else { ofst = 0; ds_lseek(wrhdl, 2, &ofst);
   if (!k_nowrf) ds_write(wrhdl, 2, "\r\n"); }
k_out = 1;

return 0; s91: return 1; // return 0 if opened successfully, 1 if not
}

#if win32

/********************/
fl_kbstr()
{
// accept string from keyboard (terminated with Enter)

// _cgets uses Windows console API, kbuf[0] limit is 127 bytes
// returned string starts at kbuf+2, length in kbuf[1] (excluding terminating 0)
// shift it to start at kbuf+1, length in kbuf[0]
kbuf[0] = sz_kbuf - 3; _cgets(kbuf);
fl_move(kbuf + 1, kbuf, (uns)*(kbuf + 1) + 2);

// replace tabs with spaces, delete leading & trailing spaces
for (kptr = kbuf + 1; *kptr; kptr++) if (*kptr == '\t') *kptr = ' ';
for (kptr = kbuf + 1; *kptr == ' '; kptr++) (*kbuf)--;
fl_move(kptr, kbuf + 1, *kbuf + 1);
for (kptr = kbuf + *kbuf; *kptr == ' '; kptr--) { (*kbuf)--; *kptr = 0; }
}

#else

/********************/
fl_kbstr()
{
#define sz_obuf 800
#define key_up 137
#define key_dn 138
static short onum = 0;
static char obuf[sz_obuf] = { 0 }; // only first byte must be 0
short n, on, sz;  char *lp, ch;

// accept string from keyboard (terminated with Enter)
// replaces 16-bit _cgets, which does not store/recall strings

*kbuf = 0; kptr = kbuf + 1; on = 0;
while (1)
{
while (1)
  {
  while (1)
    {
    if (!((ch = ds_chrd()) == 0 || ch == 0xe0)) break; ch = ds_chrd();
    if (ch == 72) { ch = key_up; break; }
    if (ch == 80) { ch = key_dn; break; }
    // ignore other extended codes, wait for next character
    }
  if (ch >= ' ' && ch <= '~' || ch =='\t' || ch =='\r' || ch == '\b'
    || ch == key_up || ch == key_dn) break;
  }

if (ch == '\r') { ds_strout("\r\n"); *kptr = 0; break; }
if (ch == '\b') { fl_kbclr(1); continue; }
if (ch == '\t') ch = ' ';
if (kptr == kbuf + sz_kbuf - 1) continue;
if (ch == key_up || ch == key_dn) // get string from recall buffer
  {
  if (ch == key_up) { if (onum == 0 || on == onum) continue; on++; }
    else if (on == 0 || --on == 0) { fl_kbclr(0); continue; }
  fl_kbclr(0); for (lp = obuf, n = on; --n > 0; lp += *lp);
  fl_move(lp, kbuf, *lp); *(kbuf + *kbuf) = 0; (*kbuf)--;
  ds_strout(kbuf + 1); kptr += *kbuf; continue;
  }
ds_chout(ch); *kptr++ = ch; (*kbuf)++;
}
// string starts at kbuf+1, 0-terminated, length in kbuf[0] (excluding the 0)

// delete leading & trailing spaces
for (kptr = kbuf + 1; *kptr == ' '; kptr++) (*kbuf)--;
fl_move(kptr, kbuf + 1, *kbuf + 1);
for (kptr = kbuf + *kbuf; *kptr == ' '; kptr--) { (*kbuf)--; *kptr = 0; }

// add string to recall buffer (no consecutive duplicates)
if ((sz = *kbuf) > 0
  && (*obuf == 0 || sz != *obuf - 1 || fl_ncomp(kbuf + 1, obuf + 1, sz) != 0))
{
fl_move(obuf, obuf + 1 + sz, sz_obuf - 1 - sz);
fl_move(kbuf, obuf, ++sz); *obuf = sz;
// count strings in buffer (ignore oldest string if incomplete)
for (onum = 0, lp = obuf;
  (sz = *lp) != 0 && lp + sz < obuf + sz_obuf; onum++, lp += sz);
}
}

/********************/
fl_kbclr(kn) short kn;
{
// kn=0 erase line, kn=1 erase last character
while (*kbuf > 0) { ds_strout("\b \b"); kptr--; (*kbuf)--; if (kn) break; }
}

#endif

/********************/

/*** ds functions ***/

/********************/
ds_strout(s) char *s;
{
char ch;
while (ch = *s++) ds_chout(ch); // display 0-terminated string
}

/********************/
ds_chout(ch) char ch;
{
putch(ch); // write ch to console, unbuffered
}

/********************/
ds_chin()
{
char ch;

// accept character from keyboard, unbuffered, no echo
if ((ch = ds_chrd()) == 0 || ch == 0xe0) { ds_chrd(); ch = 0; }
return ch; // return character, or 0 if extended key code
}

/********************/
ds_chrd()
{
return getch(); // accept character from keyboard, unbuffered, no echo
}

/********************/
ds_kbchk()
{
return kbhit(); // return 0 if no character available in keyboard buffer
}

/********************/
ds_create(fp, ahd) char *fp; uns *ahd;
{
short r;

// create file, permission read & write
if ((r = creat(fp, 0x0100 | 0x0080)) == -1) return 1; // error
*ahd = r; return 0;
}

/********************/
ds_open(fp, ahd) char *fp; uns *ahd;
{
short r;

// open file, binary & read-write
if ((r = open(fp, 0x8000 | 0x0002)) == -1) return 1; // error
*ahd = r; return 0;
}

/********************/
ds_close(hd) uns hd;
{
close(hd); // close file, ignore error (invalid handle)
}

/********************/
ds_read(hd, n, buf) uns hd, n; char *buf;
{
short r;

// read n bytes from file
if ((r = read(hd, buf, n)) == 0) return -1; if (r == -1) return 0;
return r; // return number of bytes read, or 0 if error, or -1 if eof
}

/********************/
ds_write(hd, n, buf) uns hd, n; char *buf;
{
short r;

// write n bytes to file
if ((r = write(hd, buf, n)) == -1)
  { k_wrf = 0; ds_strout("Cannot write "); ds_strout(wrfile); cb_exit(); }
return r; // number of bytes written
}

/********************/
long ds_lseek(hd, m, aofst) uns hd, m; long *aofst;
{
long p;

// position file pointer according to given offset and method
if ((p = lseek(hd, *aofst, m)) == -1) { k_wrf = 0; ds_strout("Error accessing ");
  ds_strout(hd == rdhdl ? rdfile : hd == wrhdl ? wrfile : ""); cb_exit(); }
return p; // new offset
}

/********************/
long ds_time()
{
struct tm_time tm;

// return current time in hundredths of second
gettime(&tm); // 32-bit version uses Windows API GetLocalTime
return ((long)(tm.hh * 60 + tm.mm) * 60 + tm.ss) * 100 + tm.cc;
}

/********************/
ds_hour()
{
struct tm_time tm;

gettime(&tm); return tm.hh; // current hour
}

/********************/
ds_tvar()
{
struct tm_time tm;
short n;

// return variable time-dependent value 0 - 999
gettime(&tm); n = tm.ss % 10; return 100 * n + tm.cc;
}

/********************/

